// SWXCH01.cpp : Implementation of CSWXCH01

#include "stdafx.h"
#include "SWXCH01.h"

/////////////////////////////////////////////////////////////////////////////
// CSWXCH01

// **** INTERFACE METHODS

/*
   Integer commands and their equivalent swxch methods:

    10 = set_view()
    20 = set_table()
    30 = remove_table()
    40 = apply_simple_predicate()
    50 = apply_extent()
    60 = apply_trail()
    70 = clear_predicate()

   110 = get_next_record()
   120 = get_join_value()
   130 = get_string_value()
   140 = get_geometry_value()
   150 = get_annotation_value()
   160 = get_dimension_value()

   210 = delete()
   220 = create_detached_record()
   230 = insert_record()
   240 = set_join_value()
   250 = set_string_value()
   260 = set_geometry_value()
   270 = set_annotation_value()
   280 = set_dimension_value()
*/

STDMETHODIMP CSWXCH01::Connect(BSTR ServerName, LONG *Result)
{

   // Connect to TICS "swxch" service
   // Environment variable SWXCH_TCP_PORT must be set, e.g. to 3002

   _bstr_t b;
   b = _bstr_t(ServerName, true);
   char * sname = (char *) b;
   char service[] = "swxch";

   int iStatus;
   thHandle = NULL;

   iStatus = ticsConnect((Str) sname, (Str) service, &thHandle, NULL);
   if (iStatus != TICS_OK)
   {
      // ticsConnect produced error code
      * Result = iStatus;
      return S_OK;
   }

   // Disable autoflush

   ticsSetAutoFlush(thHandle, 0);

   // Initialise acpt

   acptInit();
   ticsMakeAcptStreams(thHandle);

   *Result = 0;

   return S_OK;

}

STDMETHODIMP CSWXCH01::SetView(BSTR ViewName, LONG *Result)
{

   _bstr_t b;
   b = _bstr_t(ViewName, true);
   wchar_t * vname = (wchar_t *) b;

   // Send integer command

   ticsPutInt(thHandle, 10);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) vname);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::SetTable(BSTR TableName, LONG *Result)
{

   _bstr_t b;
   b = _bstr_t(TableName, true);
   wchar_t * tname = (wchar_t *) b;

   // Send integer command

   ticsPutInt(thHandle, 20);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) tname);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::RemoveTable(LONG *Result)
{

   // Send integer command

   ticsPutInt(thHandle, 30);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::ApplySimplePredicate(BSTR FieldName, BSTR Operator, BSTR StringValue, BSTR Option, LONG *Result)
{

   _bstr_t bFName, bOp, bSV, bOpt;
   bFName = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) bFName;
   bOp = _bstr_t(Operator, true);
   wchar_t * op = (wchar_t *) bOp;
   bSV = _bstr_t(StringValue, true);
   wchar_t * sv = (wchar_t *) bSV;
   bOpt = _bstr_t(Option, true);
   wchar_t * opt = (wchar_t *) bOpt;

   // Send integer command

   ticsPutInt(thHandle, 40);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) op);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) sv);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) opt);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::ApplyExtent(BSTR FieldName, DOUBLE XMin, DOUBLE YMin, DOUBLE XMax, DOUBLE YMax, BOOL IncludeOverlaps, BSTR Option, LONG *Result)
{

   _bstr_t bFName, bOpt;
   bFName = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) bFName;
   bOpt = _bstr_t(Option, true);
   wchar_t * opt = (wchar_t *) bOpt;

   // Send integer command

   ticsPutInt(thHandle, 50);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);
   ticsPutFloat(thHandle, XMin);
   ticsFlush(thHandle);
   ticsPutFloat(thHandle, YMin);
   ticsFlush(thHandle);
   ticsPutFloat(thHandle, XMax);
   ticsFlush(thHandle);
   ticsPutFloat(thHandle, YMax);
   ticsFlush(thHandle);
   ticsPutBool(thHandle, IncludeOverlaps);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) opt);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::ApplyTrail(BSTR FieldName, BSTR AppName, BOOL IncludeOverlaps, BSTR Option, LONG* Result)
{

   _bstr_t bFName, bGName, bOpt;
   bFName = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) bFName;
   bGName = _bstr_t(AppName, true);
   wchar_t * gname = (wchar_t *) bGName;
   bOpt = _bstr_t(Option, true);
   wchar_t * opt = (wchar_t *) bOpt;

   // Send integer command

   ticsPutInt(thHandle, 60);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) gname);
   ticsFlush(thHandle);

   ticsPutBool(thHandle, IncludeOverlaps);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) opt);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::ClearPredicate(LONG *Result)
{

   // Send integer command

   ticsPutInt(thHandle, 70);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::GetNextRecord(LONG *Result)
{

   // Send integer command

   ticsPutInt(thHandle, 110);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::GetJoin(BSTR FieldName, LONG *Result)
{

   _bstr_t b;
   b = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) b;

   // Send command

   ticsPutInt(thHandle, 120);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::GetString(BSTR FieldName, BSTR *Value, LONG *Result)
{

   // We're going to assume that any physical or derived field value
   // can be formatted as a string <= 2048 characters.
   // This simplifies the interface.

   _bstr_t b;
   b = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) b;

   // Send integer command

   ticsPutInt(thHandle, 130);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   if (iStatus != 0)
   {
      * Result = iStatus;
      return S_OK;
   }

   // Get value up to 2048 characters
   // Str16 ticsGetString16(TicsHandle handle, Str16 s, UShort maxLength);

   Str16 * s = new Str16[2048];
   Str16 sRet = NULL;

   sRet = ticsGetString16(thHandle, (Str16) s, 2048);
   if (sRet == NULL)
      b = _bstr_t(_T(""));
   else
      b = _bstr_t(sRet);

   * Value = b.copy();
   * Result = 0;
   delete s;
   return S_OK;

}

STDMETHODIMP CSWXCH01::GetGeometry(BSTR FieldName, IGeometry **Value, LONG *Result)
{

   /*

   Retrieves a Smallworld geometry value (simple_point,
   simple_chain, simple_area, point, chain, or area)
   and builds the corresponding Point, Polyline, or Polygon.

   Any curves are densified to line segments.

   */

   _bstr_t b;
   b = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) b;

   // Send integer command

   ticsPutInt(thHandle, 140);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;
   if (iStatus <= 0)
   {
      return S_OK;
   }

   // Get Value

   switch(iStatus)
   {
      case 3:      // simple_point
         iStatus = BuildPoint(Value);
         break;
      case 4:      // simple_chain
         iStatus = BuildLine(Value);
         break;
      case 5:      // simple_area
         iStatus = BuildPoly(Value);
         break;
      case 6:      // point
         iStatus = BuildPoint(Value);
         break;
      case 7:      // chain
         iStatus = BuildLine(Value);
         break;
      case 8:      // area
         iStatus = BuildPoly(Value);
         break;
      default:     // unsupported (you should never reach this)
         * Result = -3;
         break;
   }

   return S_OK;

}

STDMETHODIMP CSWXCH01::GetAnnotation(BSTR FieldName, ITextElement **Value, LONG *Result)
{
   /* Retrieves data from a Smallworld text (annotation) geometry
      and builds a corresponding TextElement.
      Rotation and justification data will be incorporated
      into the associated TextSymbol. */

   _bstr_t b;
   b = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) b;

   // Send integer command

   ticsPutInt(thHandle, 150);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;
   if (iStatus != 1)
      return S_OK;

   // Get coord_1, string, orientation, just

   AcptCoord * c = new AcptCoord;
   Str16 * s = new Str16[2048];
   Str16 sText = NULL;
   double dOrient, dAngle;
   double pi = asin(1.0) * 2.0;
   int iJust, iVert, iHoriz;

   ticsGetObject(thHandle, ACPT_TYPE_COORD, c, sizeof(AcptCoord));
   sText = ticsGetString16(thHandle, (Str16) s, 2048);
   dOrient = ticsGetFloat(thHandle);
   iJust = ticsGetInt(thHandle);

   // Build TextElement and set text

   ITextElementPtr ipTextElement(CLSID_TextElement);
   b = _bstr_t(sText);
   BSTR bsText = b.copy();
   ipTextElement->put_Text(bsText);
   ipTextElement->put_ScaleText(VARIANT_FALSE);

   // Set geometry

   IPointPtr ipPoint(CLSID_Point);
   ipPoint->PutCoords(c->x, c->y);
   IElementPtr ipElement = ipTextElement;
   ipElement->put_Geometry(ipPoint);

   // Set TextSymbol

   ISimpleTextSymbolPtr ipTextSymbol(CLSID_TextSymbol);
   dAngle = dOrient * 180 / pi;
   ipTextSymbol->put_Angle(dAngle);
   iVert = (int) floor((float) iJust / 10.0);
   iHoriz = iJust % 10;
   switch(iVert)
   {
      case 1:      // bottom
         ipTextSymbol->put_VerticalAlignment(esriTVABottom);
         break;
      case 2:      // base
         ipTextSymbol->put_VerticalAlignment(esriTVABaseline);
         break;
      case 3:      // center
         ipTextSymbol->put_VerticalAlignment(esriTVACenter);
         break;
      case 4:      // top
         ipTextSymbol->put_VerticalAlignment(esriTVATop);
         break;
      case 5:      // top
         ipTextSymbol->put_VerticalAlignment(esriTVATop);
         break;
      default:
         break;
   }
   switch(iHoriz)
   {
      case 1:      // left
         ipTextSymbol->put_HorizontalAlignment(esriTHALeft);
         break;
      case 2:      // left
         ipTextSymbol->put_HorizontalAlignment(esriTHALeft);
         break;
      case 3:      // center
         ipTextSymbol->put_HorizontalAlignment(esriTHACenter);
         break;
      case 4:      // right
         ipTextSymbol->put_HorizontalAlignment(esriTHARight);
         break;
      case 5:      // right
         ipTextSymbol->put_HorizontalAlignment(esriTHARight);
         break;
      default:
         break;
   }
   ipTextElement->put_Symbol(ipTextSymbol);

   // Assign value and we're done

   * Value = ipTextElement;
   (* Value)->AddRef();
   delete c;
   delete s;

   return S_OK;

}

STDMETHODIMP CSWXCH01::GetDimension(BSTR FieldName, IGeometry **LineValue, ITextElement **AnnoValue, LONG *Result)
{

   /*

   Retrieves data for 6 Smallworld dimension types in the form of
   a type number (1-6), a text string, and a vector of coordinates,
   and builds a corrsponding Polyline and TextElement representing
   the line path and annotation.  If successful, Result will receive
   the dimension type.

   TYPE 1: direct dimension
   two points define line path
   text is placed at midpoint and angle of segment
   valign bottom, halign center

   TYPE 2: horizontal dimension                   |
   four points define line path                    ----
   text is placed at midpoint of middle segment        |
   angle 0, valign bottom, halign center

   TYPE 3: vertical dimension                          --
   four points define line path                       |
   text is placed at midpoint of middle segment       |
   angle 90, valign bottom, halign center           --

   TYPE 4: parallel dimension
   four points define line path
   text is placed at midpoint and angle of middle segment
   valign bottom, halign center

   TYPE 5: radius dimension
   two points define line path
   text is placed at third point
   angle 0, valign bottom, halign left

   TYPE 6: angle dimension
   three points define line path
   text is placed at fourth point
   angle of first segment, valign botton, halign left

   */

   _bstr_t b;
   b = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) b;

   // Send integer command

   ticsPutInt(thHandle, 160);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;
   if (iStatus != 2)
      return S_OK;

   // Get type and text

   int iType, iSize, iNumR;
   Str16 * s = new Str16[2048];
   Str16 sText = NULL;

   iType = ticsGetInt(thHandle);
   sText = ticsGetString16(thHandle, (Str16) s, 2048);

   switch(iType)
   {
      case 1:
         iSize = 2;
         break;
      case 5:
         iSize = 3;
         break;
      default:
         iSize = 4;
         break;
   }
   AcptCoord * cV = new AcptCoord[iSize];
   ticsGetObjectVector(thHandle, ACPT_TYPE_COORD, cV, sizeof(AcptCoord) * iSize, &iNumR);

   // Build Polyline

   IPolylinePtr ipPolyline(CLSID_Polyline);
   IPointCollectionPtr ipPtCol = ipPolyline;
   int i;

   switch(iType)
   {
      case 1:
      case 5:
         iSize = 2;
         break;
      case 6:
         iSize = 3;
         break;
      default:
         iSize = 4;
         break;
   }

   for (i = 0; i < iSize; i++)
   {
      IPointPtr ipPointB(CLSID_Point);
      ipPointB->PutCoords(cV[i].x, cV[i].y);
      ipPtCol->AddPoint(ipPointB);
   }

   * LineValue = ipPolyline;
   (* LineValue)->AddRef();

   // Build TextElement and set text

   ITextElementPtr ipTextElement(CLSID_TextElement);
   b = _bstr_t(sText);
   BSTR bsText = b.copy();
   ipTextElement->put_Text(bsText);
   ipTextElement->put_ScaleText(VARIANT_FALSE);

   // Set geometry

   double x, y;

   switch(iType)
   {
      case 1:
         x = (cV[0].x + cV[1].x) / 2.0;
         y = (cV[0].y + cV[1].y) / 2.0;
         break;
      case 5:
         x = cV[2].x;
         y = cV[2].y;
         break;
      case 6:
         x = cV[3].x;
         y = cV[3].y;
         break;
      default:
         x = (cV[1].x + cV[2].x) / 2.0;
         y = (cV[1].y + cV[2].y) / 2.0;
         break;
   }
   IPointPtr ipPoint(CLSID_Point);
   ipPoint->PutCoords(x, y);
   IElementPtr ipElement = ipTextElement;
   ipElement->put_Geometry(ipPoint);

   // Build TextSymbol and set angle

   double dAngle, d;
   double pi = asin(1.0) * 2.0;

   ISimpleTextSymbolPtr ipTextSymbol(CLSID_TextSymbol);
   switch(iType)
   {
      case 1:
      case 6:
         x = cV[1].x - cV[0].x;
         y = cV[1].y - cV[0].y;
         break;
      case 3:
         x = 0;
         y = 1;
         break;
      case 4:
         x = cV[2].x - cV[1].x;
         y = cV[2].y - cV[1].y;
         break;
      default:
         x = 1;
         y = 0;
         break;
   }
   d = sqrt((x * x) + (y * y));
   x = x / d;
   y = y / d;
   dAngle = asin(y);
   if (x < 0)
      dAngle = pi - dAngle;
   ipTextSymbol->put_Angle(dAngle);

   // Set justification and finish up

   ipTextSymbol->put_VerticalAlignment(esriTVABottom);
   switch(iType)
   {
      case 6:
         ipTextSymbol->put_HorizontalAlignment(esriTHALeft);
         break;
      default:
         ipTextSymbol->put_HorizontalAlignment(esriTHACenter);
         break;
   }

   ipTextElement->put_Symbol(ipTextSymbol);
   * AnnoValue = ipTextElement;
   (* AnnoValue)->AddRef();
   delete s;
   delete cV;

   * Result = iType;

   return S_OK;

}

STDMETHODIMP CSWXCH01::Delete(LONG *Result)
{

   // Send integer command

   ticsPutInt(thHandle, 210);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::NewDetachedRecord(LONG *Result)
{

   // Send integer command

   ticsPutInt(thHandle, 220);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::InsertRecord(LONG *Result)
{

   // Send integer command

   ticsPutInt(thHandle, 230);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::NewDatalessRWO(DOUBLE XMin, DOUBLE YMin, DOUBLE XMax, DOUBLE YMax, LONG *Result)
{

   // Send integer command

   ticsPutInt(thHandle, 290);
   ticsFlush(thHandle);
   ticsPutFloat(thHandle, XMin);
   ticsFlush(thHandle);
   ticsPutFloat(thHandle, YMin);
   ticsFlush(thHandle);
   ticsPutFloat(thHandle, XMax);
   ticsFlush(thHandle);
   ticsPutFloat(thHandle, YMax);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::PutJoin(BSTR FieldName, BSTR Direction, LONG *Result)
{

   _bstr_t bFName, bDir;
   bFName = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) bFName;
   bDir = _bstr_t(Direction, true);
   wchar_t * dir = (wchar_t *) bDir;

   // Send integer command

   ticsPutInt(thHandle, 240);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) dir);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::PutString(BSTR FieldName, BSTR Value, LONG *Result)
{

   _bstr_t bFName, bVal;
   bFName = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) bFName;
   bVal = _bstr_t(Value, true);
   wchar_t * val = (wchar_t *) bVal;

   // Send integer command

   ticsPutInt(thHandle, 250);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) val);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::PutGeometry(BSTR FieldName, IGeometry *Value, LONG *Result)
{

   /*

   Attempts to send an ESRI geometry to the SWXCH server.  If Value is
   a Point, an AcptCoord is sent.  If Value is a Polyline, an
   AcptSectorRope is sent.  If Value is a Polygon, a series of
   AcptSectorRopes is sent.  Other geometry types are not supported.

   Any curves are densified to line segments.

   If Value is a Polyline with more than one part, only the first part
   is sent.

   Returns 1 if geometry type is unsupported
           2 if there was an error in building the output geometry
           0 if successful
         < 0 if SWXCH objected

   */

   _bstr_t bFName;
   bFName = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) bFName;

   // Package geometry as coordinate, sector_rope, or sector_rope_vector

   HRESULT hr;
   esriGeometryType ShapeType;
   wchar_t * gtype;

   hr = Value->get_GeometryType(&ShapeType);
   switch (ShapeType)
   {

      case esriGeometryPoint:

         gtype = L"point";
         break;

      case esriGeometryPolyline:

         gtype = L"line";
         break;

      case esriGeometryPolygon:

         gtype = L"poly";
         break;

      default:

         // Unsupported type

         * Result = 1;
         return S_OK;

   }

   // Send integer command

   ticsPutInt(thHandle, 260);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) gtype);
   ticsFlush(thHandle);

   switch (ShapeType)
   {

      case esriGeometryPoint:

         PutPoint(Value);
         break;

      case esriGeometryPolyline:

         PutLine(Value);
         break;

      case esriGeometryPolygon:

         PutPoly(Value);
         break;

      default:

         // This will not be reached
         break;

   }

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::PutAnnotation(BSTR FieldName, IPoint *Coord1, BSTR Text, DOUBLE Orientation, LONG Justification, LONG *Result)
{

   _bstr_t bFName;
   bFName = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) bFName;
   _bstr_t bText;
   bText = _bstr_t(Text, true);
   wchar_t * txt = (wchar_t *) bText;
   IGeometryPtr Value = Coord1;

   // Send integer command

   ticsPutInt(thHandle, 270);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);
   PutPoint(Value);
   ticsPutString16(thHandle, (Str16) txt);
   ticsFlush(thHandle);
   ticsPutFloat(thHandle, Orientation);
   ticsFlush(thHandle);
   ticsPutInt(thHandle, Justification);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::PutDimension(BSTR FieldName, LONG Type, BSTR Text, IPointCollection *ControlPoints, LONG *Result)
{

   /*

   FieldName - name of dimension field to set
   Type - Smallworld annotation type
   Text - dimension text
   ControlPoints - control points for dimension
            type 1: Points 1 and 2 indicate feature to dimension.
                 2/3/4: Points 1 and 2 indicate feature to dimension.
                        Point 3 indicates position of dimension line.
                 5: Points 1 and 2 indicate the centre and circumference
                    of the circle to dimension, and the angle of the leader line.
                    Point 3 determines the length of the dimension line.
                 6: Points 1, 2 and 3 indicate the angle to be dimensioned.
                    Point 4 determines the position of the dimension line.

   */

   _bstr_t bFName;
   bFName = _bstr_t(FieldName, true);
   wchar_t * fname = (wchar_t *) bFName;
   _bstr_t bText;
   bText = _bstr_t(Text, true);
   wchar_t * txt = (wchar_t *) bText;

   HRESULT hr;
   long lSize;
   hr = ControlPoints->get_PointCount(&lSize);
   AcptCoord * cList;
   cList = new AcptCoord[lSize];
   IPointPtr ipPoint;
   double dX, dY;
   long i;
   for (i = 0; i < lSize; i++)
   {
      hr = ControlPoints->get_Point(i, &ipPoint);
      hr = ipPoint->get_X(&dX);
      hr = ipPoint->get_Y(&dY);
      cList[i].x = dX;
      cList[i].y = dY;
   }

   // Send integer command

   ticsPutInt(thHandle, 280);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) fname);
   ticsFlush(thHandle);
   ticsPutInt(thHandle, Type);
   ticsFlush(thHandle);
   ticsPutString16(thHandle, (Str16) txt);
   ticsFlush(thHandle);
   ticsPutObjectVector(thHandle, ACPT_TYPE_COORD, cList, lSize);
   ticsFlush(thHandle);

   // Get result

   int iStatus;
   iStatus = ticsGetInt(thHandle);
   * Result = iStatus;

   return S_OK;

}

STDMETHODIMP CSWXCH01::Disconnect()
{

   // Disconnect from TICS

   ticsFreeAcptStreams(thHandle);
   ticsDisconnect(thHandle);
   thHandle = NULL;

   return S_OK;

}

// **** PRIVATE METHODS

int CSWXCH01::BuildPoint(IGeometry ** igValue)
{

   // Get coordinate

   AcptCoord * c = new AcptCoord;
   ticsGetObject(thHandle, ACPT_TYPE_COORD, c, sizeof(AcptCoord));

   // Build Point

   IPointPtr ipPoint(CLSID_Point);
   ipPoint->PutCoords(c->x, c->y);

   * igValue = ipPoint;
   (* igValue)->AddRef();
   delete c;

   return 0;

}

int CSWXCH01::BuildLine(IGeometry ** igValue)
{

   // Get sector_rope

   AcptSectorRope * sr = new AcptSectorRope;
   ticsGetObject(thHandle, ACPT_TYPE_SECTOR_ROPE, sr, sizeof(AcptSectorRope));

   // Build single-part Polyline via IPointCollection

   IPolylinePtr ipPolyline(CLSID_Polyline);
   IPointCollectionPtr ipPtCol = ipPolyline;

   // Loop through sectors and add points

   int i, j, n;
   AcptSector s;

   for (i = 0; i < sr->NSectors; i++)
   {
      AcptCoord * clist = NULL;
      s = sr->Sectors[i];
      n = sg.grind(&s, &clist, 0);
      for (j = 0; j < n; j++)
      {

         // Skip 1st coord of each sector unless first sector

         if ((j != 0) || ((j == 0) && (i == 0)))
         {
         IPointPtr ipPointB(CLSID_Point);
            ipPointB->PutCoords(clist[j].x, clist[j].y);
            ipPtCol->AddPoint(ipPointB);
         }

      }
   }

   * igValue = ipPolyline;
   (* igValue)->AddRef();
   delete sr;

   return 0;

}

int CSWXCH01::BuildPoly(IGeometry ** igValue)
{

   // Get vector of sector_ropes

   int iSize, iNumR;
   iSize = ticsGetInt(thHandle);
   AcptSectorRope * srV = new AcptSectorRope[iSize];
   ticsGetObjectVector(thHandle, ACPT_TYPE_SECTOR_ROPE, srV, sizeof(AcptSectorRope) * iSize, &iNumR);

   // Build Polygon as IGeometryCollection

   IPolygonPtr ipPolygon(CLSID_Polygon);
   IGeometryCollectionPtr ipGeoCol = ipPolygon;

   // Loop through sector_ropes and build rings

   int r, i, j, n;
   AcptSectorRope sr;
   AcptSector s;

   for (r = 0; r < iNumR; r++)
   {

      sr = srV[r];
      IRingPtr ipRing(CLSID_Ring);
      IPointCollectionPtr ipPtCol = ipRing;

      for (i = 0; i < sr.NSectors; i++)
      {
         AcptCoord * clist = NULL;
         s = sr.Sectors[i];
         n = sg.grind(&s, &clist, 0);
         for (j = 0; j < n; j++)
         {

            // Skip 1st coord of each sector unless first sector

            if ((j != 0) || ((j == 0) && (i == 0)))
            {
               IPointPtr ipPointB(CLSID_Point);
               ipPointB->PutCoords(clist[j].x, clist[j].y);
               ipPtCol->AddPoint(ipPointB);
            }

         }
      }

      ipGeoCol->AddGeometry(ipRing);

   }

   ITopologicalOperator2Ptr ipTopo = ipPolygon;
   ipTopo->put_IsKnownSimple(VARIANT_FALSE);
   ipPolygon->SimplifyPreserveFromTo(); // or ipTopo->Simplify()

   * igValue = ipPolygon;
   (* igValue)->AddRef();
   delete srV;

   return 0;

}

void CSWXCH01::PutPoint(IGeometry * igValue)
{

   HRESULT hr;
   IPointPtr ipPoint;
   double dX, dY;
   ipPoint = igValue;
   hr = ipPoint->get_X(&dX);
   hr = ipPoint->get_Y(&dY);
   AcptCoord * c = new AcptCoord;
   c->x = dX;
   c->y = dY;

   ticsPutObject(thHandle, ACPT_TYPE_COORD, c);
   ticsFlush(thHandle);
   delete c;
   return;

}

void CSWXCH01::PutLine(IGeometry * igValue)
{

   HRESULT hr;
   IGeometryCollectionPtr ipGC = igValue;
   IGeometryPtr ipGeom;
   hr = ipGC->get_Geometry(0, &ipGeom);

   PutPath(ipGeom);

}

void CSWXCH01::PutPoly(IGeometry * igValue)
{

   // Make sure polygon is simplified

   HRESULT hr;

   /* IPointCollectionPtr ipPC;
   IPointPtr ipPoint;
   double dX, dY;
	ipPC = igValue;
   hr = ipPC->get_Point(0, &ipPoint);
   hr = ipPoint->get_X(&dX);
   hr = ipPoint->get_Y(&dY); */

   IPolygon4Ptr ipPoly = igValue;
   // ITopologicalOperator2Ptr ipTopo = ipPoly;
   // hr = ipTopo->put_IsKnownSimple(VARIANT_FALSE);
   hr = ipPoly->SimplifyPreserveFromTo();

	/* ipPC = igValue;
   hr = ipPC->get_Point(0, &ipPoint);
   hr = ipPoint->get_X(&dX);
   hr = ipPoint->get_Y(&dY); */

   // Put number of sector_ropes

   IGeometryCollectionPtr ipGC = ipPoly;
   long lSize;
   hr = ipGC->get_GeometryCount(&lSize);
   ticsPutInt(thHandle, lSize);
   ticsFlush(thHandle);

	// If only one ring, send it

   IGeometryPtr ipGeom;
   IPathPtr ipPath;
	if (lSize == 1)
	{
      hr = ipGC->get_Geometry(0, &ipGeom);
		ipPath = ipGeom;
		PutPath(ipPath);
		return;
	}

   // Send polygon rings in order that swxch expects

   long lNumOR, lNumIR;
	IRingPtr ipRing;
	IRingPtr * ipOuterRingArray, * ipInnerRingArray;
   long i, j;

   hr = ipPoly->get_ExteriorRingCount(&lNumOR);
	ipOuterRingArray = new IRingPtr[lNumOR];
	hr = ipPoly->QueryExteriorRingsEx(lNumOR, (IRing **) ipOuterRingArray);
   for (i = 0; i < lNumOR; i++)
   {

		ipRing = ipOuterRingArray[i];
      ipPath = ipRing;
		PutPath(ipPath);
      hr = ipPoly->get_InteriorRingCount(ipRing, &lNumIR);
      if (lNumIR == 0)
         continue;

		ipInnerRingArray = new IRingPtr[lNumIR];
		hr = ipPoly->QueryInteriorRingsEx(ipRing, lNumIR, (IRing **) ipInnerRingArray);
      for (j = 0; j < lNumIR; j++)
      {
			ipRing = ipInnerRingArray[j];
         ipPath = ipRing;
         PutPath(ipPath);
      }
		delete [] ipInnerRingArray;

	}

	delete [] ipOuterRingArray;

}

void CSWXCH01::PutPath(IGeometry * igPath)
{

   HRESULT hr;
   ISegmentCollectionPtr ipSC = igPath;
   IPointCollectionPtr ipPC;
   VARIANT_BOOL vbHasCurves;
   hr = ipSC->HasNonLinearSegments(&vbHasCurves);
   if (vbHasCurves)
      GrindPath(igPath, &ipPC);
   else
      ipPC = igPath;

   long lNumP;
   hr = ipPC->get_PointCount(&lNumP);
   int iNumSecs;
   if (lNumP > 512)
      iNumSecs = 1 + (int) ceil( (double) (lNumP - 512) / 511.0 );
   else
      iNumSecs = 1;

   // Build sector rope

   AcptSectorRope * sr = new AcptSectorRope;
   sr->NSectors = iNumSecs;
   sr->Sectors = new AcptSector[iNumSecs];
   AcptSector * sList = sr->Sectors;

   AcptSector * s;
   AcptSimpleSector * ss;
   AcptCoord * cList;

   int iS = 0;
   int iSP = 0;
   int iNumSP = 0;
   bool bFirstSec = true;
   IPointPtr ipPoint;
   long i;
   double dX, dY;
   for (i = 0; i < lNumP; i++)
   {

      if ((iSP == 0) || (iSP == 512))
      {

         if (lNumP - i > 511)
            iNumSP = 512;
         else
         {
            iNumSP = lNumP - i;
            if (!bFirstSec)
               iNumSP++;
         }

         s = &(sList[iS]);
         s->type = 8;
         ss = &(s->u.SimpleSector);
         ss->NCoords = iNumSP;
         ss->coords = new AcptCoord[iNumSP];
         cList = ss->coords;
         iS++;

         iSP = 0;
         if (bFirstSec)
            bFirstSec = false;
         else
         {
            cList[0].x = dX;
            cList[0].y = dY;
            iSP++;
         }

      }

      hr = ipPC->get_Point(i, &ipPoint);
      hr = ipPoint->get_X(&dX);
      hr = ipPoint->get_Y(&dY);
      cList[iSP].x = dX;
      cList[iSP].y = dY;
      iSP++;

   }

   // Send it out

   ticsPutObject(thHandle, ACPT_TYPE_SECTOR_ROPE, sr);
   ticsFlush(thHandle);

   // Now delete it

   cList = NULL;
   for (i = 0; i < iNumSecs; i++)
   {

      s = &(sList[i]);
      ss = &(s->u.SimpleSector);
      delete ss->coords;

   }

   delete sList;
   delete sr;

}

void CSWXCH01::GrindPath(IGeometry * igPath, IPointCollection ** ipPath)
{

   // Create new path, converting curves to vertices

   int iDefaultNSegs = 32;

   HRESULT hr;
   IPointCollectionPtr ipPC(CLSID_Path);
   ISegmentCollectionPtr ipSC = igPath;
   long lNumSeg;
   hr = ipSC->get_SegmentCount(&lNumSeg);

   ISegmentPtr ipSeg;
   esriGeometryType GeomType;
   ICurvePtr ipCurve;
   ICircularArcPtr ipCArc;
   IEllipticArcPtr ipEArc;

   double dAngle;
   double pi = acos(-1.0);
   int iNumV;
   double dSegLen;
   IPointPtr ipPt;

   // Loop through segments and get vertices

   long i, j;
   for (i = 0; i < lNumSeg; i++)
   {

      hr = ipSC->get_Segment(i, &ipSeg);
      hr = ipSeg->get_GeometryType(&GeomType);
      if ((GeomType == 14) || (GeomType == 15) || (GeomType == 16))
      {

         // GeomType: 14 = CircularArc, 15 = BezierCurve, 16 = EllipticArc

         // Calculate number of vertices (iNumV)

         switch (GeomType)
         {
            case 14:

               ipCArc = ipSeg;
               hr = ipCArc->get_CentralAngle(&dAngle);
               iNumV = (int) ceil(2.0 * (double) iDefaultNSegs * fabs(dAngle) / pi);
               break;

            case 16:

               ipEArc = ipSeg;
               hr = ipEArc->get_CentralAngle(&dAngle);
               iNumV = (int) ceil(2.0 * (double) iDefaultNSegs * fabs(dAngle) / pi);
               break;

            default:

               iNumV = iDefaultNSegs + 1;

         }
			if (iNumV < 2)
				iNumV = 2;

         // Grind curve into iNumV points

         ipCurve = ipSeg;
         for (j = 0; j < iNumV; j++)
         {
            dSegLen = (double) j / (double) iNumV;
            IPointPtr ipPoint(CLSID_Point);
            hr = ipCurve->QueryPoint(esriNoExtension, dSegLen, true, ipPoint);
            ipPC->AddPoint(ipPoint, NULL, NULL);
				// ipPoint->AddRef();
         }

      }
      else
      {
         hr = ipSeg->get_FromPoint(&ipPt);
         ipPC->AddPoint(ipPt, NULL, NULL);
      }

      if (i == lNumSeg - 1)
      {

         // Add final vertex

         hr = ipSeg->get_ToPoint(&ipPt);
         ipPC->AddPoint(ipPt, NULL, NULL);

      }

   }

   * ipPath = ipPC;
	(* ipPath)->AddRef();

}

