/*
 * Source file acpt_tics.cpp
 */

#include "acpt_tics.h"

/**********************************************************************
/######################################################################
/#### tics32 implementation
/######################################################################
/*********************************************************************/

/**********************************************************************
/**** BEGIN tics.c
/**********************************************************************/

#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <stdarg.h>
#include <time.h>
#include <limits.h>
// #include <winsock.h>

/**********************************************************************
/**** BEGIN eswap.c
/**********************************************************************/

/* -------------------------------------------------------------------------
 * FUNCTION   : endSwap2
 * DESCRIPTION: End-swaps a series of 2 byte values
 * NOTE       : Does swap in place
 * ARGUMENTS  :   ptr      Pointer to short aligned values to be swapped
 *    nItems      Number of values to swap
 * ------------------------------------------------------------------------- */

Void            endSwap2(Ptr ptr, UInt nItems)
{
    UShort         *p = (UShort *) ptr;

    while (nItems--) {
   *p = *p << 8 | *p >> 8;        /* right shift *should* be unsigned */
   p++;
    }
}

/* -------------------------------------------------------------------------
 * FUNCTION   : endSwap4
 * DESCRIPTION: End-swaps a series of 4 byte values
 * NOTE       : Does swap in place
 * ARGUMENTS  :   ptr      Pointer to word aligned values to be swapped
 *    nItems      Number of values to swap
 * ------------------------------------------------------------------------- */

Void            endSwap4(Ptr ptr, UInt nItems)
{
    UInt32           *p = (UInt32 *) ptr;

    while (nItems--) {
   *p = *p << 24 | (*p & 0xff00) << 8 | (*p >> 8) & 0xff00 | *p >> 24;
   p++;
    }
}

/* -------------------------------------------------------------------------
 * FUNCTION   : endSwap8
 * DESCRIPTION: End-swaps a series of 8 byte values
 * NOTE       : Does swap in place
 * ARGUMENTS  :   ptr      Pointer to 8 byte aligned values to be swapped
 *    nItems      Number of values to swap
 * ------------------------------------------------------------------------- */

Void            endSwap8(Ptr ptr, UInt nItems)
{
    UInt32           *p = (UInt32 *) ptr,
                    t;

    while (nItems--) {
   t = *p << 24 | (*p & 0xff00) << 8 | (*p >> 8) & 0xff00 | *p >> 24;
   p++;
   p[-1] = *p << 24 | (*p & 0xff00) << 8 | (*p >> 8) & 0xff00 | *p >> 24;
   *p++ = t;
    }
}

/**********************************************************************
/**** END eswap.c
/**********************************************************************/

/* -------------------------------------------------------------------------
 * PRIVATE DEFINITIONS:
 * ------------------------------------------------------------------------- */

/* TICS library version numbers */

#define TICS_VER_MAJOR 1
#define TICS_VER_MINOR 4

/*
 * Change the diagnostic level number to get (lower for more) (higher for
 * less) diagnostic output
 * NOTE: Only ever done when DEBUG defined
 * 1 = Display actual data sent/recv'd
 * 2 = Display data sent/recv'd in local format
 * 4 = Display protocol information from client
 */

#define DIAGNOSTIC_LEVEL 4

#define EXACTLY            1
#define ONCE            2

#define MAX_BUFFER_SIZE       16384

#define SERVER_BIG_ENDIAN(h)     ((h)->endianness == IS_BIG_ENDIAN)
#define SERVER_LITTLE_ENDIAN(h)     ((h)->endianness == IS_LITTLE_ENDIAN)
#define SWAP_ENDIAN(h)        (h->endianness != iThisEndian)

#define SWAP_AND_PUT_INTS(h,p,c,t,f)            \
  if (SWAP_ENDIAN ((h))) {             \
      void *buf;                                                \
      UInt32 uiHowMuch = (UInt32)c * sizeof (t);            \
      if (uiHowMuch>UINT_MAX) {              \
          h->eofHandler(thHandle, TICS_NO_MEMORY);              \
      } else {                   \
     buf = malloc((UInt)uiHowMuch);                        \
     if (buf==NULL) {                                      \
         h->eofHandler(thHandle, TICS_NO_MEMORY);          \
     } else {                 \
         memcpy(buf, (p), (UInt)uiHowMuch);     \
              (f) (buf, c);                    \
              ticsPut ((h), buf, sizeof (t), c);           \
              free(buf);                                        \
     }                     \
      }                       \
  } else                   \
      ticsPut ((h), (p), sizeof (t), (c));

#define SWAP_AND_PUT_FLOATS(h,p,c,t,f)          \
  if (SWAP_ENDIAN ((h))) {             \
      void *buf;                                                \
      UInt32 uiHowMuch = (UInt32)c * sizeof (t);                 \
      if (uiHowMuch>UINT_MAX) {              \
          h->eofHandler(thHandle, TICS_NO_MEMORY);              \
      } else {                   \
          buf = malloc((UInt)uiHowMuch);                        \
          if (buf==NULL) {                                      \
              h->eofHandler(thHandle, TICS_NO_MEMORY);          \
     } else {                 \
               memcpy(buf, (p), (UInt)uiHowMuch);     \
              (f) ((h), (t *)buf, c);           \
              ticsPut ((h), buf, sizeof (t), c);           \
              free(buf);                                        \
     }                     \
      }                       \
  } else                   \
      ticsPut ((h), (p), sizeof (t), (c));

#define SEND_TYPE_DATA(type, number)            \
    if (thHandle->TypeCheck) {               \
   UInt32 DataType=type;               \
   UInt32 Elements=number;             \
   if (SWAP_ENDIAN(thHandle))          \
       endSwap4(&DataType, 1);            \
   ticsPut(thHandle, (Ptr)&DataType, sizeof(DataType), 1);  \
   if (type & TICS_MULTIPLE) {            \
       if (SWAP_ENDIAN(thHandle))            \
           endSwap4(&Elements, 1);           \
       ticsPut(thHandle, (Ptr)&Elements, sizeof(Elements), 1); \
   }                    \
    }

/* -------------------------------------------------------------------------
 * PRIVATE DATA:
 * ------------------------------------------------------------------------- */

static char     RcsId[] = "$Id: tics.c,v 220.7 2000/10/09 10:04:00 grant Exp $";
static char     Sw_Copyright[] = "Copyright (c) Smallworld Systems Ltd 1994";

/* Endian flag for local machine: -1 = Unset, else IS_{BIG,LITTLE}_ENDIAN */
static int      iThisEndian = -1;

/* Number of open connections */
static UInt      ConnectCount = 0;

/* This buffer is used as a temporary scratch area */
#define SCRATCH_BUF_SIZE 1024
static Byte     bScratchBuffer[SCRATCH_BUF_SIZE];

/* -------------------------------------------------------------------------
 * PRIVATE PROTOTYPES:
 * ------------------------------------------------------------------------- */

static void     CheckType(TicsHandle thHandle, UShort DataType,
           UShort Elements);
static Int32    AddToBuffer(TicsHandle thHandle, Str Buffer, Int32 mBuffer);
static Int32    ReadFromSocket(SOCKET soId, Str bBuffer, Int32 mBuffer,
                Int iHow);
static Int32    WriteToSocket(SOCKET soId, Str bBuffer, Int32 nBuffer,
               Int iHow);
static void     eofDefaultHandler(TicsHandle thHandle, Int32 Status);
static void     SwapFloatToServer(TicsHandle thHandle, Float *fpData,
              UShort i);
static void     SwapFloatFromServer(TicsHandle thHandle, Float *fpData,
                UShort i);
static void     SwapDoubleToServer(TicsHandle thHandle, Double *dpData,
               UShort i);
static void     SwapDoubleFromServer(TicsHandle thHandle, Double *dpData,
                 UShort i);
static void     PrintLog(TicsHandle thHandle, char *fnname, char *format,...);

/* =========================================================================
 * WINDOWS DLL INIT/SHUTDOWN FUNCTIONS:
 * ========================================================================= */

#if defined(TICSDLL)
int DLL_CALL    LibMain(hModule, wDataSeg, cbHeapSize, lpszCmdLine)
    HANDLE          hModule;
    WORD            wDataSeg;
    WORD            cbHeapSize;
    LPSTR           lpszCmdLine;
{
    /* nothing special required */
    return 1;
}


int DLL_CALL    _WEP(bSystemExit)
    int             bSystemExit;
{
    /* nothing special required */
    return (1);
}

#endif

/*
 * ========================================================================
 * PRIVATE ROUTINE TO CONNECT & SET UP A SOCKET.
 * ======================================================================== */
static UInt ConnectToTcpService(struct sockaddr_in *siSockAddr, SOCKET *sOut)
{
    SOCKET sNew;
    Int    iNagleOff = 1;
    /* Create a socket */

    sNew = socket(PF_INET, SOCK_STREAM, 0);
    if (sNew == INVALID_SOCKET)
   return TICS_ACCESS_DENIED;

    if (connect(sNew, (struct sockaddr *) siSockAddr, sizeof(*siSockAddr)) != 0) {
   closesocket(sNew);
   return TICS_SERVICE_UNAVAILABLE;
    }
    /*
     * Now disable the Nagle algorithm.
     */
    setsockopt(sNew, IPPROTO_TCP, TCP_NODELAY,
               (char *)&iNagleOff, sizeof(iNagleOff));
    *sOut = sNew;
    return TICS_OK;

}

/* =========================================================================
 * PUBLIC CONNECTION/DISCONNECTION FUNCTIONS:
 * ========================================================================= */

Int32 DLL_CALL    ticsConnectRegisteredService(Str registryRoot, Str szService, TicsHandle *pthHandle,
                 Str szOpts)
{
#ifdef WIN32

#define REG_BUF_SIZE 256
    LONG status;
    HKEY hKey;
    DWORD index, dataSize, dataType, keySize;
    int portNo;
    Char portstr[8];

    TCHAR valueKey[REG_BUF_SIZE];

    /* The default must match that in connection_service_manager */
    if (registryRoot == NULL || strlen((char *)registryRoot) == 0)
      registryRoot = (Str) "Software\\Smallworld\\Active Services";

    if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_CURRENT_USER,(LPCSTR)registryRoot,0,KEY_EXECUTE,&hKey))
      return TICS_ACCESS_DENIED;

    index = 0;
    do {
   dataSize = sizeof(portNo);
   keySize = REG_BUF_SIZE;
   status = RegEnumValue(hKey,index++,valueKey,&keySize,
               NULL,&dataType,(LPBYTE)&portNo,&dataSize);
   if (ERROR_SUCCESS != status) break;

   /* Make sure we have the expected data type */
   if (!strcmp((char *) valueKey,(char *) szService) && dataType == REG_DWORD) {
       break;
   }
    } while (ERROR_SUCCESS == status);

    RegCloseKey(hKey);

    if (ERROR_SUCCESS != status) return TICS_SERVICE_UNKNOWN;

    sprintf((char *)portstr,"%d",portNo);

    return ticsConnect((Str) "",portstr,pthHandle,szOpts);
#else
    /* Not supported under Unix */
    return TICS_SERVICE_UNKNOWN;
#endif
}

Int32 DLL_CALL    ticsConnect(Str szHost, Str szService, TicsHandle *pthHandle,
               Str szOpts)
/*
 * DESCRIPTION:
 * Connect to a listening TICS server.
 *
 * ARGUMENTS:
 * Char     *szHost     Hostname of machine running server
 *       *szService  Name of listening service
 * TicsHandle  *pthHandle  Pointer for returned handle
 * Char     *szOpts     Any further options
 *
 * RETURNS:
 * Int32       iResult     Status as per tics.h
 *
 */
{
    struct sockaddr_in siAddress;
    struct hostent *phHost;
    struct servent *psService;
    Char            szEnvU[100];
    Char            szEnvL[100];
    Char           *pDispatcherService;
    Char            szDispatcherPort[100];
    Char           *pServiceToLookup;
    Char           *cEnvU,
                   *cEnvL,
                   *cService,
                   *pEnv;
    Int             iService,
                    lEnv,
                    iEndianness;
    UShort          usService;
    SOCKET          sIn;
    UInt            uiStatus = TICS_OK;
    UShort          usBufferSize,
                    usTemp;
    Char            bFormat,
                    bLen,
                    bFlag;
    Byte            bData[256];
    Char            bInitial[3],
                    bProtocol;
    Byte            bConnectMsg[272];
    Bool            TypeCheck;
    Ptr             pBuffer;
    ULong           ulServerAddr;
    UInt            uiServiceIdLen;
    UInt            uiConnectMsgLen;
#ifdef   WINDOWS
    WSADATA         WSAData;
#endif

    if (ConnectCount == 0) {

   Int             iTest = 1;

   iThisEndian = (*((Byte *) &iTest)) ? IS_BIG_ENDIAN : IS_LITTLE_ENDIAN;


    }
    ConnectCount++;

#ifdef   WINDOWS
    /* Windows has special startup and shutdown code */
    if (WSAStartup(0x0101, &WSAData) != 0)
   return TICS_NETWORK_DOWN;
#endif

    /* Deal with the host name */

    if (szHost == NULL) {
   /* This could do a name lookup for the specified service */
   return TICS_HOST_UNKNOWN;
    }
    if (szHost[0] == '\0') {
   /* Assume the application means "localhost" */
   Char            szName[MAXHOSTNAMELEN + 1];

   if (gethostname((char *)szName, sizeof(szName) - 1) != 0)
       return TICS_HOST_UNKNOWN;
   if ((phHost = gethostbyname((char *) szName)) == NULL)
       return TICS_HOST_UNKNOWN;
   ulServerAddr = *(ULong *) phHost->h_addr_list[0];

    } else {
   /* A host name was hopefully specified */
   if ((phHost = gethostbyname((char *) szHost)) != NULL) {
       ulServerAddr = *(ULong *) phHost->h_addr_list[0];
   } else {
       /* Maybe it is an address in xxx.xxx.xxx.xxx notation? */
       ulServerAddr = inet_addr((char *) szHost);
       if (ulServerAddr < 0) {
      /* Exhausted possabilites as to server address format */
      return TICS_HOST_UNKNOWN;
       }
   }
    }

    pDispatcherService = (Char *)strchr((char *)szService, '/');
    if (pDispatcherService == NULL) {
        pServiceToLookup = szService;
        lEnv =  strlen((char *) szService) + 20;     /* For _TCP_PORT etc */
    } else {
        pServiceToLookup = szDispatcherPort;
        lEnv = pDispatcherService - szService;
        if (lEnv > sizeof(szDispatcherPort) )
            return TICS_NO_MEMORY;
        memcpy(pServiceToLookup, szService, lEnv);
        pServiceToLookup[lEnv] = '\0';
        lEnv += 20;
        /* Advance over the delimiter */
        pDispatcherService++;
        uiServiceIdLen = strlen((char *)pDispatcherService);
        uiConnectMsgLen = uiServiceIdLen + 12; /* 12 is the size of the header
                                                * in bytes. */

        /* The service name length is passed as a counted string in the
         * connection message with the count stored in a single byte.
         * So the maximum service name length is the maximum value
         * representable in a single byte i.e. 255
         */
        if (uiServiceIdLen > 255)
            return TICS_NO_MEMORY;

        memcpy(bConnectMsg+12, pDispatcherService, uiServiceIdLen);
        bConnectMsg[0] = 0; /* Must be zero. */
        bConnectMsg[1] = (Byte)  (uiConnectMsgLen >> 8) ;
        bConnectMsg[2] = (Byte) uiConnectMsgLen;
        bConnectMsg[3] = 1; /* Major Protocol Version */
        bConnectMsg[4] = 0; /* Minor Protocol Version */
        /*
         * The next 6 bytes must zero to ensure compatibility with
         * future enhancements.
         */
        bConnectMsg[5] = 0;
        bConnectMsg[6] = 0;
        bConnectMsg[7] = 0;
        bConnectMsg[8] = 0;
        bConnectMsg[9] = 0;
        bConnectMsg[10] = 0;
        bConnectMsg[11] = uiServiceIdLen;
    }
    if (lEnv > sizeof(szEnvU))
   return TICS_NO_MEMORY;

    /* Deal with the service name (env <= uppercased and appended _TCP_PORT) */

    for (cEnvU = szEnvU, cEnvL = szEnvL, cService = pServiceToLookup;
    *cService != '\0';
    cEnvU++, cEnvL++, cService++) {
   *cEnvU = toupper(*cService);
   *cEnvL = tolower(*cService);
    }
    *cEnvU=*cEnvL='\0';
    strcpy((char *) cEnvU, "_TCP_PORT");
    pEnv = (Char *) getenv((char *) szEnvU);
    if (pEnv) {
   /* Environment variable set so use that */
   iService = (Int) strtol((char *) pEnv, (char **) &cEnvU, 0);
   if ((iService < 1024)||(iService > 65535))
       return TICS_SERVICE_UNKNOWN;
   usService = iService;
   iService = htons(usService);
    } else {
   /* try services file - lower case */
   psService = getservbyname((char *) szEnvL, "tcp" /* FOR NOW */ );
   if (psService != NULL) {
       iService = psService->s_port;
   } else {
       /* try interpreting name as a port number */
       iService = (Int) strtol((char *)pServiceToLookup, (char **)&cService, 0);
       if ((iService < 1024)||(iService > 65535))
      return TICS_SERVICE_UNKNOWN;
       usService = iService;
       iService = htons(usService);
   }
    }

    /* Go for it, try and connect to this service */

    memcpy(&(siAddress.sin_addr), &ulServerAddr, 4);
    siAddress.sin_family = AF_INET;
    siAddress.sin_port = iService;

    uiStatus = ConnectToTcpService(&siAddress, &sIn);
    if (uiStatus != TICS_OK)
      return uiStatus;
    /*
     * If we are really got to connect to the service via a dispatcher,
     * the service name is pointed to by pDispatcherService.
     */
    if (pDispatcherService != NULL) {
        /*
         * The dispatcher protocol packet is now read in and compared against
         * what is expected.
         */
const static Byte bExpectedDispatcherInitial[] = {DISPATCHER_PROTOCOL,
                                               DISPATCHER_NAME_LENGTH,
                                               's', 'w', 'd', 'p'};

        Byte bDispatcherInitial[sizeof(bExpectedDispatcherInitial)];
        Byte bDispatcherMsg[256]; /* Should be more than enough. */
        Int32 iLength;
        Int32 iReadMsgLen;
        iLength = ReadFromSocket(sIn, bDispatcherInitial, sizeof(bDispatcherInitial), EXACTLY);
      if (iLength != sizeof(bDispatcherInitial)) {
       closesocket(sIn);
       return TICS_ACCESS_DENIED;
        }
        if (memcmp(bDispatcherInitial, bExpectedDispatcherInitial,
                   sizeof(bDispatcherInitial)) ) {
       closesocket(sIn);
       return TICS_ACCESS_DENIED;
        }
        /*
         * Now send the dispatcher the ID of the service required. This
         * counted string with a 1 byte count which is sent first.
         */
        if (WriteToSocket(sIn, (Str) bConnectMsg,
                          uiConnectMsgLen, EXACTLY) != (Int)uiConnectMsgLen) {
            closesocket(sIn);
       return TICS_ACCESS_DENIED;
        }

        iLength = ReadFromSocket(sIn, bDispatcherMsg, 2, EXACTLY);
        if (iLength != 2) {
       closesocket(sIn);
       return TICS_ACCESS_DENIED;
        }

        iReadMsgLen = ((UInt32)bDispatcherMsg[0] << 8) | bDispatcherMsg[1];
        if (iReadMsgLen > sizeof(bDispatcherMsg) ) {
            /*
             * The dispatcher and the client have ended up talking at
             * cross purposes. The dispatcher has returned far too much
             * data. So abort the connection.
             */
            DEBUG_PRINTF(1,(stderr, "ticsConnect - dispatcher returning too much data -%d.\n",  iReadMsgLen));
            closesocket(sIn);
            return TICS_ACCESS_DENIED;
        }
        iLength = ReadFromSocket(sIn,  bDispatcherMsg+2, iReadMsgLen-2, EXACTLY);
        if (iLength != iReadMsgLen - 2) {
       closesocket(sIn);
      return TICS_ACCESS_DENIED;
        }
        DEBUG_PRINTF(1, (stderr,"Dispatcher returned message type %d Len %d\n",
                         iReadMsgLen, bDispatcherMsg[3]));
        switch (bDispatcherMsg[2]) {
            case DISPATCHER_CONNECTION_OK:
                DEBUG_PRINTF(1, (stderr, "Using the dispatcher to route the data.\n"));
                break;

            case DISPATCHER_CONNECTION_REDIRECT:
                DEBUG_PRINTF(1, (stderr, "Dispatcher has issued a redirection request.\n"));
                if (iReadMsgLen < 10) {
                    closesocket(sIn);
                    return TICS_ACCESS_DENIED;
                }
                else {
                    SOCKET sServer;
                    struct sockaddr_in siServer;
                    Byte bReply[3];

                    memset(&siServer, 0, sizeof(siServer));
                    memcpy(&siServer.sin_port, bDispatcherMsg + 4, 2);
                    memcpy(&siServer.sin_addr, bDispatcherMsg + 6, 4);
                    siServer.sin_family = AF_INET;

                    uiStatus = ConnectToTcpService(&siServer, &sServer);
                    bReply[0] = 0;
                    bReply[1] = 3;
                    if ( uiStatus == TICS_OK )
                      bReply[2] = DISPATCHER_REDIRECTION_OK;
                    else
                      bReply[2] = DISPATCHER_REDIRECTION_FAILED;

                    if (WriteToSocket(sIn, bReply, sizeof(bReply),
                                      EXACTLY) != (Int)sizeof(bReply)) {
                        if (uiStatus == TICS_OK)
                          closesocket(sServer);
                        closesocket(sIn);
                   return TICS_ACCESS_DENIED;
                    }
                    closesocket(sIn);
                    sIn = sServer;
                }
                break;

            case DISPATCHER_CONNECTION_PROTOCOL_ERROR:
                DEBUG_PRINTF(1, (stderr,
                             "Dispatcher protocol error - max protocol is %d\n",
                             bDispatcherMsg[4]));
                closesocket(sIn);
                return TICS_PROTOCOL_MISMATCH;
                break;

            case DISPATCHER_CONNECTION_ERROR:
                DEBUG_PRINTF(1, (stderr, "Dispatcher has detected an error.\n"));
                closesocket(sIn);
                return TICS_ACCESS_DENIED;
                break;

            default:
                DEBUG_PRINTF(1, (stderr, "Dispatcher has returned an unknown status.\n"));
                closesocket(sIn);
                return TICS_ACCESS_DENIED;
                 break;
        }
    }
    /*
     * Read protocol header and transport and connection specific options. We
     * have to be careful about returning early from this routine from now.
     * remember there is a guy at the other end of our socket waiting for
     * replies from us. Only exit if the comms break down, or when the
     * handshaking is finished and its safe to do so.
     */

    /*
     * Initial handshaking is as follows:
     *
     * BYTE NO.   CONTENTS
     * 0    Protocol Id.
     * 1    Format (endianess & float format)
     *
     * 2    length of 1st header block=H1
     * 3    start of 1st header block
     * 3+H1    end of 1st header block
     *
     * 4+H1       length of 2nd header block=H2
     * 5+H1    start of 2nd header block
     * 5+H1+H2       end of 2nd header block
     *
     */

    /* get 1st 3 bytes (see above ) */

    if (ReadFromSocket(sIn, bInitial, sizeof(bInitial), EXACTLY)
   != sizeof(bInitial)) {
   closesocket(sIn);
   return TICS_ACCESS_DENIED;
    }
    bProtocol = bInitial[0];
    bFormat = bInitial[1];

    iEndianness = (bFormat & 0x03) == 0x00 ? IS_LITTLE_ENDIAN :
   (bFormat & 0x03) == 0x01 ? IS_BIG_ENDIAN :
   (bFormat & 0x03) == 0x02 ? IS_PDP_11_INTEGER :
   IS_UNKNOWN;

    if ( (bProtocol<MIN_PROTOCOL) || (bProtocol>MAX_PROTOCOL) ||
    (iEndianness == IS_UNKNOWN) )
   uiStatus = TICS_PROTOCOL_MISMATCH;

    /* blen is length of 1st block + 1 for length of second block */

    bLen = bInitial[2] + 1;          /* Extra byte for "next" block count */

    DEBUG_PRINTF(1, (stderr, "First block %d, %d, %d\n",
           bProtocol, bFormat, bLen));
    DEBUG_PRINTF(4, (stderr, "Host is %s Endian\n",
           HOST_BIG_ENDIAN ? "Big" : "Little"));
    DEBUG_PRINTF(4, (stderr, "Server is %s Endian\n",
           iEndianness == IS_BIG_ENDIAN ? "Big" :
           iEndianness == IS_LITTLE_ENDIAN ? "Little" : "PDP-11"));

    /* get 1st block */

    if (ReadFromSocket(sIn, bData, bLen, EXACTLY) != bLen) {
   closesocket(sIn);
   return TICS_ACCESS_DENIED;
    }

    /* Get type_check option */

    if (bLen>1)
   TypeCheck = bData[0];
    else
   TypeCheck = 0;    /* server or swtics doesn't support type_check */

    /* blen is length of 2nd block + 2 for length of usBufferSize param */

    bLen = bData[bLen - 1] + 2;         /* Extra bytes for max. buffer size */

    DEBUG_PRINTF(1, (stderr, "Second block %d\n", bLen));

    if (ReadFromSocket(sIn, bData, bLen, EXACTLY) != bLen) {
   closesocket(sIn);
   return TICS_ACCESS_DENIED;
    }
    memcpy(&usBufferSize, &(bData[bLen - 2]), 2);
    if (iEndianness != iThisEndian)
   endSwap2(&usBufferSize, 1);

    DEBUG_PRINTF(1, (stderr, "Third block %d %d\n", usBufferSize, TypeCheck));

    if (usBufferSize > MAX_BUFFER_SIZE)
   usBufferSize = MAX_BUFFER_SIZE;

    if (uiStatus == TICS_OK) {
   if ((pBuffer = malloc(usBufferSize)) == NULL) {
       uiStatus = TICS_NO_MEMORY;
   }
    }
    if (uiStatus != TICS_OK)
   usBufferSize = 0;
    usTemp = usBufferSize;
    if (iEndianness != iThisEndian)
   endSwap2(&usTemp, 1);

    /* Reply with appropriate status or buffer size */

    if (WriteToSocket(sIn, (Str) &usTemp, sizeof(usTemp), EXACTLY)
   != sizeof(usTemp) ||
   ReadFromSocket(sIn, &bFlag, sizeof(bFlag), EXACTLY)
   != sizeof(bFlag) ||
   bFlag != TICS_OK) {
   closesocket(sIn);
   free(pBuffer);
   return TICS_ACCESS_DENIED;
    }
    if ((*pthHandle = (TicsHandle) malloc(sizeof(TicsHandleRec))) == NULL) {
   closesocket(sIn);
   free(pBuffer);
   return TICS_NO_MEMORY;
    }

    (*pthHandle)->in = sIn;
    (*pthHandle)->out = sIn;
    (*pthHandle)->eofHandler = eofDefaultHandler;
    (*pthHandle)->endianness = iEndianness;
    (*pthHandle)->autoFlush = 1;
    (*pthHandle)->buffer = (Char *) pBuffer;
    (*pthHandle)->curBufferSize = 0;
    (*pthHandle)->maxBufferSize = usBufferSize;
    (*pthHandle)->bProtocol = bProtocol;
    (*pthHandle)->Logfile = NULL;
    (*pthHandle)->TypeCheck = TypeCheck;
    (*pthHandle)->Status = TICS_OK;

    return TICS_OK;
}

UShort DLL_CALL   ticsEstablishProtocol(TicsHandle thHandle, UShort uMin, UShort uMax)
/*
 * DESCRIPTION:
 * Verify client and server are compatible
 *
 * ARGUMENTS:
 * TicsHandle  *pthHandle  Pointer for TICS handle
 * UShort   Min      Minimum protocol understood
 * UShort   Max      Maximum protocol understood
 *
 * RETURNS:
 * UShort         iResult     Selected protocol
 */
{
    UShort          usProtocol;

    while (TRUE) {
   usProtocol = ticsGetUShort(thHandle);
   /* Proxy may disconnect if protocol not suitable so.
    * Ordinarily this should call error handler which should exit.
    * But with DLL may not exit, so return if an error occurred to
    * prevent an infinite loop */
   if (thHandle->Status != TICS_OK) return 0;
   if (usProtocol >= uMin && usProtocol <= uMax) {
       ticsPutBool(thHandle, 1);
       ticsFlush(thHandle);
       PrintLog(thHandle, "EstablishProtocol",
           "min=%d, max=%d, selected=%d",
           uMin, uMax, usProtocol);
       return usProtocol;
   }
   ticsPutBool(thHandle, 0);
   ticsPutUShort(thHandle, uMin);
   ticsPutUShort(thHandle, uMax);
   ticsFlush(thHandle);
    }
}

Void DLL_CALL   ticsDisconnect(TicsHandle thHandle)
/*
 * DESCRIPTION:
 * Disconnect from server
 *
 * ARGUMENTS:
 * TicsHandle  *pthHandle  Pointer for TICS handle
 */
{
    PrintLog(thHandle, "Disconnect", "");

    if ((thHandle->in == INVALID_SOCKET) || (ConnectCount == 0))
   return;
    shutdown(thHandle->in, 2);
    closesocket(thHandle->in);
    if (thHandle->buffer)
   free(thHandle->buffer);

    /*
     * This next line looks pointless since we are about to free the memory
     * block the slot is in, however it will often catch the error of trying
     * to disconnect the same connection twice. Note: doing this is not
     * supported!
     */

    thHandle->in = INVALID_SOCKET;

    free(thHandle);

    ConnectCount--;

#ifdef WINDOWS
    if (WSACleanup() != 0) {
   return;
    }
#endif

    return;
}

/* =========================================================================
 * PUBLIC GET FUNCTIONS:
 * ========================================================================= */

Void DLL_CALL   ticsGet(TicsHandle thHandle, Ptr vpBuffer,
         UShort uiElementSize, UShort uiNElements)
{
    Int32           nBuffer,
                    mBuffer = (Int32)uiElementSize * uiNElements,
                l;
    Char           *p;

    if (thHandle->autoFlush)
   ticsFlush(thHandle);
    nBuffer = ReadFromSocket(thHandle->in, (Str) vpBuffer, mBuffer, EXACTLY);

    PrintLog(thHandle, "Get", "element size=%d, number=%d, bytes read=%d", uiElementSize, uiNElements, nBuffer);

    if (nBuffer == mBuffer)
   return;

    DEBUG_PRINTF(1, (stderr, "ticsGet, read %d, expected %d\n",
           nBuffer, mBuffer));
    thHandle->eofHandler(thHandle, TICS_DISCONNECTED);

    /* Must be continuing after error? Return zeros */

    l = (Int32)uiElementSize*uiNElements;
    for (p=(Char *)vpBuffer; l>0; p++,l--) *p=0;

    return;
}

Byte DLL_CALL    ticsGetByte(TicsHandle thHandle)
{
    Byte            bOne;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_BYTE, 1);

    ticsGet(thHandle, (Ptr) &bOne, sizeof(Byte), 1);

    PrintLog(thHandle, "GetByte", "%d", bOne);

    return bOne;
}

Short DLL_CALL    ticsGetShort(TicsHandle thHandle)
{
    Short           sOne;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_SHORT, 1);

    ticsGet(thHandle, (Ptr) &sOne, sizeof(Short), 1);
    if (SWAP_ENDIAN(thHandle))
   endSwap2(&sOne, 1);

    PrintLog(thHandle, "GetShort", "%d", sOne);

    return sOne;
}

UShort DLL_CALL   ticsGetUShort(TicsHandle thHandle)
{
    UShort          usOne;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_USHORT, 1);

    ticsGet(thHandle, (Ptr) &usOne, sizeof(UShort), 1);
    if (SWAP_ENDIAN(thHandle))
   endSwap2(&usOne, 1);

    PrintLog(thHandle, "GetUShort", "%d", usOne);

    return usOne;
}

Int32 DLL_CALL  ticsGetInt(TicsHandle thHandle)
{
    Int32           iOne;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_INT, 1);

    ticsGet(thHandle, (Ptr) &iOne, sizeof(Int32), 1);
    if (SWAP_ENDIAN(thHandle))
   endSwap4(&iOne, 1);

    PrintLog(thHandle, "GetInt", "%d", iOne);

    return iOne;
}

UInt32 DLL_CALL ticsGetUInt(TicsHandle thHandle)
{
    UInt32          uiOne;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_UINT, 1);

    ticsGet(thHandle, (Ptr) &uiOne, sizeof(UInt32), 1);
    if (SWAP_ENDIAN(thHandle))
   endSwap4(&uiOne, 1);

    PrintLog(thHandle, "GetUInt", "%d", uiOne);

    return uiOne;
}

Float DLL_CALL ticsGetShortFloat(TicsHandle thHandle)
{
    Float           fOne;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_SFLOAT, 1);

    ticsGet(thHandle, (Ptr) &fOne, sizeof(Float), 1);
    SwapFloatFromServer(thHandle, &fOne, 1);
    DEBUG_PRINTF(2, (stderr, "GetShortFloat => %f\n", fOne));

    PrintLog(thHandle, "GetShortFloat", "%f", fOne);

    return fOne;
}

Double DLL_CALL ticsGetFloat(TicsHandle thHandle)
{
    Double          dOne;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_FLOAT, 1);

    ticsGet(thHandle, (Ptr) &dOne, sizeof(Double), 1);
    SwapDoubleFromServer(thHandle, &dOne, 1);
    DEBUG_PRINTF(2, (stderr, "GetFloat => %f\n", dOne));

    PrintLog(thHandle, "GetFloat", "%f", dOne);

    return dOne;
}

#if defined(HAVE_64_BIT_INTS)
Int64 DLL_CALL  ticsGetLong(TicsHandle thHandle)
{
    Int64           iOne;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_LONG, 1);

    ticsGet(thHandle, (Ptr) &iOne, sizeof(Int64), 1);
    if (SWAP_ENDIAN(thHandle))
   endSwap8(&iOne, 1);

    PrintLog(thHandle, "GetLong", FORMAT_LONG, iOne);

    return iOne;
}

UInt64 DLL_CALL ticsGetULong(TicsHandle thHandle)
{
    UInt64          uiOne;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_ULONG, 1);

    ticsGet(thHandle, (Ptr) &uiOne, sizeof(UInt64), 1);
    if (SWAP_ENDIAN(thHandle))
   endSwap8(&uiOne, 1);

    PrintLog(thHandle, "GetULong", FORMAT_ULONG, uiOne);

    return uiOne;
}
#endif

Void DLL_CALL   ticsGetShorts(TicsHandle thHandle, Short *ptr, UShort count)
{
    PrintLog(thHandle, "GetShorts", "number=%d", count);

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_SHORTS, count);

    ticsGet(thHandle, (Ptr) ptr, sizeof(Short), count);
    if (SWAP_ENDIAN(thHandle))
   endSwap2(ptr, count);
}

Void DLL_CALL   ticsGetUShorts(TicsHandle thHandle, UShort *ptr, UShort count)
{
    PrintLog(thHandle, "GetUShorts", "number=%d", count);

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_USHORTS, count);

    ticsGet(thHandle, (Ptr) ptr, sizeof(UShort), count);
    if (SWAP_ENDIAN(thHandle))
   endSwap2(ptr, count);
}

Void DLL_CALL   ticsGetInts(TicsHandle thHandle, Int32 *ptr, UShort count)
{
    PrintLog(thHandle, "GetInts", "number=%d", count);

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_INTS, count);

    ticsGet(thHandle, (Ptr) ptr, sizeof(Int32), count);
    if (SWAP_ENDIAN(thHandle))
   endSwap4(ptr, count);
}

Void DLL_CALL   ticsGetUInts(TicsHandle thHandle, UInt32 *ptr, UShort count)
{
    PrintLog(thHandle, "GetUInts", "number=%d", count);

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_UINTS, count);

    ticsGet(thHandle, (Ptr) ptr, sizeof(UInt32), count);
    if (SWAP_ENDIAN(thHandle))
   endSwap4(ptr, count);
}

Void DLL_CALL   ticsGetShortFloats(TicsHandle thHandle, Float *ptr, UShort count)
{
    PrintLog(thHandle, "GetShortFloats", "number=%d", count);

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_SFLOATS, count);

    ticsGet(thHandle, (Ptr) ptr, sizeof(Float), count);
    SwapFloatFromServer(thHandle, ptr, count);
}

Void DLL_CALL   ticsGetFloats(TicsHandle thHandle, Double *ptr, UShort count)
{
    PrintLog(thHandle, "GetFloats", "number=%d", count);

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_FLOATS, count);

    ticsGet(thHandle, (Ptr) ptr, sizeof(Double), count);
    SwapDoubleFromServer(thHandle, ptr, count);
}

#if defined(HAVE_64_BIT_INTS)
Void DLL_CALL   ticsGetLongs(TicsHandle thHandle, Int64 *ptr, UShort count)
{
    PrintLog(thHandle, "GetLongs", "number=%d", count);

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_LONGS, count);

    ticsGet(thHandle, (Ptr) ptr, sizeof(Int64), count);
    if (SWAP_ENDIAN(thHandle))
   endSwap8(ptr, count);
}

Void DLL_CALL   ticsGetULongs(TicsHandle thHandle, UInt64 *ptr, UShort count)
{
    PrintLog(thHandle, "GetULongs", "number=%d", count);

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_ULONGS, count);

    ticsGet(thHandle, (Ptr) ptr, sizeof(UInt64), count);
    if (SWAP_ENDIAN(thHandle))
   endSwap8(ptr, count);
}
#endif

Str DLL_CALL    ticsGetString(TicsHandle thHandle, Str s, UShort maxLength)
/*
 * DESCRIPTION:
 * Retrieve string from server
 *
 * NOTE:
 * If string is too long to fit in space provided then the string is
 * truncated and the remainder thrown away.
 *
 * ARGUMENTS:
 * TicsHandle  *pthHandle  Pointer for TICS handle
 * Str       s    Storage for returned string
 * UShort    maxLength  Size of storage area
 *
 * RETURNS:
 * Str       Result          Returned string or NULL if truncated
 */
{
    UShort          usLength,
                    usReturn;
    Int             iTruncate;
    UShort          uiRemain;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_STRING, 1);

    /* Get length of string */

    ticsGet(thHandle, &usLength, sizeof(UShort), 1);
    if (SWAP_ENDIAN(thHandle))
   endSwap2(&usLength, 1);
    iTruncate = usLength > (maxLength - 1);
    usReturn = iTruncate ? maxLength - 1 : usLength;

    /* Get string */

    if (usReturn > 0)
   ticsGet(thHandle, s, sizeof(Char), usReturn);
    s[usReturn] = '\0';

    if (iTruncate) {
   uiRemain = usLength - (maxLength - 1); /* For NULL terminator */
   /* Get the protocol back in step */
   while (uiRemain > SCRATCH_BUF_SIZE) {
       ticsGet(thHandle, bScratchBuffer, sizeof(Byte), SCRATCH_BUF_SIZE);
       uiRemain -= SCRATCH_BUF_SIZE;
   }
   ticsGet(thHandle, bScratchBuffer, sizeof(Byte), uiRemain);
    }
    PrintLog(thHandle, "GetString", "maxlen=%d, sent=%d, read=%d, \"%s\"", maxLength, usLength, usReturn, s);

    return iTruncate ? NULL : s;
}

Str16 DLL_CALL    ticsGetString16(TicsHandle thHandle, Str16 s, UShort maxLength)
/*
 * DESCRIPTION:
 * Retrieve string from server
 *
 * NOTE:
 * If string is too long to fit in space provided then the string is
 * truncated and the remainder thrown away.
 *
 * ARGUMENTS:
 * TicsHandle  *pthHandle  Pointer for TICS handle
 * Str16  s    Storage for returned string
 * UShort    maxLength  Size of storage area
 *
 * RETURNS:
 * Str16  Result          Returned string or NULL if truncated
 */
{
    UShort          usLength,
                    usReturn;
    Int             iTruncate;
    UShort          uiRemain;

    if (thHandle->TypeCheck) CheckType(thHandle, TICS_STRING16, 1);

    /* Get length of string */

    ticsGet(thHandle, &usLength, sizeof(UShort), 1);
    if (SWAP_ENDIAN(thHandle))
   endSwap2(&usLength, 1);
    iTruncate = usLength > (maxLength - 1);
    usReturn = iTruncate ? maxLength - 1 : usLength;

    /* Get string */

    if (usReturn > 0) {
   ticsGet(thHandle, s, sizeof(Char16), usReturn);
   if (SWAP_ENDIAN(thHandle)) endSwap2(s, usReturn);
    }

    s[usReturn] = '\0';

    if (iTruncate) {
   uiRemain = (usLength - maxLength + 1)*2; /* +1 For NULL terminator */
   /* Get the protocol back in step */
   while (uiRemain > SCRATCH_BUF_SIZE) {
       ticsGet(thHandle, bScratchBuffer, sizeof(Byte), SCRATCH_BUF_SIZE);
       uiRemain -= SCRATCH_BUF_SIZE;
   }
   ticsGet(thHandle, bScratchBuffer, sizeof(Byte), uiRemain);
    }
#ifdef WIN32
    PrintLog(thHandle, "GetString16", "maxlen=%d, sent=%d, read=%d, \"%s\"", maxLength, usLength, usReturn, s);
#else
    PrintLog(thHandle, "GetString16", "maxlen=%d, sent=%d, read=%d", maxLength, usLength, usReturn);
#endif


    return iTruncate ? NULL : s;
}

/* =========================================================================
 * PUBLIC PUT FUNCTIONS:
 * ========================================================================= */

Int32 DLL_CALL    ticsCondition(TicsHandle thHandle, Byte Message)
/*
 * DESCRIPTION:
 * Send a condition to server tics_proxy_client. This is
 * typically used when CTRL-C is hit to abort output on client.
 *
 * ARGUMENTS:
 * TicsHandle  *pthHandle  Pointer for TICS handle
 * UShort    Message User defined message number
 */
{
    fd_set          fsWrite;
    Byte     msg=(Byte)Message;

    PrintLog(thHandle, "Condition", "message=%d", Message);

    DEBUG_PRINTF(1, (stderr, "Sending condition %d\n", Message));

    FD_ZERO(&fsWrite);
    FD_SET(thHandle->out, &fsWrite);
    if (SELECT(FD_SETSIZE, NULL, &fsWrite, NULL, NULL) < 0)
   return -1;

    /* Send as out-of-band data */

    if (send(thHandle->out, (char *)&msg, 1, MSG_OOB) != 1) {
   return GetLastSockError();
    }

    return 0;

}

Void DLL_CALL   ticsPut(TicsHandle thHandle, Ptr vpBuffer,
         UShort uiElementSize, UShort uiNElements)
{
    Int32           nBuffer,
                    mBuffer = (Int32)uiElementSize * uiNElements;

    PrintLog(thHandle, "Put", "element size=%d, number=%d", uiElementSize, uiNElements);

    nBuffer = AddToBuffer(thHandle, (Str) vpBuffer, mBuffer);
    if (nBuffer == mBuffer)
   return;

    thHandle->eofHandler(thHandle, TICS_DISCONNECTED);
    return;
}

Void DLL_CALL   ticsPutByte(TicsHandle thHandle, Byte b)
{
    Byte            bOne = b;

    PrintLog(thHandle, "PutByte", "%d", b);

    SEND_TYPE_DATA(TICS_BYTE, 1);

    ticsPut(thHandle, (Ptr) &bOne, sizeof(Byte), 1);
}

Void DLL_CALL   ticsPutShort(TicsHandle thHandle, Short s)
{
    Short           sOne = s;

    PrintLog(thHandle, "PutShort", "%d", s);

    SEND_TYPE_DATA(TICS_SHORT, 1);

    if (SWAP_ENDIAN(thHandle))
   endSwap2(&sOne, 1);
    ticsPut(thHandle, (Ptr) &sOne, sizeof(Short), 1);
}

Void DLL_CALL   ticsPutUShort(TicsHandle thHandle, UShort us)
{
    UShort          usOne = us;

    PrintLog(thHandle, "PutUShort", "%d", us);

    SEND_TYPE_DATA(TICS_USHORT, 1);

    if (SWAP_ENDIAN(thHandle))
   endSwap2(&usOne, 1);
    ticsPut(thHandle, (Ptr) &usOne, sizeof(UShort), 1);
}

Void DLL_CALL   ticsPutInt(TicsHandle thHandle, Int32 i)
{
    Int32           iOne = i;

    PrintLog(thHandle, "PutInt", "%d", i);

    SEND_TYPE_DATA(TICS_INT, 1);

    if (SWAP_ENDIAN(thHandle))
   endSwap4(&iOne, 1);
    ticsPut(thHandle, (Ptr) &iOne, sizeof(Int32), 1);
}

Void DLL_CALL   ticsPutUInt(TicsHandle thHandle, UInt32 ui)
{
    UInt32          uiOne = ui;

    PrintLog(thHandle, "PutUInt", "%d", ui);

    SEND_TYPE_DATA(TICS_UINT, 1);

    if (SWAP_ENDIAN(thHandle))
   endSwap4(&uiOne, 1);
    ticsPut(thHandle, (Ptr) &uiOne, sizeof(UInt32), 1);
}

Void DLL_CALL   ticsPutShortFloat(TicsHandle thHandle, Float f)
{
    Float           fOne = (Float) f;

    PrintLog(thHandle, "PutShortFloat", "%f", f);

    SEND_TYPE_DATA(TICS_SFLOAT, 1);

    SwapFloatToServer(thHandle, &fOne, 1);
    ticsPut(thHandle, (Ptr) &fOne, sizeof(Float), 1);
}

Void DLL_CALL   ticsPutFloat(TicsHandle thHandle, Double d)
{
    Double          dOne = d;

    PrintLog(thHandle, "PutFloat", "%f", d);

    SEND_TYPE_DATA(TICS_FLOAT, 1);

    SwapDoubleToServer(thHandle, &dOne, 1);
    ticsPut(thHandle, (Ptr) &dOne, sizeof(Double), 1);
}

#if defined(HAVE_64_BIT_INTS)
Void DLL_CALL   ticsPutLong(TicsHandle thHandle, Int64 i)
{
    Int64           iOne = i;

    PrintLog(thHandle, "PutLong", FORMAT_LONG, i);

    SEND_TYPE_DATA(TICS_LONG, 1);

    if (SWAP_ENDIAN(thHandle))
   endSwap8(&iOne, 1);
    ticsPut(thHandle, (Ptr) &iOne, sizeof(Int64), 1);
}

Void DLL_CALL   ticsPutULong(TicsHandle thHandle, UInt64 ui)
{
    UInt64          uiOne = ui;

    PrintLog(thHandle, "PutULong", FORMAT_ULONG, ui);

    SEND_TYPE_DATA(TICS_ULONG, 1);

    if (SWAP_ENDIAN(thHandle))
   endSwap8(&uiOne, 1);
    ticsPut(thHandle, (Ptr) &uiOne, sizeof(UInt64), 1);
}
#endif

Void DLL_CALL   ticsPutShorts(TicsHandle thHandle, Short *ptr, UShort count)
{
    Short          *spPtr = ptr;

    PrintLog(thHandle, "PutShorts", "number=%d", count);

    SEND_TYPE_DATA(TICS_SHORTS, count);

    SWAP_AND_PUT_INTS(thHandle, spPtr, count, Short, endSwap2)
}

Void DLL_CALL   ticsPutUShorts(TicsHandle thHandle, UShort *ptr, UShort count)
{
    UShort         *uspPtr = ptr;

    PrintLog(thHandle, "PutUShorts", "number=%d", count);

    SEND_TYPE_DATA(TICS_USHORTS, count);

    SWAP_AND_PUT_INTS(thHandle, uspPtr, count, UShort, endSwap2)
}

Void DLL_CALL   ticsPutInts(TicsHandle thHandle, Int32 *ptr, UShort count)
{
    Int32          *ipPtr = ptr;

    PrintLog(thHandle, "PutInts", "number=%d", count);

    SEND_TYPE_DATA(TICS_INTS, count);

    SWAP_AND_PUT_INTS(thHandle, ipPtr, count, Int32, endSwap4)
}

Void DLL_CALL   ticsPutUInts(TicsHandle thHandle, UInt32 *ptr, UShort count)
{
    UInt32         *uipPtr = ptr;

    PrintLog(thHandle, "PutUInts", "number=%d", count);

    SEND_TYPE_DATA(TICS_UINTS, count);

    SWAP_AND_PUT_INTS(thHandle, uipPtr, count, UInt32, endSwap4);
}

Void DLL_CALL   ticsPutShortFloats(TicsHandle thHandle, Float *ptr, UShort count)
{
    Float          *fpPtr = ptr;

    PrintLog(thHandle, "PutShortFloats", "number=%d", count);

    SEND_TYPE_DATA(TICS_SFLOATS, count);

    SWAP_AND_PUT_FLOATS(thHandle, fpPtr, count, Float, SwapFloatToServer)
}

Void DLL_CALL   ticsPutFloats(TicsHandle thHandle, Double *ptr, UShort count)
{
    Double         *dpPtr = ptr;

    PrintLog(thHandle, "PutFloats", "number=%d", count);

    SEND_TYPE_DATA(TICS_FLOATS, count);

    SWAP_AND_PUT_FLOATS(thHandle, dpPtr, count, Double, SwapDoubleToServer)
}

#if defined(HAVE_64_BIT_INTS)
Void DLL_CALL   ticsPutLongs(TicsHandle thHandle, Int64 *ptr, UShort count)
{
    Int64          *ipPtr = ptr;

    PrintLog(thHandle, "PutLongs", "number=%d", count);

    SEND_TYPE_DATA(TICS_LONGS, count);

    SWAP_AND_PUT_INTS(thHandle, ipPtr, count, Int64, endSwap8)
}

Void DLL_CALL   ticsPutULongs(TicsHandle thHandle, UInt64 *ptr, UShort count)
{
    UInt64         *uipPtr = ptr;

    PrintLog(thHandle, "PutULongs", "number=%d", count);

    SEND_TYPE_DATA(TICS_UINTS, count);

    SWAP_AND_PUT_INTS(thHandle, uipPtr, count, UInt64, endSwap8);
}
#endif

Void DLL_CALL   ticsPutString(TicsHandle thHandle, Str s)
{
    Int i;
    PrintLog(thHandle, "PutString", "\"%s\"", s);

    i = strlen((char *)s);
    if (i>USHRT_MAX)
        thHandle->eofHandler(thHandle, TICS_NO_MEMORY);
    else
        ticsPutLenString(thHandle, s, (UShort)i);
}

Void DLL_CALL   ticsPutLenString(TicsHandle thHandle, Str s, UShort usLen)
{
    UShort usSLen;

    PrintLog(thHandle, "PutLenString", "%d", usLen);

    SEND_TYPE_DATA(TICS_STRING, 1);

    usSLen = usLen;
    if (SWAP_ENDIAN(thHandle))
   endSwap2(&usSLen, 1);
    ticsPut(thHandle, &usSLen, sizeof(UShort), 1);
    ticsPut(thHandle, s, sizeof(Char), usLen);
}

Void DLL_CALL   ticsPutString16(TicsHandle thHandle, Str16 s)
{
    Int i;
    Str16 p;
#ifdef WIN32
    PrintLog(thHandle, "PutString16", "\"%S\"", s);
#else
    PrintLog(thHandle, "PutString16", "");
#endif

    /* find the length */
    p = s;
    while (*p) p++;
    i = p - s;
    if (i>USHRT_MAX)
        thHandle->eofHandler(thHandle, TICS_NO_MEMORY);
    else
        ticsPutLenString16(thHandle, s, (UShort)i);
}

Void DLL_CALL   ticsPutLenString16(TicsHandle thHandle, Str16 s, UShort usLen)
{
    UShort usSLen;

    PrintLog(thHandle, "PutLenString16", "%d", usLen);

    SEND_TYPE_DATA(TICS_STRING16, 1);

    usSLen = usLen;
    if (SWAP_ENDIAN(thHandle))
   endSwap2(&usSLen, 1);
    ticsPut(thHandle, &usSLen, sizeof(UShort), 1);

    SWAP_AND_PUT_INTS(thHandle, s, usLen, Char16, endSwap2);
}


/* =========================================================================
 * PUBLIC ADMIN FUNCTIONS:
 * ========================================================================= */

TicsEofHandler DLL_CALL ticsSetEofHandler(TicsHandle thHandle, TicsEofHandler func)
{
    TicsEofHandler  OldHandler = (TicsEofHandler) thHandle->eofHandler;

    PrintLog(thHandle, "SetEofHandler", "func=%p", func);

    if (func == NULL)
   thHandle->eofHandler = eofDefaultHandler;
    else
   thHandle->eofHandler = func;

    return OldHandler;
}

Int32 DLL_CALL    ticsSetAutoFlush(TicsHandle thHandle, Int32 iNewAutoFlush)
{
    Int32           iOldAutoFlush = thHandle->autoFlush;

    PrintLog(thHandle, "SetAutoFlush", "%d", iNewAutoFlush);

    thHandle->autoFlush = iNewAutoFlush;

    return iOldAutoFlush;
}

Void DLL_CALL   ticsFlush(TicsHandle thHandle)
{
    PrintLog(thHandle, "Flush", "");

    DEBUG_PRINTF(1, (stderr, "Flush %d bytes\n", thHandle->curBufferSize));
    if (thHandle->curBufferSize > 0) {
   WriteToSocket(thHandle->out,
            (Str) thHandle->buffer,
            thHandle->curBufferSize,
            EXACTLY);
   thHandle->curBufferSize = 0;
    }
}

Int32 DLL_CALL    ticsLogToFile(TicsHandle thHandle, Str FileName)
{
    time_t          t;
    Str             s;
    Char            szName[MAXHOSTNAMELEN + 1];


    DEBUG_PRINTF(1, (stderr, "Flush %d bytes\n", thHandle->curBufferSize));
    if (thHandle->Logfile != NULL) {
   PrintLog(thHandle, "LogToFile", "<CLOSED>");
   t = time(NULL);
   s = (Str) ctime(&t);
   fprintf(thHandle->Logfile, "time        = %s\n", s);
   fclose(thHandle->Logfile);
   thHandle->Logfile = NULL;
    }
    if ((FileName != NULL) && (*FileName != '\0')) {
   thHandle->Logfile = fopen((char *) FileName, "w");
   /* Make sure unbuffered */
   setvbuf(thHandle->Logfile, NULL, _IONBF, 0);
   fprintf(thHandle->Logfile, "TICS client log file\n");
   fprintf(thHandle->Logfile, "tics.c          = $Revision: 220.7 $\n");
   if (!gethostname((char *)szName, sizeof(szName) - 1))
       fprintf(thHandle->Logfile, "client host     = %s\n", szName);
   fprintf(thHandle->Logfile, "max buffer size = %d\n", thHandle->maxBufferSize);
   fprintf(thHandle->Logfile, "protocol        = %d\n", thHandle->bProtocol);
   fprintf(thHandle->Logfile, "auto flush      = %d\n", thHandle->autoFlush);
   fprintf(thHandle->Logfile, "endianness      = %d\n", thHandle->endianness);
   fprintf(thHandle->Logfile, "float ieee      = %d\n", thHandle->floatieee);
   t = time(NULL);
   s = (Str) ctime(&t);
   fprintf(thHandle->Logfile, "time            = %s\n", s);
    }
    return (Int32) (thHandle->Logfile != 0);
}

Int32 DLL_CALL    ticsError(TicsHandle thHandle)
{
    return thHandle->Status;
}

void DLL_CALL    ticsVersion(UInt *major, UInt *minor)
{
  *major = TICS_VER_MAJOR;
  *minor = TICS_VER_MINOR;
}

static void     eofDefaultHandler(TicsHandle thHandle, Int32 Status)
{
    PrintLog(thHandle, "DefaultHandler", "%d", Status);

    thHandle->Status = Status;
#ifndef TICSDLL
    fprintf(stderr, "Error %d. Lost connection to server\n", Status);
    exit(TICS_EXIT_FAILURE);
#endif

}

/* =========================================================================
 * PRIVATE FUNCTIONS:
 * ========================================================================= */

static void     CheckType(TicsHandle thHandle, UShort DataType, UShort Elements)
{
    UInt32 DataTypeSent;
    UInt32 ElementsSent;
    Int32  nRead;

    if (thHandle->autoFlush)
   ticsFlush(thHandle);

    /* get datatype sent */

    nRead=ReadFromSocket(thHandle->in, (Str)&DataTypeSent,
          sizeof(DataTypeSent), EXACTLY);
    if (nRead != sizeof(DataTypeSent)) {
#if !defined(TICSDLL)
   fprintf(stderr, "CheckType, read %d, expected %d\n",
      nRead, sizeof(DataTypeSent));
#endif
   thHandle->eofHandler(thHandle, TICS_DISCONNECTED);
    }
    if (SWAP_ENDIAN(thHandle))
   endSwap4(&DataTypeSent, 1);

    /* get number of elements sent if sending multiplies */

    if (DataType & TICS_MULTIPLE) {
   nRead=ReadFromSocket(thHandle->in, (Str)&ElementsSent,
              sizeof(ElementsSent), EXACTLY);
   if (nRead != sizeof(ElementsSent)) {
#if !defined(TICSDLL)
       fprintf(stderr, "CheckType, read %d, expected %d\n",
          nRead, sizeof(ElementsSent));
#endif
       thHandle->eofHandler(thHandle, TICS_DISCONNECTED);
   }
   if (SWAP_ENDIAN(thHandle))
       endSwap4(&ElementsSent, 1);
    } else {
   ElementsSent = 1;
    }

    /* check datatype and num elements sent against data expected */

    if ((UInt32)DataType!=DataTypeSent || (UInt32)Elements!=ElementsSent) {
#if !defined(TICSDLL)
   fprintf(stderr, "Server & client out of step:\n");
   fprintf(stderr, "  data sent does not match data expected,\n");
   fprintf(stderr, "  datatype expected=%d, datatype sent=%d\n",
      DataType, DataTypeSent);
   fprintf(stderr, "  elements expected=%d, elements sent=%d\n",
      Elements, ElementsSent);
#endif
   thHandle->eofHandler(thHandle, TICS_DISCONNECTED);
    }

}

static Int32    AddToBuffer(TicsHandle thHandle, Str Buffer, Int32 mBuffer)
{
    Int32           iRemain = mBuffer;
    Str             cFrom = Buffer;

    DEBUG_PRINTF(1, (stderr, "Add %d bytes to buffer\n", mBuffer));
    while ((thHandle->curBufferSize + iRemain) >= (Int32)thHandle->maxBufferSize) {
   UShort         iLen = thHandle->maxBufferSize - thHandle->curBufferSize;

   memcpy(&thHandle->buffer[thHandle->curBufferSize], cFrom, iLen);
   thHandle->curBufferSize += iLen;
   ticsFlush(thHandle);
   iRemain -= iLen;
   cFrom += iLen;
    }
    if (iRemain > 0) {
   memcpy(&thHandle->buffer[thHandle->curBufferSize], cFrom, (Int)iRemain);
   thHandle->curBufferSize += (UShort)iRemain;
    }
    return mBuffer;
}

static Int32    ReadFromSocket(SOCKET soId, Str bBuffer, Int32 mBuffer, Int iHow)
/*
 * Perform a read from the specified socket. If iHow is ONCE, try and read in
 * one go the exit. If iHow is EXACT, keep reading until the amount requested
 * is supplied For the moment, this is a blocking function, I may change that
 * in future.
 */
{
    fd_set          fsRead;
    Int32           nBuffer = 0;
    Int32           i;
    Int32       left;

    DEBUG_PRINTF(1, (stderr, "Read %d bytes from socket\n", mBuffer));
    while (nBuffer < mBuffer) {
   FD_ZERO(&fsRead);
   FD_SET(soId, &fsRead);
   i = SELECT(FD_SETSIZE, &fsRead, NULL, NULL, NULL);

   if (i >= 0) {
       left = mBuffer - nBuffer;
       if (left>INT_MAX) left = INT_MAX;
       i = recv(soId, (char*)&bBuffer[nBuffer], (Int)left, 0);
   }

   if (i <= 0) {
       i = GetLastSockError();
#ifdef EINTR
       if (i == EINTR)
      continue;
#endif
       return -1;
   }
   nBuffer += i;
   if (iHow == ONCE)
       break;
    }
    DEBUG_PRINTF(1, (stderr, "Done\n"));
    return nBuffer;
}

static Int32    WriteToSocket(SOCKET soId, Str bBuffer, Int32 mBuffer, Int iHow)
/*
 * Performing a write to the specified socket If iHow is ONCE, try and write
 * in one go, then exit If iHow is EXACT, keep trying until the specified
 * amount has been written For the moment this is a blocking function, I may
 * change that in future
 */
{
    fd_set          fsWrite;
    Int32           nBuffer = 0;
    Int32           i;
    Int32       left;

    DEBUG_PRINTF(1, (stderr, "Write %d bytes to socket\n", mBuffer));
    while (nBuffer < mBuffer) {
   FD_ZERO(&fsWrite);
   FD_SET(soId, &fsWrite);
   i = SELECT(FD_SETSIZE, NULL, &fsWrite, NULL, NULL);

   if (i >= 0) {
       left = mBuffer - nBuffer;
       if (left>INT_MAX) left = INT_MAX;
       i = send(soId, (char*)&bBuffer[nBuffer], (Int)left, 0);
   }

   if (i <= 0) {
       i = GetLastSockError();
#ifdef EINTR
       if (i == EINTR)
      continue;
#endif
       return -1;
   }
   nBuffer += i;
   if (iHow == ONCE)
       break;
    }
    return nBuffer;
}

/* One function to swap Float TO the server's format */

static void     SwapFloatToServer(TicsHandle thHandle, Float *fpData, UShort i)
{
    /*
     * Should handle case of Server being VAX, client being IEEE here,
     * however the TICS server is not supported on VAX.
     */

    if (SWAP_ENDIAN(thHandle))          /* Assume host and client IEEE */
   endSwap4(fpData, i);

}

/* One function to swap Double FROM the server's format */

static void     SwapFloatFromServer(TicsHandle thHandle, Float *fpData, UShort i)
{
    /*
     * Should handle case of Server being VAX, client being IEEE here,
     * however the TICS server is not supported on VAX.
     */

    if (SWAP_ENDIAN(thHandle))
   endSwap4(fpData, i);

}

/* One function to swap Double TO the server's format */

static void     SwapDoubleToServer(TicsHandle thHandle, Double *dpData, UShort i)
{
    /*
     * Should handle case of Server being VAX, client being IEEE here,
     * however the TICS server is not supported on VAX.
     */

    if (SWAP_ENDIAN(thHandle))
   endSwap8(dpData, i);

}

/* One function to swap Double FROM the server's format */

static void     SwapDoubleFromServer(TicsHandle thHandle, Double *dpData, UShort i)
{
    /*
     * Should handle case of Server being VAX, client being IEEE here,
     * however the TICS server is not supported on VAX.
     */

    if (SWAP_ENDIAN(thHandle))
   endSwap8(dpData, i);

}

static void     PrintLog(TicsHandle thHandle, char *fnname, char *format,...)
{
    va_list         ap;

    if (thHandle->Logfile != NULL) {
   fprintf(thHandle->Logfile, "%-20s:", fnname);
   va_start(ap, format);
   vfprintf(thHandle->Logfile, format, ap);
   va_end(ap);
   fprintf(thHandle->Logfile, "\n");
    }
}

/**********************************************************************
/**** END tics.c
/**********************************************************************/

/**********************************************************************
/######################################################################
/#### acpt32 implementation
/######################################################################
/*********************************************************************/

/**********************************************************************
/**** BEGIN acpt.c
/**********************************************************************/

/*
 * acpt.c: C code for use in ACPs which wish to transfer
 * objects between C and Magik.
 */

/*
 * Note: when changing any of these numbers, ensure RELEASES.h and acpt_tics.rc
 * are updated, the latter in two places.
 */
#define ACPT_V_MAJOR 1
#define ACPT_V_MINOR 4
#define ACPT_V_BUGFIX 0

/*
 * Header and protocol - used to determine we are in step with other end, and
 * determine endian-ness. The header value is the non-palindromic char set 'ACPT' -
 * If we read 'TPCA' then we're byte-swapped.
 */
#define ACPT_PROTOCOL 1
#define ACPT_MIN_READ_PROTOCOL 1

#define ACPT_HEADER 0x41435054


#define ACPT_PTR_SIZE (Int32)(sizeof(void *))  /* code demands this is always the same */
#define ACPT_STREAM_BUFFER_INCR 128     /* size stream buffer is extended by when necc.*/


#define ACPT_STREAM_TYPE_SIZE OFFSET(AcptTypeDesc, id)


static AcptResult acptInitTypes();
static AcptResult acptFreeBuiltinType(AcptType type);

static AcptResult acptMakeBuiltinType(AcptType id, Int32 size, Int32 stream_size,
                  AcptTypeMT *MT);
static AcptResult acptMakeBuiltinVectorType(AcptType id, Int32 size, AcptType el_type,
                   UInt32 vec_type,Int32 length);
static AcptResult acptMakeVectorTypeHelper(AcptTypeDesc **type_desc,
                  UInt32 vec_type, Int32 vec_offset,
                  Int32 length_detail, Int32 size,
                  Bool internal, UInt32 owner_flags);
static AcptResult acptMakeRefTypeHelper(AcptTypeDesc **type_desc, Int32 size);

static AcptResult acptMakeStructTypeHelper(AcptStructElement field_specs[],
                  Int32 n_fields,
                  Int32 size, AcptType *typep);

static AcptResult acptMakeUnionTypeHelper(AcptUnionElement field_specs[],
                 Int32 n_fields,
                Int32 data_offset, Int32 discrim_offset,
                Int32 size, AcptType *typep);

static AcptResult acptDerefType(AcptType type);

static AcptResult acptGetStreamTypeDesc(AcptStream *stream, AcptType type,
               AcptTypeDesc **type_desc);
static AcptResult acptGetObjectDetails(AcptStream *stream, AcptType *typep,
                   Int32 *lengthp, Bool *is_vectorp);
static AcptResult acptSetObjectDetails(AcptStream *stream, AcptType type, Int32 length,
                   Bool is_vector);
static AcptResult acptPrepareToSendObject(AcptStream *stream, AcptTypeDesc *type_desc);

static AcptResult acptSendType(AcptStream *stream, AcptTypeDesc *type_desc);
static AcptResult acptReceiveType(AcptStream *stream);
static AcptResult acptSendIntro(AcptStream *stream);
static AcptResult acptReceiveIntro(AcptStream *stream);
static AcptResult acptCompareTypes(AcptStream *stream, AcptTypeDesc *c_type,
               AcptTypeDesc *s_type);

static AcptResult acptReadStruct(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
             Int32 n_els);
static AcptResult acptWriteStruct(AcptTypeDesc *type, Buffer from_addr,
              AcptStream *stream, Int32 n_els);
static AcptResult acptFreeStruct(AcptTypeDesc *type, Buffer from_addr, Int32 n_els);

static AcptResult acptReadRef(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
               Int32 n_els);
static AcptResult acptWriteRef(AcptTypeDesc *type, Buffer from_addr,
                AcptStream *stream, Int32 n_els);
static AcptResult acptFreeRef(AcptTypeDesc *type, Buffer from_addr, Int32 n_els);

static AcptResult acptReadVec(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
               Int32 n_els);
static AcptResult acptWriteVec(AcptTypeDesc *type, Buffer from_addr, AcptStream *stream,
                Int32 n_els);
static AcptResult acptFreeVec(AcptTypeDesc *type, Buffer from_addr, Int32 n_els);

static AcptResult acptReadUnion(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
            Int32 n_els);
static AcptResult acptWriteUnion(AcptTypeDesc *type, Buffer from_addr,
             AcptStream *stream, Int32 n_els);
static AcptResult acptFreeUnion(AcptTypeDesc *type, Buffer from_addr, Int32 n_els);

static AcptResult acptReadKleene(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
             Int32 n);
static AcptResult acptWriteKleene(AcptTypeDesc *type, Buffer from_addr,
              AcptStream *stream, Int32 n);

static AcptResult acptReadDI(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
              Int32 n);
static AcptResult acptWriteDI(AcptTypeDesc *type, Buffer from_addr, AcptStream *stream,
               Int32 n);

static AcptResult acptByteSwap4InSitu(UInt32 *values, Int32 n);

static AcptResult acptPrepareForError(AcptStream *stream);
static AcptResult acptProcessError(AcptStream *stream, Int32 i,
               AcptTypeDesc *type, AcptResult result);
static AcptResult acptInitFlavourNames();
static AcptResult acptInitScalarNames();


static AcptTypeMT acptMT0 = {acptReadScalars, acptWriteScalars};
static AcptTypeMT acptMT1 = {acptReadScalars, acptWriteScalars, NULL, acptStringLength1};
static AcptTypeMT acptMT2 = {acptReadScalars, acptWriteScalars, NULL, acptStringLength2};
static AcptTypeMT acptMT4 = {acptReadScalars, acptWriteScalars};
static AcptTypeMT acptMT8 = {acptReadScalars, acptWriteScalars};
static AcptTypeMT acptMT1_2 = {acptRead1_2, acptWrite1_2, NULL, acptStringLength2};

static AcptTypeMT acptMTKleene = {acptReadKleene, acptWriteKleene};
static AcptTypeMT acptMTDI = {acptReadDI, acptWriteDI};
static AcptTypeMT acptMTBigInt = {acptReadBigInt, acptWriteBigInt, acptFreeBigInt};

static AcptTypeMT acptMTStruct = {acptReadStruct, acptWriteStruct, acptFreeStruct};
static AcptTypeMT acptMTUnion = {acptReadUnion, acptWriteUnion, acptFreeUnion};

static AcptTypeMT acptMTRef = {acptReadRef, acptWriteRef, acptFreeRef};
static AcptTypeMT acptMTVec = {acptReadVec, acptWriteVec, acptFreeVec};


/*
 * Defined types are held in two separate lists, builtinTypeDescList and typeDescs.
 * The former list holds types whose type_id is less than ACPT_MAX_BUILTIN_TYPES,
 * and will always represent the same thing (covering the basic scalar types).
 * The latter list is allocated and extended dynamically, and holds user defined
 * types. The type_id in this case is the index into typeDescs+ACPT_MAX_BUILTIN_TYPES.
 */

static AcptTypeDesc **typeDescs = NULL;
static Int32 typeDescListMax = 0;

#if defined(_ACPTM_C_ONLY_)
AcptTypeDesc *builtinTypeDescList[ACPT_MAX_BUILTIN_TYPES];
UInt32 matchSeq=0;      /* the source of type_desc.matching_read_seq,
             * incremented for each new C type made */
#endif

static AcptStream **writeStreams = NULL; /* list of current Write streams */
static Int32 writeStreamsMax = 0;

/*
 * ACPT_TYPE_DEFINITION is an unacheivable as a normal type, and is used to
 * herald a type definition coming down the line, ahead of some real data.
 */
#define ACPT_TYPE_DEFINITION 0xffff

/*
 * Support for multithreading. We use one mutex, which controls NewType(),
 * DerefType() and also FreeStream() and the bit which marks a type as sent
 * to a stream(). We defend against recursion with a counter.
 */
#if defined(ACPT_MULTI_THREADED)
static CRITICAL_SECTION cs;
#define INIT_MUTEX()      InitializeCriticalSection(&cs)
#define START_EXCLUSIVE   EnterCriticalSection(&cs); {
#define END_EXCLUSIVE     } LeaveCriticalSection(&cs)
#else
#define INIT_MUTEX()
#define START_EXCLUSIVE
#define END_EXCLUSIVE
#endif

/*
 * If, during a write, the definition of a type is been constructed
 * and the bit set to say it has now been sent to that stream, and
 * then the write fails (eg because a short vector is too long, for
 * example), it is necessary to unmark the type send, because the
 * other end has not, in fact, received it.
 */

/*
 * We limit the number of types to 65k, because when transmitting type info
 * the rest of the word contains flags and size info. ACPT_TYPES_INCR is the
 * size a list of types is extended by each time it needs to grow. Note that
 * type 0xff is also omitted, as this appears as the ACPT_TYPE_DEFINITION
 * marker.
 */
#define ACPT_MAX_TYPES (0xfffe - ACPT_MAX_BUILTIN_TYPES)
#define ACPT_TYPES_INCR 32

/*
 * Allocators for data space. This use effectively the same macros as
 * Alloc and Free, but are 'pluggable' so the punter can provide his
 * own one. In particular, in a single thread application which doesn't
 * want to preserve any data from one retrieval after the next, the same
 * bucket could be used each time using a custom allocator.
 *
 * Note: Argument prototypes have been deliberately omitted because the
 * precise spec of the calloc() and free() interfaces varies.

static struct {
    void * (*calloc)();
    void (*free)();
} AcptDataAllocators = {calloc, free};
 */

// #define acptAllocDataSpace(_n, _type, _loc) \
//   ((*(_loc) = (_type *)((*AcptDataAllocators.calloc)((_n) ? (_n) : 1, sizeof(_type)))) ? ACPT_OK : acptExcp(ACPT_EXCP_MEMORY_EXHAUSTED))
#define acptAllocDataSpace(_n, _type, _loc) \
   ((*(_loc) = (_type *)(calloc((_n) ? (_n) : 1, sizeof(_type)))) ? ACPT_OK : acptExcp(ACPT_EXCP_MEMORY_EXHAUSTED))

// #define acptFreeDataSpace(_loc) if (_loc) (*AcptDataAllocators.free)(_loc); else;
#define acptFreeDataSpace(_loc) if (_loc) free(_loc); else;

static void (ACPT_DLL_CALL *acptTypeMismatchHandler)(AcptStreamError *error) = NULL;


xstatic AcptResult acptExcp(AcptResult excp)
    /*
     * Does nothing but return EXCP. Provided as it is a useful hook for
     * debugging, and could possibly have tracing introduced if this
     * was helpful. Alternatively, this could be conditionalised, and,
     * for example, replaced by a macro, in an optimised version.
     */
{return excp;}

    /*
AcptResult ACPT_DLL_CALL acptSetDataAllocators(void * (*calloc)(), void (*free)())
     * Set the allocators to be used for generating and freeing space
     * used in object data. These allocators are *not* used for internal
     * space (eg for types or streams). The precise spec of CALLOC and
     * FREE is the defined spec for the ANSII standard versions, and
     * should be rigorously adhered to. In particular, CALLOC should
     * return NULL on memory_exhausted.
{
    AcptDataAllocators.calloc = calloc;
    AcptDataAllocators.free = free;

    return ACPT_OK;
}
     */

AcptResult ACPT_DLL_CALL
acptSetTypeMismatchHandler(void (ACPT_DLL_CALL *handler)(AcptStreamError *error))
    /*
     * Declare a handler which will be called when a type-mismatch
     * error occurs. The handler takes one argument, which will be
     * supplied as a pointer to an AcptStreamError struct, from which
     * various information about the error can be established. This
     * handler could then, for example, be used to print details of
     * the mismatch on standard-error. An acpt function
     * acptPrintStreamMismatchError() is provided for this purpose,
     * but this is *not* the default. The default is a NULL handler
     * (which therefore does nothing).
     */
{
    acptTypeMismatchHandler = handler;

    return ACPT_OK;
}

/*
 *       I N I T I A L I S A T I O N
 *       ___________________________
 */


AcptResult ACPT_DLL_CALL acptInit()
    /*
     * Initialise the acpt package. This call is mandatory before any other
     * acpt interfaces are called. This interface should only be called once
     * per session.
     */
{
    if (ACPT_INT_SIZE != sizeof(Int32)) return acptExcp(ACPT_EXCP_INTS_TOO_SMALL);
    INIT_MUTEX();
    acptInitFlavourNames();
    acptInitScalarNames();
    return acptInitTypes();
}

AcptResult ACPT_DLL_CALL acptEnd()
    /*
     * Clean up space allocated by acpt which cannot be freed any
     * other way.  This interface is only supplied to enable
     * programmers to ensure that there are no space-leaks, and it
     * will return ACPT_EXCP_UNFREED_TYPE if any user-defined types
     * are left outstanding. Note that without this interface, Purify
     * will not find types left lying around, as they are referenced
     * via static, which it ignores.
     */
{
    Int32 i;

    for (i=0; i<typeDescListMax && !typeDescs[i]; i++);

    if (i < typeDescListMax) return acptExcp(ACPT_EXCP_UNFREED_TYPE);

    for (i=0; i<ACPT_MAX_BUILTIN_TYPES; i++) acptFreeBuiltinType(i);

    acptFree(typeDescs);
    typeDescs = NULL;
    typeDescListMax = 0;

    return ACPT_OK;
}

static AcptResult acptFreeBuiltinType(AcptType type)
    /*
     * Free the space for a builtin type. This is only supplied for
     * use when running purify, and it is convenient to be able to
     * clean everything up at the end. TYPE can be any typeno (so
     * it is legitimate to call this even if the type is not defined)
     */
{
    AcptResult result = ACPT_OK;
    Int32 i;

    AcptTypeDesc *type_desc, *owner_desc, **pp, **e;
    AcptStream *stream;

    if (type >= ACPT_MAX_BUILTIN_TYPES) return ACPT_OK;

    if (!(type_desc = builtinTypeDescList[type])) return ACPT_OK;

    acptFree(type_desc->flavours);
    acptFree(type_desc->fields);
    acptFree(type_desc);
    builtinTypeDescList[type] = NULL;

    return ACPT_OK;
}

AcptResult ACPT_DLL_CALL acptVersion(UInt32 *major, UInt32 *minor, UInt32 *bugfix)
    /*
     * Populates the three arguments with Major, Minor and Bugfix version numbers.
     */
{
    *major = ACPT_V_MAJOR;
    *minor = ACPT_V_MINOR;
    *bugfix = ACPT_V_BUGFIX;

    return ACPT_OK;
}

static AcptResult acptInitTypes()
    /*
     * Initialise the builtin types.
     */
{
    AcptResult result;
    static AcptType type_holder;
    static AcptStructElement date_time_fields[] = {
   {{ACPT_TYPE_INT32}, OFFSIZ(AcptDateTime, day)},
   {{ACPT_TYPE_INT32}, OFFSIZ(AcptDateTime, msecond)}};

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_UNSET, 0, 0, &acptMT0), result);

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_UINT8,  1, 1, &acptMT1), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_UINT16, 2, 2, &acptMT2), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_UINT32, 4, 4, &acptMT4), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_UINT64, 8, 8, &acptMT8), result);

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_INT8,  1, 1, &acptMT1), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_INT16, 2, 2, &acptMT2), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_INT32, 4, 4, &acptMT4), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_INT64, 8, 8, &acptMT8), result);

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_FLOAT32, 4, 4, &acptMT4), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_FLOAT64, 8, 8, &acptMT8), result);

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_CHAR8,  1, 1, &acptMT1), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_CHAR16, 2, 2, &acptMT2), result);

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_BOOL,   4, 1, &acptMTKleene), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_KLEENE, 4, 1, &acptMTKleene), result);

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_SIMPLE_TIME, 4, 4, &acptMT4), result);
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_DATE,        4, 4, &acptMT4), result);

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_DOUBLE_AS_INT, 8, 8, &acptMTDI), result);

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_BIGINT, ACPT_PTR_SIZE, 0,
                &acptMTBigInt), result);

    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_SHORT_STRING8, ACPT_PTR_SIZE,
                 ACPT_TYPE_CHAR8, ACPT_SHORT_V_STRING, 0),
      result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_SHORT_STRING16, ACPT_PTR_SIZE,
                 ACPT_TYPE_CHAR16, ACPT_SHORT_V_STRING, 0),
      result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_LONG_STRING8, ACPT_PTR_SIZE,
                 ACPT_TYPE_CHAR8, ACPT_LONG_V_STRING, 0),
      result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_LONG_STRING16, ACPT_PTR_SIZE,
                 ACPT_TYPE_CHAR16, ACPT_LONG_V_STRING, 0),
      result);

    type_holder = ACPT_TYPE_DATE_TIME;
    ACPT_CHECKR(acptMakeStructTypeHelper(date_time_fields,
                ARRAY_LEN(date_time_fields),
                sizeof(AcptDateTime), &type_holder),
      result);

    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_EMPTY_VECTOR, 0, 0, &acptMT0), result);

    /*
     * The Char8to16 types permit a Char8 on the stream to manifest itself locally
     * as a Char16. It therefore matches an incoming Char8 (and so has that as
     * its base_id) and simply stores the data locally as Char16.
     */
    ACPT_CHECKR(acptMakeBuiltinType(ACPT_TYPE_CHAR8_TO_16,  2, 1, &acptMT1_2),
      result);

    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_SHORT_STRING8_TO_16, ACPT_PTR_SIZE,
                 ACPT_TYPE_CHAR8_TO_16, ACPT_SHORT_V_STRING,
                 0), result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_LONG_STRING8_TO_16, ACPT_PTR_SIZE,
                 ACPT_TYPE_CHAR8_TO_16, ACPT_LONG_V_STRING, 0),
      result);

    builtinTypeDescList[ACPT_TYPE_CHAR8_TO_16]->base_id = ACPT_TYPE_CHAR8;
    builtinTypeDescList[ACPT_TYPE_SHORT_STRING8_TO_16]->base_id = ACPT_TYPE_SHORT_STRING8;
    builtinTypeDescList[ACPT_TYPE_LONG_STRING8_TO_16]->base_id = ACPT_TYPE_LONG_STRING8;

    /*
     * Sector geometry types. These are built-in because they are very basic,
     * and the Magik end needs to use them.
     */
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_COORD, 8*2, ACPT_TYPE_FLOAT64,
                 ACPT_FIXED_VECTOR, 2), result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_COORD3D, 8*3, ACPT_TYPE_FLOAT64,
                 ACPT_FIXED_VECTOR, 3), result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_COORD2H, 8*3, ACPT_TYPE_FLOAT64,
                 ACPT_FIXED_VECTOR, 3), result);


    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_SIMPLE_SECTOR,
                 sizeof(AcptSimpleSector), ACPT_TYPE_COORD,
                 ACPT_LONG_V_VECTOR,
                 OFFSET(AcptSimpleSector, NCoords)), result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_SECTOR_Z,
                 sizeof(AcptSectorZ), ACPT_TYPE_COORD3D,
                 ACPT_LONG_V_VECTOR,
                 OFFSET(AcptSectorZ, NCoords)), result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_SECTOR_H,
                 sizeof(AcptSectorH), ACPT_TYPE_COORD2H,
                 ACPT_LONG_V_VECTOR,
                 OFFSET(AcptSectorH, NCoords)), result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_CIRCLE,
                 sizeof(AcptSector), ACPT_TYPE_COORD,
                 ACPT_LONG_V_VECTOR,
                 OFFSET(AcptSimpleSector, NCoords)), result);
    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_TAN_ARC,
                 sizeof(AcptSectorH), ACPT_TYPE_COORD2H,
                 ACPT_LONG_V_VECTOR,
                 OFFSET(AcptSectorH, NCoords)), result);
        ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_CP_ARC,
                 sizeof(AcptSectorH), ACPT_TYPE_COORD2H,
                 ACPT_LONG_V_VECTOR,
                 OFFSET(AcptSectorH, NCoords)), result);

    type_holder = ACPT_TYPE_SECTOR; /* very tacky, but saves more code */
    ACPT_CHECKR(acptMakeUnionTypeHelper(SectorDef, ARRAY_LEN(SectorDef),
               OFFSET(AcptSector, u), OFFSET(AcptSector, type),
               sizeof(AcptSector), &type_holder), result);

    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_SECTOR_ROPE,
                 sizeof(AcptSectorRope), ACPT_TYPE_SECTOR,
                 ACPT_LONG_V_VECTOR,
                 OFFSET(AcptSectorRope, NSectors)), result);

    ACPT_CHECKR(acptMakeBuiltinVectorType(ACPT_TYPE_GIS_ID,
                 sizeof(AcptGIS_ID), ACPT_TYPE_UINT32,
                 ACPT_FIXED_VECTOR, 3), result);


    return ACPT_OK;
}


static AcptResult acptMakeBuiltinType(AcptType id, Int32 size, Int32 stream_size,
                  AcptTypeMT *MT)
  /*
   * Make a virgin type with the given ID, SIZE in bytes and method-table MT.
   */
{
    AcptResult result;
    AcptTypeDesc *type_desc;

    ACPT_CHECKR(acptAllocType(1, AcptTypeDesc, &type_desc), result);

    type_desc->id = id;
    type_desc->base_id = id;
    type_desc->size = size;
    type_desc->stream_size = stream_size;
    type_desc->MT = MT;
    type_desc->compound_length = 1;
    type_desc->matching_read_seq = ++matchSeq;

    builtinTypeDescList[id] = type_desc;

    return ACPT_OK;
}

static AcptResult acptMakeBuiltinVectorType(AcptType id, Int32 size, AcptType el_type,
                   UInt32 vec_type,Int32 length_detail)
  /*
   * Make a virgin builtin vector type with the given ID, SIZE in bytes, element
   * type EL_TYPE, specific type VEC_TYPE and length, if FIXED_LENGTH. Arguments
   * are *not* validated for consistency, and EL_TYPE must already have been declared.
   */
{
    AcptResult result;
    AcptTypeDesc *el_type_desc, *type_desc;

    ACPT_CHECKR(acptGetTypeDesc(el_type, &el_type_desc), result);
    ACPT_CHECKR(acptAllocType(1, AcptTypeDesc, &type_desc), result);

    type_desc->id = id;
    type_desc->base_id = id;
    type_desc->size = size;
    type_desc->el_subtype = el_type_desc;
    type_desc->type_detail = vec_type;
    type_desc->data_offset = 0;  /* always */
    type_desc->data_detail = length_detail;
    type_desc->matching_read_seq = ++matchSeq;

    if (vec_type == ACPT_FIXED_VECTOR) {
   type_desc->MT = el_type_desc->MT;
   type_desc->stream_size = el_type_desc->stream_size;
   type_desc->compound_length = el_type_desc->compound_length*length_detail;
    } else {
   type_desc->MT = &acptMTVec;
   type_desc->compound_length = 0;
    }

    builtinTypeDescList[id] = type_desc;

    return ACPT_OK;
}


/*
 *     N E W   T Y P E   C R E A T I O N
 *    ____________________________________
 */
AcptResult ACPT_DLL_CALL
acptNewStructType(AcptStructElement field_specs[], Int32 n_fields,
        Int32 size, AcptType *typep)
    /*
     * Populates TYPEP with a reference to a new type which defines a
     * C struct which can be transferred. FIELD_SPECS (N_FIELDS long)
     * describes each field in the struct, and the order in which they
     * will be transferred. SIZE is the size of the C struct, in bytes.
     */
{
    *typep = 0;
    return acptMakeStructTypeHelper(field_specs, n_fields, size, typep);
}

static AcptResult acptMakeStructTypeHelper(AcptStructElement field_specs[],
                  Int32 n_fields,
                  Int32 size, AcptType *typep)
  /*
   * Does the actual work for NewStructType.
   */
{
    Int32 i;
    AcptResult result;
    UInt32 flags, vec_type;
    AcptStructElement *spec;
    AcptTypeField *fields = NULL;
    AcptTypeDesc *subtype_desc, *new_type;
    AcptTypeDesc temp_type;
    AcptType subtype;

    /*
     * Make a new field vector: the desc itself is made in temporary space.
     */
    ACPT_CHECKR(acptAllocType(n_fields, AcptTypeField, &fields), result);

    for (i=0; i<n_fields; i++) {
   spec = &field_specs[i];
   subtype = spec->type.type;
   vec_type = spec->type.vec_type;
   flags = spec->flags;

   ACPT_CHECK((flags & ~(ACPT_UNSET | ACPT_REF))==0, ACPT_EXCP_BAD_FLAGS, result);
   /*
    * We do not permit an in-line REF to a VVec field, because the in-line
    * definition says where the length goes, and this must be part of the
    * same thing as the ref to the data.
    */
   ACPT_CHECK(!((flags & ACPT_REF) && acptIsVVector(vec_type)), ACPT_EXCP_BAD_FLAGS,
         result);

   ACPT_CHECKF(acptGetTypeDesc(subtype, &subtype_desc), result);

   /*
    * Make the new subtype if more than just type specified. SUBTYPE_DESC
    * is re-populated with the new flavour by each helper.
    */
   if (acptIsVector(vec_type))
     ACPT_CHECKF(acptMakeVectorTypeHelper(&subtype_desc, spec->type.vec_type,
                      spec->offset, spec->type.length_detail,
                      spec->size, TRUE, flags), result);
   if (flags & ACPT_REF)
     ACPT_CHECKF(acptMakeRefTypeHelper(&subtype_desc, spec->size), result);

   /*
    * Set up the reference to the underlying type. We lose the REF flag
    * (if present) because it would have been coalesced into the new subtype.
    */
   fields[i].flags = flags & ~ACPT_REF;
   fields[i].offset = spec->offset;
   fields[i].extra_info = spec->detail_offset;

   /*
    * Set the space taken up by the field in the body of the struct (so
    * it can be cleared if necc - this is always the size the man gave.
    */
   fields[i].inline_size = spec->size;

   fields[i].subtype = subtype_desc;
    }

    /*
     * Get the actual type id for the new type, and set the copiers.
     */
    memset(&temp_type, 0, sizeof(temp_type));

    temp_type.id = *typep; /* outer wrapper ensures this is OK */
    temp_type.size = size;
    temp_type.MT = &acptMTStruct;
    temp_type.data_detail = 0;   /* genuinely not used */
    temp_type.compound_length = 1;
    temp_type.type_detail = ACPT_STRUCT;
    temp_type.fields = fields;
    temp_type.n_fields = n_fields;

    ACPT_CHECKF(acptNewType(NULL, &temp_type, &new_type, FALSE), result);

    *typep = new_type->base_id = new_type->id;

    return ACPT_OK;

 errors:
    acptFree(fields);

    return result;
}


AcptResult ACPT_DLL_CALL
acptNewUnionType(AcptUnionElement field_specs[], Int32 n_fields,
       Int32 data_offset, Int32 discrim_offset,
       Int32 size, AcptType *typep)
    /*
     * Populates TYPEP with a reference to a new type which defines a
     * C discriminated union which can be transferred. FIELD_SPECS
     * (N_FIELDS long) describes each element in the union. The union
     * will manifest itself as a structure with two fields, the actual
     * data (at DATA_OFFSET) and the Int32 discriminator value at
     * DISCRIM_OFFSET.  SIZE should be provided as the size of the
     * struct which encapsulates the discriminated union.
     */
{
    *typep = 0;
    return acptMakeUnionTypeHelper(field_specs, n_fields, data_offset, discrim_offset,
              size, typep);
}

static AcptResult acptMakeUnionTypeHelper(AcptUnionElement field_specs[],
                 Int32 n_fields, Int32 data_offset,
                 Int32 discrim_offset,
                 Int32 size, AcptType *typep)
    /*
     * Does the actual work for NewUnionType.
     */
{
    Int32 i;
    Int32 result;
    UInt32 flags, vec_type;
    AcptUnionElement *spec;
    AcptTypeField *fields = NULL;
    AcptTypeDesc *subtype_desc, *new_type;
    AcptTypeDesc temp_type;
    AcptType subtype;

    /*
     * Make a new field vector: the desc itself is made in temporary space.
     */
    ACPT_CHECKR(acptAllocType(n_fields, AcptTypeField, &fields), result);

    for (i=0; i<n_fields; i++) {
   spec = &field_specs[i];
   subtype = spec->type.type;
   vec_type = spec->type.vec_type;
   flags = spec->flags;

   ACPT_CHECK((flags & ~ACPT_REF)==0, ACPT_EXCP_BAD_FLAGS, result);
   /*
    * We do not permit an in-line REF to a VVec field, because the in-line
    * definition says where the length goes, and this must be part of the
    * same thing as the ref to the data.
    */
   ACPT_CHECK(!((flags & ACPT_REF) && acptIsVVector(vec_type)), ACPT_EXCP_BAD_FLAGS,
         result);

   ACPT_CHECKF(acptGetTypeDesc(subtype, &subtype_desc), result);

   /*
    * Make the new subtype if more than just type specified. SUBTYPE_DESC
    * is re-populated with the new flavour by each helper.
    */
   if (acptIsVector(vec_type))
     ACPT_CHECKF(acptMakeVectorTypeHelper(&subtype_desc, spec->type.vec_type,
                      0, spec->type.length_detail,
                      spec->size, TRUE, flags), result);
   if (flags & ACPT_REF)
     ACPT_CHECKF(acptMakeRefTypeHelper(&subtype_desc, spec->size), result);

   /*
    * Set up the reference to the underlying type. We lose the REF flag
    * (if present) because it would have been coalesced into the new subtype.
    */
   fields[i].flags = flags & ~ACPT_REF; /* will be 0 */
   fields[i].extra_info = spec->discrim_value;

   /*
    * Set the space taken up by the field in the body of the struct (so
    * it can be cleared if necc - this is always the size the man gave.
    */
   fields[i].inline_size = spec->size;

   fields[i].subtype = subtype_desc;
    }

    /*
     * Get the actual type id for the new type, and set the copiers.
     */
    memset(&temp_type, 0, sizeof(temp_type));

    temp_type.id = *typep; /* outer wrapper will ensure this is 0 unless builtin */
    temp_type.size = size;
    temp_type.MT = &acptMTUnion;
    temp_type.compound_length = 1;
    temp_type.data_detail = discrim_offset;
    temp_type.data_offset = data_offset;
    temp_type.type_detail = ACPT_UNION;
    temp_type.fields = fields;
    temp_type.n_fields = n_fields;

    ACPT_CHECKF(acptNewType(NULL, &temp_type, &new_type, FALSE), result);

    *typep = new_type->base_id = new_type->id;

    return ACPT_OK;

 errors:
    acptFree(fields);

    return result;
}


AcptResult ACPT_DLL_CALL
acptNewFixedVectorType(AcptType el_type, UInt32 vec_type,
             Int32 length, AcptType *typep)
    /*
     * Populates TYPEP with a reference to a new type which defines a
     * fixed vector type field of length LENGTH. VEC_TYPE describes the
     * field, and should be one of:
     *   ACPT_FIXED_VECTOR: field is fixed length array of TYPE, the
     *       length being supplied in LENGTH.
     *   ACPT_BOUNDED_STRING: only valid in conjunction with CHAR8 and
     *       CHAR16 types, this option is used to define a null
     *       terminated 'string' field of fixed maximum length (supplied
     *       in LENGTH), populated in-situ (ie the data is not via a ref).
     *   ACPT_FIXED_VECTOR_REF:  The vector is a vector of ref's to the data
     *       rather than holding the data in-line. Note that the 'REF' aspect
     *       is ignored when type matching (so data sent as FIXED_VECTOR
     *       can be received as FIXED_VECTOR_REF and vice-versa).
     *
     * The size the field will occupy is assumed to be
     * LENGTH*element_size (where element-size is the size of a
     * pointer in the case of VECTOR_REF).  LENGTH must be >= 0.
     */
{
    AcptResult result;
    AcptTypeDesc *type_desc;
    Int32 size;

    ACPT_CHECKR(acptGetTypeDesc(el_type, &type_desc), result);

    size = length *(vec_type == ACPT_FIXED_VECTOR_REF ?
          ACPT_PTR_SIZE : type_desc->size);

    ACPT_CHECKR(acptMakeVectorTypeHelper(&type_desc, vec_type, 0, length,
                size, FALSE, 0), result);

    *typep = type_desc->id;   /* type_desc was updated in-situ */

    return ACPT_OK;
}

AcptResult ACPT_DLL_CALL
acptNewVVectorType(AcptType el_type, UInt32 vec_type,
         Int32 vec_offset, Int32 length_offset, Int32 size,
         AcptType *typep)
    /*
     * Populates TYPEP with a reference to a new type which defines a
     * variable length vector type field. The variable length array
     * will manifest itself as a structure with two fields, the length
     * (Int32) (at LENGTH_OFFSET) and the vector data (which will be a
     * pointer to dynamically allocated memory) at VEC_OFFSET. SIZE
     * should be provided as the size of the struct (which does not
     * have to comprise simply an Int32 and a pointer).
     *
     * VEC_TYPE describes the field, and should be one of:
     *   ACPT_SHORT_V_VECTOR: the field is a variable length array of TYPE,
     *       which is never longer than 255 elements long. LENGTH_OFFSET must
     *       be set to an offset in an owning struct which refers to an Int32
     *       field which will contain the length. The vector field itself
     *       will always be a reference to dynamically allocated memory.
     *   ACPT_LONG_V_VECTOR: As for ACPT_SHORT_V_VECTOR,
     *       except that there is no limit on field length.
     *   ACPT_SHORT_V_VECTOR_REF:} equivalent types, but each element of the
     *   ACPT_LONG_V_VECTOR_REF: } array is a ref to the value.
     */
{
    AcptResult result;
    AcptTypeDesc *type_desc;

    ACPT_CHECKR(acptGetTypeDesc(el_type, &type_desc), result);

    ACPT_CHECKR(acptMakeVectorTypeHelper(&type_desc, vec_type, vec_offset,
                length_offset, size, FALSE, 0), result);

    *typep = type_desc->id;   /* type_desc was updated in-situ */

    return ACPT_OK;
}

static AcptResult acptMakeVectorTypeHelper(AcptTypeDesc **type_desc,
                  UInt32 vec_type, Int32 vec_offset,
                  Int32 length_detail, Int32 size,
                  Bool internal, UInt32 owner_flags)
    /*
     * Make a new vector type, whose element type is described by
     * TYPE_DESC, and reset TYPE_DESC to refer to the new
     * type. INTERNAL should be FALSE if this is a type the user has
     * specifically asked for, TRUE if it is a sub-field being
     * constructed on the fly. SIZE should be supplied as the size of
     * the type, and the result ACPT_EXCP_FIELD_SIZE_TOO_SMALL will be
     * returned if it doesn't look big enough.
     *
     * If this is a field in a struct, OWNER_FLAGS should be the
     * accompanying flags (of which the REF aspect is of interest).
     */
{
    AcptResult result;
    AcptTypeDesc *el_type = *type_desc, *new_type = NULL;
    AcptTypeDesc temp_type;
    Int32 compound_length = 0;

    /*
     * check that the type is one of the valid ones, and that the size seems OK.
     */

    if (vec_type == ACPT_FIXED_VECTOR || vec_type == ACPT_BOUNDED_STRING) {
   if (owner_flags & ACPT_REF) {
       if (size < ACPT_PTR_SIZE)    return acptExcp(ACPT_EXCP_FIELD_SIZE_TOO_SMALL);
   } else if (size < el_type->size*length_detail)
                                    return acptExcp(ACPT_EXCP_FIELD_SIZE_TOO_SMALL);
    } else if (vec_type == ACPT_FIXED_VECTOR_REF) {
   if (owner_flags & ACPT_REF) {
       if (size < ACPT_PTR_SIZE)    return acptExcp(ACPT_EXCP_FIELD_SIZE_TOO_SMALL);
   } else if (size < ACPT_PTR_SIZE*length_detail)
                                    return acptExcp(ACPT_EXCP_FIELD_SIZE_TOO_SMALL);
    } else if (vec_type == ACPT_SHORT_V_VECTOR ||
          vec_type == ACPT_LONG_V_VECTOR ||
          vec_type == ACPT_SHORT_V_VECTOR_REF ||
          vec_type == ACPT_LONG_V_VECTOR_REF) {
   if (internal && size < ACPT_PTR_SIZE ||
       !internal && size < ACPT_PTR_SIZE+ACPT_INT_SIZE)
                                    return acptExcp(ACPT_EXCP_FIELD_SIZE_TOO_SMALL);
    } else return acptExcp(ACPT_EXCP_BAD_VECTOR_TYPE);

    /*
     * Special processing for bounded strings: can only be used with
     * underlying char types, and as we know length and size, can
     * decide whether long or short.
     */
    if (vec_type == ACPT_BOUNDED_STRING) {
   if (el_type->base_id != ACPT_TYPE_CHAR8 && el_type->base_id != ACPT_TYPE_CHAR16)
     return acptExcp(ACPT_EXCP_BAD_BOUNDED_STRING);
   if (length_detail > 255) vec_type = ACPT_LONG_BOUNDED_STRING;
   else vec_type = ACPT_SHORT_BOUNDED_STRING;
    }

    /*
     * Set the size. For in-line fixed size things, this is the given size;
     * For REF'd fixed size things, calculate what the size will be. For
     * internally generated V things, set to 0, as we compute the size on the fly.
     */
    if (internal && acptIsVVector(vec_type)) size = 0;
    else if (owner_flags & ACPT_REF) {
   if (vec_type == ACPT_FIXED_VECTOR_REF) size = ACPT_PTR_SIZE*length_detail;
   else if (vec_type == ACPT_FIXED_VECTOR ||
       vec_type == ACPT_BOUNDED_STRING) size = el_type->size*length_detail;
    }

    /*
     * If all that remains on the stream can be simply read by a raw copier,
     * set a compound length for it, and set its method table to be the subtype's
     * MT, so we simply drop in and start processing.
     */
    memset(&temp_type, 0, sizeof(temp_type));

    if (vec_type == ACPT_FIXED_VECTOR && size == el_type->size*length_detail) {
   temp_type.compound_length = length_detail*el_type->compound_length;
   temp_type.MT = el_type->MT;
   temp_type.stream_size = el_type->stream_size;

   /*
    * We could have a fixed vector of structs or unions... if so,
    * replicate the structure information locally.
    */
   if (el_type->n_fields) {
       temp_type.n_fields = el_type->n_fields;
       ACPT_CHECKR(acptAllocType(el_type->n_fields, AcptTypeField,
                  &temp_type.fields), result);
       memcpy(temp_type.fields, el_type->fields,
         el_type->n_fields*sizeof(AcptTypeField));
   }
    } else {
   temp_type.compound_length = 0;
   temp_type.MT = &acptMTVec;
    }
    /*
     * If this is an internal (ie struct-field, defined on-the-fly) V vector,
     * then offset the data and length so the data is treated as being at 0
     * (as this is where we'll be referencing in the copier).
     */
    if (internal && acptIsVVector(vec_type)) {
   length_detail -= vec_offset;
   vec_offset = 0;
    }

    /*
     * Now construct the type.
     */
    temp_type.id = 0;      /* ensure new one each time */
    temp_type.type_detail = vec_type;
    temp_type.data_offset = vec_offset;
    temp_type.data_detail = length_detail;
    temp_type.el_subtype = el_type;
    temp_type.size = size;

    ACPT_CHECKF(acptNewType(*type_desc, &temp_type, type_desc, internal), result);

    (*type_desc)->base_id = (*type_desc)->id;

    return ACPT_OK;
 errors:
    acptFree(temp_type.fields);
    return result;
}


static AcptResult acptMakeRefTypeHelper(AcptTypeDesc **type_desc, Int32 size)
    /*
     * Make a new type which is a REF to the type in TYPE_DESC, repopulating
     * TYPE_DESC with the result.
     */
{
    AcptResult result;
    Bool t_ref = acptIsRef((*type_desc)->type_detail);
    AcptTypeDesc temp_type, *el_type = *type_desc;

    if (size < ACPT_PTR_SIZE) return acptExcp(ACPT_EXCP_FIELD_SIZE_TOO_SMALL);

    if (t_ref) return acptExcp(ACPT_EXCP_UNSUPPORTED_REF_REF); /* shouldn't be possible */
    /*
     * Construct a new type referring to the old one, with the REF bit set. Note
     * that the base_id is retained from the original type.
     */
    memset(&temp_type, 0, sizeof(temp_type));
    temp_type.MT = &acptMTRef;
    temp_type.type_detail = ACPT_REF_MASK;
    temp_type.el_subtype = el_type;
    temp_type.size = ACPT_PTR_SIZE;
    temp_type.compound_length = 0;
    temp_type.base_id = (*type_desc)->base_id;

    ACPT_CHECKR(acptNewType(*type_desc, &temp_type, type_desc, TRUE), result);

    return ACPT_OK;
}


AcptResult ACPT_DLL_CALL acptFreeType(AcptType type)
    /*
     * Free TYPE so that space it required can be reused.
     *
     * Types are internally heavyweight, so it is desirable to free a
     * type which is no longer required as soon as possible. Note that
     * it is legitimate to Free a type as soon as access is no longer
     * required to it explicitly - even though it is referred to as
     * part of another type. Internal ref-counting ensures that the
     * type will only really die when there are genuinely no
     * references. This means that it is legitimate to define a type
     * temporarily, incorporate it into something else (eg a vector or
     * struct) and then immediately free it again.
     */
{
    return acptDerefType(type);
}


/*
 *       R E A D I N G   A N D   W R I T I N G
 *       ______________________________________
 */

AcptResult ACPT_DLL_CALL
acptReadObject(AcptStream *stream, AcptType type, void *datum, Int32 datum_size)
    /*
     * Read the next object from STREAM, which must be compatible with TYPE,
     * into the space at address DATUM which is DATUM_SIZE bytes long. The
     * data must have been sent using writeObject functionality: if
     * writeObjectVector functionality was used, then readObjectVector
     * must be used to read the data.
     *
     * Note: In the majority of cases where this function returns an
     * exception, the inbound type can queried (using
     * acptGetTypeAndLength()) and another attempt made to read the
     * data (eg using readObjectVector, or providing a different type
     * or size) - (so read-failure at this juncture is typically
     * non-destructive).
     */
{
    AcptResult result;
    Int32 length;
    AcptTypeDesc *type_desc;

    ACPT_CHECKR(acptPrepareToReadObject(stream, type, FALSE, &length, &type_desc),
      result);
    if (type_desc->size > datum_size) return acptExcp(ACPT_EXCP_SIZE_TOO_SMALL);

    ACPT_CHECKR(acptDoRead(type_desc, stream, (Buffer) datum, 1), result);

    acptInvalidateCurrentObject(stream);

    return ACPT_OK;
}

AcptResult ACPT_DLL_CALL
acptReadObjectVector(AcptStream *stream, AcptType type,
           void *datums, Int32 space, Int32 *n_objects)
    /*
     * Read a homogeneous vector of objects from STREAM, which must be
     * of a type compatible with TYPE, into the space at address
     * DATUMS which is SPACE bytes long. The data must have been sent
     * using writeObjectVector functionality: if writeObject
     * functionality was used, then readObject must be used to read
     * the data. N_OBJECTS (if not NULL) will be populated with the
     * number of objects retrieved.
     *
     * DATUMS will be populated with in-line objects.
     *
     * Note: In the majority of cases where this function returns an
     * exception, the inbound type can queried (using
     * acptGetTypeAndLength()) and another attempt made to read the
     * data (eg using readObject, or providing a different type
     * or size) - (so read-failure at this juncture is typically
     * non-destructive).
     */
{
    AcptResult result;
    Int32 length;
    AcptTypeDesc *type_desc;

    ACPT_CHECKR(acptPrepareToReadObject(stream, type, TRUE, &length, &type_desc),
      result);
    if (type_desc->size*length > space) return acptExcp(ACPT_EXCP_SIZE_TOO_SMALL);

    /*
     * Now transfer to the user's space.
     */
    if (n_objects) *n_objects = length;

    ACPT_CHECKR(acptDoRead(type_desc, stream, (Buffer) datums, length), result);

    acptInvalidateCurrentObject(stream);

    return ACPT_OK;
}

xstatic AcptResult acptPrepareToReadObject(AcptStream *stream, AcptType type,
                  Bool want_vector, Int32 *length,
                  AcptTypeDesc **type_desc)
    /*
     * Do all the preamble to reading an object. WANT_VECTOR says
     * whether the caller wants a vector or a scalar get, and it must
     * match what was sent.
     */
{
    AcptResult result;
    Bool is_vector;
    AcptType stream_type;
    AcptTypeDesc *stream_type_desc;

    /*
     * Find out what kind of object we have, without getting the data yet.
     */
    ACPT_CHECKR(acptGetObjectDetails(stream, &stream_type, length, &is_vector),
      result);

    if (is_vector && !want_vector) return acptExcp(ACPT_EXCP_VECTOR_SENT);
    if (!is_vector && want_vector) return acptExcp(ACPT_EXCP_SCALAR_SENT);

    ACPT_CHECKR(acptGetTypeDesc(type, type_desc), result);

    /*
     * validate the stream type: if an empty vector was sent, always permit
     * EMPTY_VECTOR as a permissible type either way.
     */
    if (is_vector && *length==0 &&
   (stream_type == ACPT_TYPE_EMPTY_VECTOR || type == ACPT_TYPE_EMPTY_VECTOR))
      return ACPT_OK;

    ACPT_CHECKR(acptGetStreamTypeDesc(stream, stream_type, &stream_type_desc),
      result);

    if (stream_type_desc->matching_read_seq == (*type_desc)->matching_read_seq)
      return ACPT_OK;

    /*
     * Verify TYPE's shape is compatible with what's on the stream, preparing
     * error record first.
     */
    stream->Errors.depth = 0;
    stream->Errors.code = ACPT_OK;
    ACPT_CHECKF(acptCompareTypes(stream, (*type_desc), stream_type_desc), result);

    return ACPT_OK;

 errors:
    if (acptTypeMismatchHandler) (*acptTypeMismatchHandler)(&stream->Errors);
    return result;
}


AcptResult ACPT_DLL_CALL
acptWriteObject(AcptStream *stream, AcptType type, void *datum)
    /*
     * Write the object at DATUM which matches type TYPE to STREAM. Note that
     * any errors which arise are non-destructive - in all cases other than
     * the stream itself being unwritable, nothing is written, and if it was
     * physically unwritable nothing would be written on the stream.
     */
{
    AcptResult result;
    AcptTypeDesc *type_desc;

    stream->any_type_sent = FALSE;
    acptInitStreamForWrite(stream);

    ACPT_CHECKF(acptPrepareToWriteObject(stream, type, 1, FALSE, &type_desc),
      result);

    ACPT_CHECKF(acptDoWrite(type_desc, (Buffer) datum, stream, 1), result);

    ACPT_CHECKF(acptSendBuffer(stream), result);
    if (stream->any_type_sent) acptClearSentThisGo(stream);

    return ACPT_OK;

 errors:
    if (stream->any_type_sent) acptUnsendTypes(stream);
    return result;
}

AcptResult ACPT_DLL_CALL
acptWriteObjectVector(AcptStream *stream, AcptType type, void *datums,
            Int32 length)
    /*
     * Write LENGTH objects in DATUMS which match type TYPE to STREAM.
     *
     * The objects in DATUMS are assumed to be in-line (ie not references).
     *
     * Note that any errors which arise are non-destructive - in all
     * cases other than the stream itself being unwritable, nothing is
     * written, and if it was physically unwritable nothing would be
     * written on the stream.
     */
{
    AcptResult result;
    AcptTypeDesc *type_desc;

    stream->any_type_sent = FALSE;
    acptInitStreamForWrite(stream);

    ACPT_CHECKF(acptPrepareToWriteObject(stream, type, length, TRUE, &type_desc),
      result);

    ACPT_CHECKF(acptDoWrite(type_desc, (Buffer) datums, stream, length), result);

    ACPT_CHECKF(acptSendBuffer(stream), result);
    if (stream->any_type_sent) acptClearSentThisGo(stream);

    return ACPT_OK;

 errors:
    if (stream->any_type_sent) acptUnsendTypes(stream);
    return result;
}

xstatic AcptResult acptPrepareToWriteObject(AcptStream *stream, AcptType type,
                   Int32 length, Bool is_vector,
                   AcptTypeDesc **type_desc)
    /*
     * Actually write the object(s) in DATUMS.
     */
{
    AcptResult result;

    ACPT_CHECKR(acptGetTypeDesc(type, type_desc), result);
    ACPT_CHECKR(acptPrepareToSendObject(stream, (*type_desc)), result);
    ACPT_CHECKR(acptSetObjectDetails(stream, (*type_desc)->base_id,
                 length, is_vector), result);

    return ACPT_OK;
}

AcptResult ACPT_DLL_CALL acptFreeObject(AcptType type, void *datum)
  /*
   * Free any space allocated during acptReadObject() to DATUM which
   * is an object of type TYPE.
   *
   * Note that the object memory itself is *not* freed - only space which it
   * refers to. For successful operation, the _unset flag on a field, and the
   * length field of a variable length array, must be used correctly:
   * If a field can be unset, and *is* unset, then space it may potentially
   * refer to is ignored (so if a field is marked _unset, the caller should
   * also ensure that the object it referred to is freed at the same time).
   * Similarly, if the length of a variable length array is 0, the space
   * reference by the array is not touched.
   */
{
    AcptResult result;
    AcptTypeDesc *type_desc;

    ACPT_CHECKR(acptGetTypeDesc(type, &type_desc), result);
    ACPT_CHECKR(acptDoFree(type_desc, (Buffer) datum, 1), result);
    return ACPT_OK;
}


AcptResult ACPT_DLL_CALL
acptFreeObjectVector(AcptType type, void *datums, Int32 length)
    /*
     * Free any space allocated during acptReadObjectVector() to DATUMs which
     * is an in-line vector of LENGTH objects of type TYPE.
     *
     * Note that DATUMS itself is *not* freed - only space which it
     * refers to. For successful operation, the _unset flag on a field, and the
     * length field of a variable length array, must be used correctly:
     * If a field can be unset, and *is* unset, then space it may potentially
     * refer to is ignored (so if a field is marked _unset, the caller should
     * also ensure that the object it referred to is freed at the same time).
     * Similarly, if the length of a variable length array is 0, the space
     * reference by the array is not touched.
     */
{
    AcptResult result;
    AcptTypeDesc *type_desc;

    ACPT_CHECKR(acptGetTypeDesc(type, &type_desc), result);
    ACPT_CHECKR(acptDoFree(type_desc, (Buffer) datums, length), result);
    return ACPT_OK;
}


AcptResult ACPT_DLL_CALL acptSkipObject(AcptStream *stream)
    /*
     * Skip the next or pending (if the previous request was a failed Read,
     * or a Peek) object (or vector of objects) on STREAM.
     */
{
    AcptResult result;

    ACPT_CHECKR(acptGetObjectDetails(stream, NULL, NULL, NULL), result);

    acptInvalidateCurrentObject(stream);

    return ACPT_OK;
}

AcptResult ACPT_DLL_CALL
acptPeekObject(AcptStream *stream, UInt32 *typeno, Bool *is_vectorp,
          Int32 *lengthp)
    /*
     * Interrogates the item on the stream which would be consumed by
     * the next call to ReadObject, ReadObjectVector or SkipNext. This
     * call is non-destructive, and may be made repeatedly before one
     * of the above calls, and also after a failed call to ReadObject
     * or ReadObjectVector. TYPENO is populated with the *stream type*
     * - this will only match a local type if it is one of the
     * built-in ones; IS_VECTORP is populated with TRUE if
     * ReadObjectVector() is required to read the object, FALSE if
     * ReadObject is required; and LENGTHP with the length if
     * IS_VECTORP is TRUE. Any of these three arguments may
     * legitimately be NULL. Note that under some circumstances, if
     * IS_VECTORP is TRUE and if LENGTHP is 0, TYPENO will be
     * ACPT_TYPE_EMPTY_VECTOR.
     */
{
    return acptGetObjectDetails(stream, (AcptType *)typeno, lengthp, is_vectorp);
}

AcptResult ACPT_DLL_CALL acptTypenoForType(AcptType type, UInt32 *typeno)
    /*
     * Populates TYPENO with the type number which will be sent
     * associated with TYPE. Note that although this is often TYPE, it
     * isn't guaranteed to be so.
     */
{
    AcptResult result;
    AcptTypeDesc *type_desc;

    ACPT_CHECKR(acptGetTypeDesc(type, &type_desc), result);
    *typeno = type_desc->base_id;

    return ACPT_OK;
}



/*
 *          H E L P E R S
 *          _____________
 */
xstatic AcptResult acptNewType(AcptTypeDesc *owner_type, AcptTypeDesc *temp_type,
                AcptTypeDesc **new_type, Bool internal)
    /*
     * Creates a new type in NEW_TYPE which has all the properties defined in
     * TEMP_TYPE. The new type is a flavour of OWNER_TYPE (if this is set),
     * and if a flavour which matches is found, then this is reused. If INTERNAL
     * is TRUE, then the new type isn't ref-counted on creation, as it is only
     * being made to be referenced from elsewhere. NEW_TYPE's id is either the
     * ID specified in TEMP_TYPE, or, if this is 0 (as is more common) a new
     * (or reused) value. The caller must ensure TEMP_TYPE.ID is set to 0 unless
     * a value is explicitly to be reused.
     */
{
    AcptResult result = ACPT_OK;
    AcptTypeDesc *p, **pp, **e, **sp=NULL;
    AcptType id = temp_type->id;
    Int32 index, new_max;
    Int32 i;

    /*
     * We are going to mess with both the global typeDescs list and with the
     * flavours vector in OWNER_TYPE, so protect wherever necc.
     */
    START_EXCLUSIVE{
   *new_type = NULL;    /* in case we have to leave early */

   /*
    * First see if there is a matching type under owner, unless this is
    * a type number being reused.
    */
   if (owner_type && owner_type->n_flavours && id == 0) {
       for (pp=owner_type->flavours, e=pp+owner_type->n_flavours; pp<e; pp++) {
      if (!(p=*pp)) {      /* remember empty slot for later */
          sp = pp;
          continue;
      }

      if (p->MT != temp_type->MT ||
          p->type_detail != temp_type->type_detail ||
          p->data_offset != temp_type->data_offset ||
          p->data_detail != temp_type->data_detail ||
          p->size != temp_type->size ||
          p->compound_length != temp_type->compound_length) continue;
      /*
       * We don't need to check fields, because OWNER_TYPE constrains what
       * the fields are.
       */
      *new_type = p;
      if (!internal) p->n_refs++;

      /*
       * temp_type is now being discarded, so if any space had been
       * allocated within it it should be freed.
       */
      acptFree(temp_type->fields);

      goto end;
       }
   }

   /*
    * No match. If generating an id, look for a NULL entry in the
    * typeDescs list. If none found before the end then attempt to
    * grow the list and then use the first new entry.
    */
   if (id == 0) {
       for(pp=typeDescs, e=pp+typeDescListMax; pp<e && *pp; pp++);

       if (pp==e) {        /* Grow the list */
      ACPT_CHECK((pp-typeDescs)+ACPT_TYPES_INCR<=ACPT_MAX_TYPES,
            ACPT_EXCP_NO_MORE_TYPES, result);

      ACPT_CHECKF(acptRealloc(typeDescListMax+ACPT_TYPES_INCR,
               AcptTypeDesc *, &typeDescs), result);

      pp=typeDescs+typeDescListMax; /* reset after realloc */
      typeDescListMax+=ACPT_TYPES_INCR;
      memset(pp, 0, ACPT_TYPES_INCR*sizeof(AcptTypeDesc **));
       }
       id = (AcptType)((pp-typeDescs)+ACPT_MAX_BUILTIN_TYPES);

   } else {       /* number given - ensure list long enough */
       if (id < ACPT_MAX_BUILTIN_TYPES) pp = builtinTypeDescList+id;
       else {
      index = id-ACPT_MAX_BUILTIN_TYPES;
      if (index >= typeDescListMax) {
          ACPT_CHECK(index <= ACPT_MAX_TYPES, ACPT_EXCP_NO_MORE_TYPES,
                result);
          if ((new_max = index+ACPT_TYPES_INCR) > ACPT_MAX_TYPES)
            new_max = index+1;

          ACPT_CHECKF(acptRealloc(new_max, AcptTypeDesc *, &typeDescs),
            result);

          memset(typeDescs+typeDescListMax, 0,
            (new_max-typeDescListMax)*sizeof(AcptTypeDesc **));
          typeDescListMax = new_max;
      }
      pp = typeDescs+index;
       }

       /*
        * See if type already defined... if so we really have no
        * choice but to assume that it's the right one, and
        * simply return the information.
        */
       if (*pp) {
      *new_type = *pp;
      if (!internal) (*pp)->n_refs++;

      goto end;
       }
   }

   /*
    * Make the new type, and copy over the detail from temp_type.
    */
   ACPT_CHECKF(acptAllocType(1, AcptTypeDesc, new_type), result);
   memcpy((Buffer)(*new_type), (Buffer)temp_type, sizeof(AcptTypeDesc));

   *pp = (*new_type);

   (*new_type)->id = id;
   (*new_type)->matching_read_seq = ++matchSeq; /* unique on each new type */
   if (!internal) (*new_type)->n_refs++;

   /*
    * Reference any substructure here. We always do this, so Free can always
    * dumbly de-reference.
    */
   if ((*new_type)->el_subtype) (*new_type)->el_subtype->n_refs++;

   for (i=0; i<(*new_type)->n_fields; i++)
     (*new_type)->fields[i].subtype->n_refs++;

   /*
    * Now hang this new type off its parent (if defined).
    */
   if (owner_type) {
       if (sp) *sp = (*new_type);
       else {
      ACPT_CHECKF(acptRealloc(owner_type->n_flavours+1, AcptTypeDesc *,
               &owner_type->flavours), result);
      owner_type->flavours[owner_type->n_flavours++] = (*new_type);
       }
       (*new_type)->owner_id = owner_type->id;
   }

   goto end;

    errors:
   acptFree(*new_type);

    end:;

    } END_EXCLUSIVE;

    return result;
}

static AcptResult acptDerefType(AcptType type)
    /*
     * Dereference Type, and if the ref count reaches 0 then free the space for
     * it, dereferencing anything it refers to first. Ignore builtin types.
     */
{
    AcptResult result = ACPT_OK;
    Int32 i;
    UInt32 id, sent_byte;
    unsigned char sent_mask;

    AcptTypeDesc *type_desc, *owner_desc, **pp, **e;
    AcptStream *stream;

    if (type < ACPT_MAX_BUILTIN_TYPES) return ACPT_OK;

    START_EXCLUSIVE {

   ACPT_CHECKF(acptGetTypeDesc(type, &type_desc), result);

   if (--type_desc->n_refs > 0) goto end;

   /*
    * We want to free it, so knock if from the list, then get rid of any
    * substructure. We can do this dumbly, because NewType always references
    * substructure dumbly.
    */
   id = type_desc->id-ACPT_MAX_BUILTIN_TYPES;
   typeDescs[id] = NULL;

   for (i=0; i<type_desc->n_fields; i++)
     acptDerefType(type_desc->fields[i].subtype->id);

   if (type_desc->el_subtype) acptDerefType(type_desc->el_subtype->id);

   /*
    * Look for _self in owning structure - if present, nullify. Note to
    * debugger users... this piece of code may legitimately return
    * TYPE_NOT_FOUND, so raising a spurious break in acptExcp.
    */
   if (type_desc->owner_id &&
       acptGetTypeDesc(type_desc->owner_id, &owner_desc) == ACPT_OK &&
       owner_desc->n_flavours > 0) {
       for (pp=owner_desc->flavours, e=pp+owner_desc->n_flavours;
       pp<e && *pp != type_desc; pp++);
       if (pp < e) *pp = NULL;
   }

   /*
    * Remove from any SentTypes lists in write streams.
    */
   sent_byte = id >> 3;
   sent_mask = ~(1 << (id & 7));

   for (i=0; i<writeStreamsMax; i++) {
       if ((stream = writeStreams[i]) && sent_byte < (UInt32)stream->SentTypesMax)
         stream->SentTypes[sent_byte] &= sent_mask;
   }

   /*
    * If being used internally by the datastore, each Field has an extra
    * slot used to hold the unset value (if set). Free the space.
    */
#if defined(_ACPTM_C_ONLY_)
   for (i=0; i<type_desc->n_fields; i++)
     acptFree(type_desc->fields[i].unset_value);
#endif
   acptFree(type_desc->flavours);
   acptFree(type_desc->fields);
   acptFree(type_desc);

    errors:;
    end:;
    } END_EXCLUSIVE;

    return result;
}

xstatic AcptResult acptGetTypeDesc(AcptType type, AcptTypeDesc **type_desc)
    /*
     * Given type TYPE, return the corresponding desc for it from the descs
     * list TYPE_DESCS which is N_DESCS long, or, if the type is builtin,
     * from the builtin list. Returns ACPT_EXCP_NO_SUCH_TYPE if the type is
     * unknown, or ACPT_EXCP_NULL_TYPE if 0 (used as a primitive check).
     */
{
    AcptResult result = ACPT_OK;

    if (!type) return acptExcp(ACPT_EXCP_NULL_TYPE);

    if (type < ACPT_MAX_BUILTIN_TYPES) {
   if (!(*type_desc = builtinTypeDescList[type]))
     return acptExcp(ACPT_EXCP_NO_SUCH_TYPE);
    } else {
   START_EXCLUSIVE {

       type -= ACPT_MAX_BUILTIN_TYPES;
       if ((Int32)type >= typeDescListMax || !(*type_desc = typeDescs[type]))
         result = acptExcp(ACPT_EXCP_NO_SUCH_TYPE);

   } END_EXCLUSIVE;
    }

    return result;
}


static AcptResult acptGetStreamTypeDesc(AcptStream *stream, AcptType type,
               AcptTypeDesc **type_desc)
    /*
     * Given stream type TYPE, return the corresponding desc for
     * it. Returns ACPT_EXCP_NO_SUCH_TYPE if the type is unknown, or
     * ACPT_EXCP_NULL_TYPE if 0 (used as a primitive check).
     */
{
    if (!type) return acptExcp(ACPT_EXCP_NULL_TYPE);

    if ((Int32)type < ACPT_MAX_BUILTIN_TYPES) {
   if (!(*type_desc = builtinTypeDescList[type]))
     return acptExcp(ACPT_EXCP_NO_SUCH_TYPE);
    } else {
   type -= ACPT_MAX_BUILTIN_TYPES;
   if ((Int32)type >= stream->typeDescListMax ||
       !(*type_desc = stream->typeDescs[type]))
     return acptExcp(ACPT_EXCP_NO_SUCH_TYPE);
    }

    return ACPT_OK;
}

static AcptResult acptGetObjectDetails(AcptStream *stream, AcptType *typep,
                   Int32 *lengthp, Bool *is_vectorp)
    /*
     * Returns details of the next object on STREAM. This interface may
     * be called repeatedly: an object is read once, and the same information
     * is then returned for it each time. Any of the result pointers may
     * legitimately be NULL (in which case it is ignored).
     */
{
    AcptResult result;
    UInt32 details, flags;
    Int32 length=1;
    Bool is_vector;
    AcptType type;

    if (!stream->object_current) { /* need to read a new one */
   if (!stream->reader) return acptExcp(ACPT_EXCP_STREAM_NOT_READER);

   ACPT_CHECKR(acptReceiveBuffer(stream), result);

   /*
    * Check incoming data to see if it is actually a type definition, in which
    * case absorb it and try again.
    */
   while (1) {
       ACPT_CHECKR(acptReadUInt32(stream, &details, 1), result);

       if (details != ACPT_TYPE_DEFINITION) break;

       ACPT_CHECKR(acptReceiveType(stream), result);
   }

   type = details >> 16;
   flags = (details >> 8) & 0xff; /* only one flag used currently! */

   if (is_vector = acptIsVector(flags))
     ACPT_CHECKR(acptReadUInt32(stream, (UInt32 *)&length, 1), result);

   stream->object_current = TRUE;
   stream->current_object_type = type;
   stream->current_object_is_vector = is_vector;
   stream->current_object_length = length;
   stream->data_start = stream->buffer;
    } else stream->buffer = stream->data_start;

    if (typep) *typep = stream->current_object_type;
    if (is_vectorp) *is_vectorp = stream->current_object_is_vector;
    if (lengthp) *lengthp = stream->current_object_length;

    return ACPT_OK;
}


static AcptResult acptSetObjectDetails(AcptStream *stream, AcptType type, Int32 length,
                   Bool is_vector)
    /*
     * Compile the details of object type TYPE (or a vector of objects, if
     * IS_VECTOR is TRUE).
     */
{
    AcptResult result;
    UInt32 flags=0, details=0;

    /*
     * Compose first word, which contains TYPE and FLAGS.
     */
    if (is_vector) flags |= ACPT_VECTOR_MASK;

    details = (type << 16) | (flags << 8);

    ACPT_CHECKR(acptWriteUInt32(stream, &details, 1), result);

    /*
     * If a vector, add the length. This is always a whole word.
     */
    if (is_vector)
      ACPT_CHECKR(acptWriteUInt32(stream, (UInt32 *)&length, 1), result);

    return ACPT_OK;
}


static AcptResult acptPrepareToSendObject(AcptStream *stream, AcptTypeDesc *type_desc)
    /*
     * Prepare to send an object of the type TYPE_DESC to STREAM. This function
     * ensures that the other end is sent all the types necessary to define
     * TYPE_DESC.
     */
{
    if (stream->reader) return acptExcp(ACPT_EXCP_STREAM_NOT_WRITER);

    if (type_desc->base_id < ACPT_MAX_BUILTIN_TYPES) return ACPT_OK;

    return acptSendType(stream, type_desc);
}


static AcptResult acptSendType(AcptStream *stream, AcptTypeDesc *type_desc)
    /*
     * Send a type definition to the recipient. Note that this function is
     * recursive, and sends un-sent lower level types before sending TYPE_DESC.
     */
{
    AcptResult result=ACPT_OK;
    AcptTypeField *field;
    UInt32 type_detail, id, sent_byte;
    unsigned char sent_mask;
    Int32 i, n_words;
    Bool is_struct;
    UInt32 vec_buf[6], *struct_buff, *p;

    /*
     * If indirect, we have a type with local structuring information where
     * we really want its subtype.
     */
    if (acptIsIndirect(type_desc->type_detail)) type_desc = type_desc->el_subtype;

    /*
     * See if this type has already been sent. Builtin types never need to be.
     */
    if (type_desc->base_id < ACPT_MAX_BUILTIN_TYPES) return ACPT_OK;

    /*
     * See if the bit is sent in the SentTypes structure. This doesn't need
     * protecting as the only other code which could be looking here is cleanup
     * code, which will simply reset bits.
     */
    id = type_desc->base_id-ACPT_MAX_BUILTIN_TYPES;
    sent_byte = id >> 3;
    sent_mask = 1 << (id & 7);
    if (sent_byte < (UInt32)stream->SentTypesMax &&
   (stream->SentTypes[sent_byte] & sent_mask))
      return ACPT_OK;

    type_detail = type_desc->type_detail;

    /*
     * Send any substructure.
     */
    if (acptIsVector(type_detail)) {
   ACPT_CHECKR(acptSendType(stream, type_desc->el_subtype), result);
    } else if (acptHasStructure(type_detail)) {
   for (i=0; i<type_desc->n_fields; i++)
     ACPT_CHECKR(acptSendType(stream, type_desc->fields[i].subtype), result);
    } else return acptExcp(ACPT_EXCP_BAD_SEND_TYPE);

    /*
     * Now we can send self. Send TYPE_DEF, id, flags, subtype_id,
     * length for vector, marker, id, flags, n_subtypes, {extra_info,
     * subtype_id} for struct things. Make the whole lot up as a
     * vector, and send as a bunch.
     */
    if (acptIsVector(type_detail)) {
   n_words = acptIsVVector(type_detail) ? 4 : 5;
   vec_buf[0] = ACPT_TYPE_DEFINITION;
   vec_buf[1] = type_desc->base_id;
   vec_buf[2] = acptInterestingFlags(type_detail);
   vec_buf[3] = type_desc->el_subtype->base_id;
   vec_buf[4] = type_desc->data_detail;   /* will not be sent if VLen */
   ACPT_CHECKR(acptWriteUInt32(stream, vec_buf, n_words), result);

    } else {
   n_words = 4+2*type_desc->n_fields;

   /*
    * Don't use malloc here, because we can get in a mess. Simple write
    * directly into the stream buffer.
    */
   ACPT_CHECKR(acptMakeStreamSpaceFor(stream, n_words*ACPT_INT_SIZE), result);
   struct_buff = (UInt32 *)stream->buffer;
   struct_buff[0] = ACPT_TYPE_DEFINITION;
   struct_buff[1] = type_desc->base_id;
   struct_buff[2] = acptInterestingFlags(type_detail);
   struct_buff[3] = type_desc->n_fields;

   is_struct = acptIsStruct(type_detail);

   for (p=&struct_buff[4],i=0; i<type_desc->n_fields; i++) {
       field = &type_desc->fields[i];
       if (is_struct) *p++ = (field->flags & ACPT_UNSET);
       else *p++ = field->extra_info; /* discrim value */
       *p++ = field->subtype->base_id;
   }
   stream->buffer += n_words*ACPT_INT_SIZE;
    }

    /*
     * Now go exclusive, as we are updating the SentTypes vector and we don't
     * want anyone else in at the same time.
     */
    START_EXCLUSIVE {
   if (sent_byte >= (UInt32)stream->SentTypesMax) {
       if (((result = acptRealloc(sent_byte+1, unsigned char,
                   &stream->SentTypes)) != ACPT_OK) ||
      ((result = acptRealloc(sent_byte+1, unsigned char,
                   &stream->SentTypesThisGo)) != ACPT_OK)) {
      goto end;
       }
       memset(stream->SentTypes+stream->SentTypesMax, 0,
         sent_byte+1-stream->SentTypesMax);
       memset(stream->SentTypesThisGo+stream->SentTypesMax, 0,
         sent_byte+1-stream->SentTypesMax);
       stream->SentTypesMax = sent_byte+1;
   }

   stream->SentTypes[sent_byte] |= sent_mask;
   stream->SentTypesThisGo[sent_byte] |= sent_mask;
   stream->any_type_sent = TRUE;

    end:;
    } END_EXCLUSIVE;

    return result;
}


xstatic AcptResult acptUnsendTypes(AcptStream *stream)
    /*
     * Unset the last set of types sent to this stream.
     */
{
    Int32 i;
    unsigned char *p1, *p2;

    for (i=0,p1=stream->SentTypes,p2=stream->SentTypesThisGo;
    i<stream->SentTypesMax; i++,p1++,p2++) {
   *p1 &= ~*p2;
   *p2 = 0;
    }

    return ACPT_OK;
}

static AcptResult acptReceiveType(AcptStream *stream)
    /*
     * Construct a new stream type from data coming in off the stream.
     */
{
    AcptResult result;
    UInt32 type_detail;
    AcptType id;
    Int32 n_fields, i;
    UInt32 buffer[3], *p, *extra_info=NULL;
    Bool is_struct;
    AcptTypeField *fields = NULL, *field;
    AcptTypeDesc *type_desc = NULL, *el_subtype;
    char *c;

    /* allocate only the minimum space necessary */
    ACPT_CHECKF(acptAllocType(ACPT_STREAM_TYPE_SIZE, char, &c), result);
    type_desc = (AcptTypeDesc *)c;

    ACPT_CHECKF(acptReadUInt32(stream, buffer, 3), result); /* id, type_detail, sub/nf */

    type_desc->base_id = id = buffer[0];
    type_desc->type_detail = type_detail = buffer[1];

    /*
     * NOTE: we deliberately don't bother ref-counting stream types, because
     * this is being done for us at the other end, and the counting would
     * only serve as a validation... and to be done properly would require
     * much more complexity in the sending code. Consider: we have a struct
     * which references the same type twice. The sender frees both the struct
     * type and the referenced type. It then reuses the referenced type, and
     * the receiver gets a request to re-use it: however, it can't, because
     * it hasn't had word yet that the struct type isn't required, so it is
     * still referenced. The only real soloution would be to send lists of
     * types which had become free, before any attempt at re-use. However,
     * this would be both complex and memory hungry, and so isn't implemented.
     */
    if (acptIsVector(type_detail)) {
   ACPT_CHECKF(acptGetStreamTypeDesc(stream, buffer[2], &el_subtype), result);
   type_desc->el_subtype = el_subtype;

   if (!acptIsVVector(type_detail))
     ACPT_CHECKF(acptReadUInt32(stream, (UInt32 *)&type_desc->data_detail, 1), result);

    } else {
   type_desc->n_fields = n_fields = buffer[2];
   is_struct = acptIsStruct(type_detail);

   /*
    * The remaining detail contains unset_flag/discrim_value, sub_type pairs
    */
   ACPT_CHECKF(acptAllocType(n_fields, AcptTypeField, &fields), result);
   type_desc->fields = fields;

   ACPT_CHECKF(acptAllocType(n_fields*2, UInt32, &extra_info), result);
   ACPT_CHECKF(acptReadUInt32(stream, extra_info, n_fields*2), result);

   for (i=0, p=extra_info; i<n_fields; i++) {
       field = &fields[i];

       if (is_struct) {
      if (*p++) field->flags = ACPT_UNSET;
       } else {
      field->extra_info = *p++;
       }

       ACPT_CHECKF(acptGetStreamTypeDesc(stream, (AcptType)*p++, &el_subtype),
         result);
       field->subtype = el_subtype;

   }
    }

    /*
     * Finally, populate stream type table with this new type. If there already was
     * such a type, dereference it. A builtin type should never come through.
     */
    id -= ACPT_MAX_BUILTIN_TYPES;

    if ((Int32)id >= stream->typeDescListMax) { /* grow the table */
   ACPT_CHECKF(acptRealloc(id+ACPT_TYPES_INCR, AcptTypeDesc *,
            &stream->typeDescs), result);
   memset(stream->typeDescs+stream->typeDescListMax, 0,
          (id+ACPT_TYPES_INCR-stream->typeDescListMax)*sizeof(AcptTypeDesc *));
   stream->typeDescListMax = id+ACPT_TYPES_INCR;
    }

    /*
     * Simple free the old type if there was one, relying on the other end to
     * ensure we don't remove something we need.
     */
    if (stream->typeDescs[id]) {
   acptFree(stream->typeDescs[id]->fields);
   acptFree(stream->typeDescs[id]);
    }

    stream->typeDescs[id] = type_desc; /* ignore refcounting entirely */

    acptFree(extra_info);

    return ACPT_OK;

 errors:
    acptFree(type_desc);
    acptFree(extra_info);
    acptFree(fields);

    return result;
}


AcptResult ACPT_DLL_CALL
acptQueryStreamType(AcptStream *stream, UInt32 typeno,
          UInt32 *flavour, Int32 *data_detail,
          Int32 *subtype, Int32 **subtypes)
    /*
     * Returns details of the stream type from STREAM whose number is
     * TYPENO.  FLAVOUR will be populated with the flags which define
     * what sort of type this is, and will be one of:
     *     ACPT_SCALAR
     *     ACPT_FIXED_VECTOR
     *     ACPT_SHORT_V_VECTOR
     *     ACPT_LONG_V_VECTOR
     *     ACPT_SHORT_BOUNDED_STRING
     *     ACPT_LONG_BOUNDED_STRING
     *     ACPT_SHORT_V_STRING
     *     ACPT_LONG_V_STRING
     *     ACPT_STRUCT
     *     ACPT_UNION
     *
     * DATA_DETAIL will be populated with a value depending on
     * FLAVOUR: if FLAVOUR is is ACPT_FIXED_VECTOR or
     * ACPT_[LONG/SHORT]_BOUNDED_STRING then DATA_DETAIL is populated
     * with the length value; if FLAVOUR is ACPT_STRUCT or ACPT_UNION
     * then it is populated with the number of subtypes. If FLAVOUR
     * indicates a vector type, SUBTYPE is populated with the typeno
     * of the subtype (otherwise it's NULLed). If FLAVOUR is
     * ACPT_STRUCT or ACPT_UNION, SUBTYPES is populated with an
     * integer vector containing the typnos or the underlying fields
     * and followed by the same number of elements containing flags
     * (indicating _unset allowed, if FLAVOUR is ACPT_STRUCT) or
     * discriminators (if FLAVOUR is ACPT_UNION) - so the vector is
     * 2*data_detail long.
     *
     * If SUBTYPES is not relevant the argument is populated with
     * NULL; if it is relevant then it is the caller's responsibility
     * to free the space using acptFree() when no longer required.
     */
{
    AcptResult result;
    AcptTypeDesc *type_desc;
    UInt32 type_detail;
    Int32 n_fields, i, *p1, *p2;
    AcptTypeField field;

    ACPT_CHECKR(acptGetStreamTypeDesc(stream, typeno, &type_desc), result);
    type_detail = type_desc->type_detail;

    *flavour = acptInterestingFlags(type_detail);

    if (acptIsVector(*flavour)) {
   *data_detail = acptIsFixedVector(*flavour) ? type_desc->data_detail : 0;

   *subtype = type_desc->el_subtype->base_id;
   *subtypes = NULL;

    } else if (acptHasStructure(*flavour)) {
   n_fields = *data_detail = type_desc->n_fields;
   *subtype = 0;
   ACPT_CHECKR(acptAllocType(2*type_desc->n_fields, Int32, subtypes),
          result);

   for (p1=*subtypes, p2=*subtypes+n_fields, i=0; i<n_fields; p1++,p2++,i++) {
       field = type_desc->fields[i];
       *p1 = field.subtype->base_id;
       *p2 = acptIsStruct(*flavour) ? (Int32)field.flags : field.extra_info;
   }
    } else {
   *data_detail = *subtype = 0;
   *subtypes = NULL;
    }

    return ACPT_OK;
}



/*
 *          H A N D S H A K I N G
 *          _____________________
 */
/*
 * The following interfaces are called only the first time the stream is used,
 * and verify that the stream is behaving correctly and soot out byte-swapping.
 */
static AcptResult acptSendIntro(AcptStream *stream)
    /*
     * The first ever call for a Write stream, places a header (so the other end can
     * verify everything's in step, and work out byte-swapping) and also a
     * format code, which must match, into the front of the buffer. P is populated
     * with the position in the buffer where the count value will be placed.
     */
{
    AcptResult result;
    /*                     nbytes                                    flags */
    UInt32 b[]={ACPT_HEADER, 0, ACPT_PROTOCOL, ACPT_MIN_READ_PROTOCOL, 0};
    Int32 header_size = sizeof(b);
    Int32 diff = stream->buffer-stream->buffer_start;
    Int32 total_size = diff+header_size,
          incr = total_size <= stream->buffer_max ? total_size :
       total_size+ACPT_STREAM_BUFFER_INCR;
    Buffer new_buffer;

    /*
     * Set the NBYTES field manually. This ignores the HEADER field and
     * NBYTES itself.
     */
    b[1] = sizeof(b)-sizeof(UInt32)*2;

    /*
     * The very first time around, we actually shove the buffer along by the
     * size of the header, so there's room for it. We always do this by
     * re-mallocing, as this is probably faster than manually copying backwards.
     */
    ACPT_CHECKR(acptAllocType(incr, BuffEl, &new_buffer), result);

    memcpy(new_buffer, b, header_size);
    memcpy(new_buffer+header_size, stream->buffer_start, diff);

    acptFree(stream->buffer_start);

    stream->buffer_max = incr;
    stream->buffer_start = new_buffer;
    stream->buffer = new_buffer+total_size;

    stream->buffer_count_offset = header_size;

    return ACPT_OK;
}

static AcptResult acptReceiveIntro(AcptStream *stream)
    /*
     * The first ever call for a Read stream, receives the header and format code.
     */
{
    AcptResult result;
    UInt32 h, count, p, flags, hs=sizeof(UInt32)*2; /* read HEADER, NBYTES first */

    ACPT_CHECKR(acptMakeStreamSpaceFor(stream, hs), result);
    ACPT_CHECKR((*stream->Transfer)(stream->handle, stream->buffer, hs), result);

    stream->started = TRUE;   /* can't do much about errors */
    stream->buffer_count = hs;

    ACPT_CHECKR(acptReadUInt32(stream, &h, 1), result);

    if (h != ACPT_HEADER) {
   acptByteSwap4InSitu(&h, 1);
   if (h != ACPT_HEADER) return acptExcp(ACPT_EXCP_BAD_INTRO);
   stream->byte_swapped = TRUE;
    }

    ACPT_CHECKR(acptReadUInt32(stream, &count, 1), result);
    ACPT_CHECKR(acptMakeStreamSpaceFor(stream, count), result);

    ACPT_CHECKR((*stream->Transfer)(stream->handle, stream->buffer, count),
      result);
    stream->buffer_count += count;

    /*
     * Read both the protocol at which the data was written, and also
     * the min current protocol which can read the data. Only the latter
     * is relevant in this rev.
     */
    ACPT_CHECKR(acptReadUInt32(stream, &p, 1), result);  /* write */
    ACPT_CHECKR(acptReadUInt32(stream, &p, 1), result);  /* min-read */

    if (p > ACPT_PROTOCOL) return acptExcp(ACPT_EXCP_BAD_PROTOCOL);

    ACPT_CHECKR(acptReadUInt32(stream, &flags, 1), result); /* ignore for now */

    /*
     * Skip any data we don't know about.
     */
    stream->buffer = stream->buffer_start+stream->buffer_count;

    return ACPT_OK;
}


static AcptResult acptCompareTypes(AcptStream *stream, AcptTypeDesc *c_type,
               AcptTypeDesc *s_type)
    /*
     * Compare the two types C_TYPE and S_TYPE: if they correspond in
     * underlying stream structure return ACPT_OK otherwise return
     * ACPT_EXCP_INCOMPATIBLE_TYPE. If the types are successfully matched
     * s_type->matching_read_seq is set to c_type_matching_read_seq, so
     * the two types can be rapidly spotted as matching again, next time
     * around. STREAM is provided so that if a mismatch error occurs, the
     * details can be recorded.
     */
{
    AcptResult result;
    Int32 i, last_error_detail;
    Bool is_struct;
    AcptTypeField *c_field, *s_field;
    UInt32 c_flags, s_flags;

    if ((c_type->base_id < ACPT_MAX_BUILTIN_TYPES &&
    c_type->base_id == s_type->base_id) ||
   s_type->matching_read_seq == c_type->matching_read_seq) return ACPT_OK;

    if (acptIsIndirect(c_type->type_detail))  /* ignore local structuring info */
      c_type = c_type->el_subtype;

    c_flags = c_type->type_detail;
    s_flags = s_type->type_detail;

    stream->Errors.depth++;
    last_error_detail = -1;   /* if error, then default to local */

    /*
     * Match flags (ignoring REF and VEC_REF, which are for C structuring only).
     * If fixed len array, check lengths match too. Then check substructure.
     */
    if (acptInterestingFlags(c_flags) != acptInterestingFlags(s_flags)) {
   stream->Errors.code = ACPT_EXCP_FLAVOUR_MISMATCH;
   stream->Errors.c_data = c_flags;
   stream->Errors.s_data = s_flags;
   goto error_here;
    }

    if (acptIsFixedVector(c_flags) && c_type->data_detail != s_type->data_detail) {
   stream->Errors.code = ACPT_EXCP_LENGTH_MISMATCH;
   stream->Errors.c_data = c_type->data_detail;
   stream->Errors.s_data = s_type->data_detail;
   goto error_here;
    }


    /*
     * If a vector, check that element shape matches, if a structured thing check
     * that each component matches.
     */
    if (acptIsVector(c_flags)) {
   last_error_detail = 0;
   ACPT_CHECKF(acptCompareTypes(stream, c_type->el_subtype,
                 s_type->el_subtype), result);
    } else if (acptHasStructure(c_flags)) {
   if (c_type->n_fields != s_type->n_fields) {
       stream->Errors.code =ACPT_EXCP_NFIELDS_MISMATCH;
       stream->Errors.c_data = c_type->n_fields;
       stream->Errors.s_data = s_type->n_fields;
       goto error_here;
   }

   is_struct = acptIsStruct(c_flags);

   for (i=0; i<c_type->n_fields; i++) {
       c_field = &c_type->fields[i];
       s_field = &s_type->fields[i];
       last_error_detail = i;

       if (is_struct) {
      if (acptInterestingFieldFlags(c_field->flags) !=
          acptInterestingFieldFlags(s_field->flags)) {
          stream->Errors.code = ACPT_EXCP_FIELD_FLAGS_MISMATCH;
          stream->Errors.c_data = c_field->flags;
          stream->Errors.s_data = s_field->flags;
          goto error_here;
      }

       } else {
      if (c_field->extra_info != s_field->extra_info) {
          stream->Errors.code = ACPT_EXCP_DISCRIM_MISMATCH;
          stream->Errors.c_data = c_field->extra_info;
          stream->Errors.s_data = s_field->extra_info;
          goto error_here;
      }
       }

       ACPT_CHECKF(acptCompareTypes(stream, c_field->subtype,
                s_field->subtype), result);
   }
    } else {
   /*
    * must be simple scalar: since the raw is test failed, types don't match
    */
   stream->Errors.code = ACPT_EXCP_SCALAR_TYPE_MISMATCH;
   stream->Errors.c_data = c_type->base_id;
   stream->Errors.s_data = s_type->base_id;
   goto error_here;
    }

    /*
     * Types match, so cache it in case we see it again.
     */
    s_type->matching_read_seq = c_type->matching_read_seq;
    stream->Errors.depth--;

    return ACPT_OK;

 error_here:
    /*
     * We get here if a mismatch has occurred *at this level*. Create space for
     * the error stack, and setup Errors.sp for higher levels.
     */
    result = acptPrepareForError(stream);

 errors:
    return acptProcessError(stream, last_error_detail, s_type, result);
}



/*
 *       D A T A   M O V E R S
 *       _____________________
 */

/*
 * This section contains the 'Methods' use to copy data to and from the
 * stream, and do any other type specific things.
 */


/*
 *       R E F _ F I E L D S
 *       ___________________
 */

static AcptResult acptReadRef(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
               Int32 n_els)
    /*
     * Create space for the type referenced by TYPE, populate TO_ADDR to
     * refer to it, and then read the subtype. Note that N_ELS is ignored,
     * as it's only possible to define scalar Ref fields currently.
     */
{
    AcptResult result;
    AcptTypeDesc *subtype = type->el_subtype;
    Buffer ref_addr;

    ACPT_CHECKR(acptAllocDataSpace(subtype->size, BuffEl, &ref_addr), result);
    *(Buffer *)to_addr = ref_addr;

    ACPT_CHECKF(acptDoRead(subtype, stream, ref_addr, 1), result);

    return ACPT_OK;

 errors:
    acptFreeDataSpace(ref_addr);
    return result;
}


static AcptResult acptWriteRef(AcptTypeDesc *type, Buffer from_addr,
                AcptStream *stream, Int32 n_els)
    /*
     * Jump thru the ref at FROM_ADDR and then write the subtype. Note
     * that N_ELS is ignored, as it's only possible to define scalar
     * Ref fields currently.
     */
{
    AcptResult result;
    AcptTypeDesc *subtype = type->el_subtype;

    stream->Errors.depth++;

    if (!(from_addr = *(Buffer *)from_addr)) {
   stream->Errors.code = acptExcp(ACPT_EXCP_NULL_ADDR);
   goto errors_here;
    }

    ACPT_CHECKF(acptDoWrite(subtype, from_addr, stream, 1), result);

    stream->Errors.depth--;
    return ACPT_OK;

 errors_here:
    result = acptPrepareForError(stream);

 errors:
    return acptProcessError(stream, -1, type, result);
}


static AcptResult acptFreeRef(AcptTypeDesc *type, Buffer from_addr, Int32 n_els)
    /*
     * Jump thru the ref at FROM_ADDR and free any data beneath it,
     * then free the subtype data itself.
     */
{
    AcptTypeDesc *subtype = type->el_subtype;

    if ((from_addr = *(Buffer *)from_addr)) {
   acptDoFree(subtype, from_addr, 1);
   acptFreeDataSpace(from_addr);
    }
    return ACPT_OK;
}



/*
 *       V E C _ F I E L D S
 *       ___________________
 */

static AcptResult acptReadVec(AcptTypeDesc *type_desc, AcptStream *stream,
               Buffer to_addr, Int32 n_els)
    /*
     * Read the vector of type defined by TYPE_DESC from STREAM into
     * TO_ADDR.
     */
{
    AcptResult result;
    Int32 i, err_i, count;
    Int32 type_detail = type_desc->type_detail;
    Int32 length_detail = type_desc->data_detail;
    Bool is_vec_ref = acptIsVecRef(type_detail);
    Int32 size = type_desc->size;
    AcptTypeDesc *subtype = type_desc->el_subtype;
    Int32 length = length_detail, sub_el_size=subtype->size;
    Buffer pos, vec_addr = NULL, to_addr1;
    Buffer saved_addr = NULL; /* for error cleanup */

    /*
     * NOTE: we will only turn up here if type_desc->compound_length is 0.
     * If it isn't, a method table for the subtype would have been used.
     */
    for (count=0; count<n_els; count++, to_addr+=size) {
   to_addr1 = to_addr;
   /*
    * Determine length if not fixed, retrieved from stream.
    */
   if (acptIsVVector(type_detail) || acptIsString(type_detail)) {
       if (acptVectorIsShort(type_detail)) {
      ACPT_CHECKF(acptCheckStreamLeft(stream, 1), result);
      length = (Int32)(*stream->buffer++);
       } else ACPT_CHECKF(acptReadUInt32(stream, (UInt32 *)&length, 1), result);

       if (acptIsBoundedString(type_detail)) {
      ACPT_CHECK(length <= length_detail, ACPT_EXCP_STRING_TOO_LONG, result);
       } else {
      if (!acptIsString(type_detail))
        *(Int32 *)(to_addr1+length_detail) = length;

      /*
       * If this is a variable array, allocate the space for it
       * unless it's 0 length.  Note that a 0 length string will
       * always come through as 1 long (for the trailing nul).
       */
      if (length) {
          ACPT_CHECKF(acptAllocDataSpace(length*(is_vec_ref ?
                        ACPT_PTR_SIZE :
                        sub_el_size),
                     BuffEl, &vec_addr), result);
          *(Buffer *)(to_addr1+type_desc->data_offset) = vec_addr;
          to_addr1 = vec_addr;
      } else {
          *(Buffer *)(to_addr1+type_desc->data_offset) = NULL;
          continue;
      }
       }
   }

   if (!is_vec_ref) {
       ACPT_CHECKF(acptDoRead(subtype, stream, to_addr1, length), result);
   } else {
       /*
        * Now read each item in turn, making space for each item of a vec_ref
        */
       for (err_i=i=0, saved_addr=to_addr1; i<length; i++) {
      ACPT_CHECKF(acptAllocDataSpace(sub_el_size, BuffEl, &pos), result);
      *(Buffer *)to_addr1 = pos;
      to_addr1 += ACPT_PTR_SIZE;
      err_i++;

      ACPT_CHECKF(acptDoRead(subtype, stream, pos, 1), result);
       }
   }
    }
    return ACPT_OK;

 errors:
    /*
     * Clean up: note that we can't rely on all memory being initialised, so
     * just undo what was done.
     */

    if (saved_addr)
      for(i=0, to_addr1=saved_addr; i<err_i; i++) {
     pos = *(Buffer *)to_addr1;

     if (i < (err_i-1)) acptDoFree(subtype, pos, 1);

     acptFreeDataSpace(pos);
     to_addr1 += ACPT_PTR_SIZE;
      }

    if (vec_addr) acptFreeDataSpace(vec_addr);

    /* Now free previous ones we've created. */
    for (to_addr-=size; count; count--) acptDoFree(type_desc, to_addr, 1);

    return result;
}

static AcptResult acptWriteVec(AcptTypeDesc *type_desc, Buffer from_addr,
                AcptStream *stream, Int32 n_els)
    /*
     * Transfer the datum at FROM_ADDR, of type defined by TYPE_DESC, into
     * the stream STREAM.
     */
{
    AcptResult result;
    Int32 i;
    Int32 type_detail = type_desc->type_detail;
    Int32 length_detail = type_desc->data_detail;
    Bool is_vec_ref = acptIsVecRef(type_detail);
    Int32 size = type_desc->size;
    AcptTypeDesc *subtype = type_desc->el_subtype;
    Int32 length = length_detail, sub_el_size = subtype->size;
    Buffer pos, from_addr1;

    /*
     * NOTE: we will only turn up here if type_desc->compound_length is 0.
     * If it isn't, a method table for the subtype would have been used.
     */
    for (;n_els; n_els--, from_addr+=size) {
   from_addr1 = from_addr;

   if (acptIsVVector(type_detail) || acptIsString(type_detail)) {
       if (acptIsString(type_detail)) {
      if (acptIsFixedVector(type_detail)) {
          (*subtype->MT->stringLength)(from_addr, &length, length_detail);
          if (length > length_detail)
            return acptExcp(ACPT_EXCP_STRING_TOO_LONG);
      } else {
          if (!*(Buffer *)from_addr) return acptExcp(ACPT_EXCP_NULL_ADDR);

          (*subtype->MT->stringLength)(*(Buffer *)from_addr, &length, 0);
      }
       } else length = *(Int32 *)(from_addr+length_detail);

       if (acptVectorIsShort(type_detail)) {
      if (length > 255) return acptExcp(ACPT_EXCP_LENGTH_TOO_LONG);
      acptMakeStreamSpaceFor(stream, 1);
      *stream->buffer++ = length;
       } else acptWriteUInt32(stream, (UInt32 *)&length, 1);

       if (length == 0) continue;
   }

   /*
    * Now write the data itself.
    */
   if (acptIsVVector(type_detail))
     if (!(from_addr1 = *(Buffer *)(from_addr1+type_desc->data_offset)))
       return acptExcp(ACPT_EXCP_NULL_ADDR);

   if (!is_vec_ref)  {
       ACPT_CHECKR(acptDoWrite(subtype, from_addr1, stream, length), result);
   } else {
       /*
        * Now write each datum in turn.
        */
       for (i=0; i<length; i++) {
      if (!(pos = *(Buffer *)from_addr1)) return acptExcp(ACPT_EXCP_NULL_ADDR);
      from_addr1 += ACPT_PTR_SIZE;

      ACPT_CHECKR(acptDoWrite(subtype, pos, stream, 1), result);
       }
   }
    }

    return ACPT_OK;
}


static AcptResult acptFreeVec(AcptTypeDesc *type_desc, Buffer from_addr, Int32 n_els)
    /*
     * Free any malloc'd data of the type define by TYPE_DESC at FROM_ADDR.
     */
{
    AcptResult result;
    Int32 i;
    Int32 type_detail = type_desc->type_detail;
    Int32 length_detail = type_desc->data_detail;
    Bool is_vec_ref = acptIsVecRef(type_detail);
    Int32 size = type_desc->size;
    AcptTypeDesc *subtype = type_desc->el_subtype;
    Int32 length = length_detail, sub_el_size = subtype->size;
    Buffer vec_addr, pos, from_addr1;

    /*
     * NOTE: we will only turn up here if type_desc->compound_length is 0.
     * If it isn't, a method table for the subtype would have been used.
     */
    for (;n_els; n_els--, from_addr+=size) {
   from_addr1 = from_addr;

   if (acptIsVVector(type_detail) || acptIsString(type_detail)) {
       if (!acptIsString(type_detail))
         length = *(Int32 *)(from_addr1+length_detail);
       else length = 1; /* ie no longer 0 */
   }

   if (acptIsVVector(type_detail)) {
       vec_addr = *(Buffer *)(from_addr1+type_desc->data_offset);
       from_addr1 = vec_addr;
   }

   /*
    * Now free the data itself. For simple cases, just call the raw function,
    * otherwise go round an element at a time, doing each datum as we go.
    */

   if (from_addr1) {
       if (!is_vec_ref)  {
      ACPT_CHECKR(acptDoFree(subtype, from_addr1, length), result);
       } else {
      for (i=0; i<length; i++) {
          pos = *(Buffer *)from_addr1;

          ACPT_CHECKR(acptDoFree(subtype, pos, 1), result);

          acptFreeDataSpace(pos);
          from_addr1 += ACPT_PTR_SIZE;
      }
       }

       if (acptIsVVector(type_detail)) acptFreeDataSpace(vec_addr);
   }
    }

    return ACPT_OK;
}


/*
 *    S T R U C T   F I E L D S
 *    _________________________
 */

static AcptResult acptReadStruct(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
             Int32 n_els)
    /*
     * Move N_ELS structs from the STREAM to the struct(s) at TO_ADDR,
     * using the details in TYPE.
     */
{
    AcptResult result;
    AcptTypeField *field;
    AcptTypeDesc *subtype;
    Int32 i, j, n_fields = type->n_fields;
    Int32 size = acptIsVector(type->type_detail) ? type->el_subtype->size : type->size;
    UInt32 flags;
    Buffer addr, to_addr_start = to_addr; /* for error cleanup */

    n_els *= type->compound_length;

    /*
     * here, we cope  with the UNSET flag, and then do each datum itself.
     */
    for(j=0; j < n_els; j++, to_addr+=size) {
   for (i=0; i<n_fields; i++) {
       field = &type->fields[i];
       addr = to_addr+field->offset;
       flags = field->flags;
       subtype = field->subtype;

       if (flags & ACPT_UNSET) {
      ACPT_CHECKF(acptCheckStreamLeft(stream, 1), result);
      if ((*stream->buffer++) & ACPT_UNSET_FLAG) {
          *(Int32 *)(to_addr+field->extra_info) = ACPT_UNSET_FLAG;
          memset(addr, 0, field->inline_size); /* always leave clean */
          if (acptIsVVector(subtype->type_detail) &&
         !acptIsString(subtype->type_detail))
            *(Int32 *)(addr+subtype->data_detail) = 0;
          continue;

      } else *(Int32 *)(to_addr+field->extra_info) = 0;
       }

       ACPT_CHECKF(acptDoRead(subtype, stream, addr, 1), result);
   }
    }

    return ACPT_OK;
 errors:
    /*
     * We failed in the j'th struct at the i'th field, so clean this up manually,
     * then use FreeStruct on the rest
     */
    while(--i >= 0) {
   field = &type->fields[i];
   addr = to_addr+field->offset;
   flags = field->flags;
   subtype = field->subtype;

   acptDoFree(subtype, addr, 1);
    }

    acptFreeStruct(type, to_addr_start, j);

    return result;
}

static AcptResult acptWriteStruct(AcptTypeDesc *type, Buffer from_addr,
              AcptStream *stream, Int32 n_els)
    /*
     * Move N_ELS fields of the struct(s) at FROM_ADDR to STREAM
     * using the details in TYPE.
     */
{
    AcptResult result;
    AcptTypeDesc *subtype;
    AcptTypeField *field;
    Int32 i, n_fields = type->n_fields;
    Int32 size = acptIsVector(type->type_detail) ? type->el_subtype->size : type->size;
    UInt32 flags;
    Buffer addr;

    n_els *= type->compound_length;

    /*
     * All we do is to copy each field for each element
     */
    for(;n_els; n_els--, from_addr+=size) {
   for (i=0; i<n_fields; i++) {
       field = &type->fields[i];
       addr = from_addr+field->offset;
       flags = field->flags;
       subtype = field->subtype;

       if (flags & ACPT_UNSET) {
      acptMakeStreamSpaceFor(stream, 1);

      if (*(Int32 *)(from_addr+field->extra_info) & ACPT_UNSET_FLAG) {
          *stream->buffer++ = ACPT_UNSET_FLAG;
          continue;
      } else *stream->buffer++ = 0;
       }

       ACPT_CHECKR(acptDoWrite(subtype, addr, stream, 1), result);
   }
    }

    return ACPT_OK;
}

static AcptResult acptFreeStruct(AcptTypeDesc *type, Buffer from_addr, Int32 n_els)
    /*
     * Free anything allocated in the N_ELS structs of type TYPE
     * starting at FROM_ADDR.
     */
{
    AcptTypeDesc *subtype;
    AcptTypeField *field;
    Int32 i, n_fields = type->n_fields;
    Int32 size = acptIsVector(type->type_detail) ? type->el_subtype->size : type->size;
    UInt32 flags;
    Buffer addr;

    n_els *= type->compound_length;

    for(;n_els; n_els--, from_addr+=size) {
   for (i=0; i<n_fields; i++) {
       field = &type->fields[i];
       addr = from_addr+field->offset;
       flags = field->flags;
       subtype = field->subtype;

       if ((flags & ACPT_UNSET) &&
      *(Int32 *)(from_addr+field->extra_info) & ACPT_UNSET_FLAG) continue;

       acptDoFree(subtype, addr, 1);
   }
    }

    return ACPT_OK;
}


/*
 *       U N I O N   F I E L D S
 *       _______________________
 */

static AcptResult acptReadUnion(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
            Int32 n_els)
    /*
     * Move N_ELS structs from STREAM to the union(s) at TO_ADDR,
     * using the details in TYPE.
     */
{
    AcptResult result;
    AcptTypeField *field;
    AcptTypeDesc *subtype;
    Int32 i, j, n_fields = type->n_fields, size = type->size;
    Int32 discrim_offset = type->data_detail, data_offset = type->data_offset;
    Int32 discrim_value;
    Buffer to_addr_start = to_addr; /* for error cleanup */

    n_els *= type->compound_length;

    for(j=0; j < n_els; j++, to_addr+=size) {

   ACPT_CHECKF(acptReadUInt32(stream, (UInt32 *)&discrim_value, 1), result);
   *(Int32 *)(to_addr+discrim_offset) = discrim_value;

   for (i=0; i<n_fields; i++)
     if ((field = &type->fields[i])->extra_info == discrim_value) break;

   /* Note that we should never get a discrim mismatch, as types match! */
   ACPT_CHECK(i != n_fields, ACPT_EXCP_BAD_DISCRIM_ON_WRITE, result);

   subtype = field->subtype;
   ACPT_CHECKF(acptDoRead(subtype, stream, to_addr+data_offset, 1), result);
    }

    return ACPT_OK;
 errors:
    /*
     * We failed in the j'th union, so clean up before that
     */
    acptFreeUnion(type, to_addr_start, j);

    return result;
}

static AcptResult acptWriteUnion(AcptTypeDesc *type,
             Buffer from_addr, AcptStream *stream,
             Int32 n_els)
    /*
     * Move N_ELS fields of the union(s) at FROM_ADDR to STREAM using
     * the details in TYPE.
     */
{
    AcptResult result;
    AcptTypeField *field;
    AcptTypeDesc *subtype;
    Int32 i, n_fields = type->n_fields, size = type->size;
    Int32 discrim_offset = type->data_detail, data_offset = type->data_offset;
    Int32 discrim_value;

    n_els *= type->compound_length;

    for(;n_els; n_els--, from_addr+=size) {
   discrim_value = *(Int32 *)(from_addr+discrim_offset);

   for (i=0; i<n_fields; i++)
     if ((field = &type->fields[i])->extra_info == discrim_value) break;

   if (i == n_fields) return acptExcp(ACPT_EXCP_BAD_DISCRIM); /* unknown value */

   ACPT_CHECKR(acptWriteUInt32(stream, (UInt32 *)&discrim_value, 1), result);

   subtype = field->subtype;
   ACPT_CHECKR(acptDoWrite(subtype, from_addr+data_offset, stream, 1), result);
    }

    return ACPT_OK;
}


static AcptResult acptFreeUnion(AcptTypeDesc *type, Buffer from_addr, Int32 n_els)
    /*
     * Free anything allocated in the N_ELS unions starting at FROM_ADDR.
     */
{
    AcptTypeField *field;
    AcptTypeDesc *subtype;
    Int32 i, n_fields = type->n_fields, size = type->size;
    Int32 discrim_offset = type->data_detail, data_offset = type->data_offset;
    Int32 discrim_value;

    n_els *= type->compound_length;

    for(;n_els; n_els--, from_addr+=size) {
   discrim_value = *(Int32 *)(from_addr+discrim_offset);

   for (i=0; i<n_fields; i++)
     if ((field = &type->fields[i])->extra_info == discrim_value) break;

   if (i == n_fields) continue; /* if fallen off end, don't know what type to do */

   subtype = field->subtype;
   acptDoFree(subtype, from_addr+data_offset, 1);
    }

    return ACPT_OK;
}

/*
 *       R A W   D A T A   C O P I E R S
 *       _______________________________
 */

/*
 * Note: End-swapping is always done by the reader - the writer always sends
 * native endianness data. The technique for endswapping is to copy the data
 * into the user's space, then swap in-situ, as his space should be correctly
 * aligned and therefore optimal transfer can be used.
 */

xstatic AcptResult acptWriteScalars(AcptTypeDesc *type, Buffer from_addr,
                AcptStream* stream, Int32 n)
    /*
     * Write scalars from C address FROM_ADDR into STREAM.
     */
{
    AcptResult result;

    n *= type->compound_length*type->stream_size;

    ACPT_CHECKR(acptMakeStreamSpaceFor(stream, n), result);
    memcpy(stream->buffer, from_addr, n);
    stream->buffer += n;

    return ACPT_OK;
}

xstatic AcptResult acptReadScalars(AcptTypeDesc *type, AcptStream *stream,
               Buffer to_addr, Int32 n)
    /*
     * Read N elements of TYPE from STREAM into TO_ADDR, byte-swapping
     * if necessary.
     */
{
    AcptResult result;
    Int32 inc;

    n *= type->compound_length;
    inc = n*type->stream_size;

    ACPT_CHECKR(acptCheckStreamLeft(stream, inc), result);
    memcpy(to_addr, stream->buffer, inc);
    stream->buffer += inc;

    if (stream->byte_swapped && type->stream_size > 1) {
   switch(type->stream_size) {
   case 2: {
       unsigned short *to = (unsigned short *)to_addr;
       while (n--) {
      unsigned short v = *to;
      *to++ = v << 8 | v >> 8;
       }
       break;
   }

   case 4: {
       UInt32 *to = (UInt32 *)to_addr;
       while (n--) {
      UInt32 v = *to;
      v = v << 16 | v >> 16;
      *to++ = (v & 0x00ff00ffU) << 8 | (v >> 8) & 0x00ff00ffU;
       }
       break;
   }

   case 8: {
       UInt32 *to = (UInt32 *)to_addr;
       while (n--) {
      UInt32 v0, v1;

      v0 = to[0];    /* allow for delayed load/store */
      v1 = to[1];
      v0 = v0 << 16 | v0 >> 16;
      to[1] = (v0 & 0x00ff00ffU) << 8 | (v0 >> 8) & 0x00ff00ffU;
      v1 = v1 << 16 | v1 >> 16;
      to[0] = (v1 & 0x00ff00ffU) << 8 | (v1 >> 8) & 0x00ff00ffU;
      to += 2;
       }
       break;
   }
   }
    }
    return ACPT_OK;
}

xstatic AcptResult acptWriteBigInt(AcptTypeDesc *type, Buffer from_addr,
               AcptStream* stream, Int32 n)
    /*
     * Write N BigInts from C address FROM_ADDR into STREAM. FROM_ADDR
     * contains the address of the byte vector.
     */
{
    AcptResult result;
    unsigned char count;

    n *= type->compound_length;

    while(n--) {
       unsigned char *bigint = *(unsigned char **)from_addr;
       count = *bigint;

       if (count == 0 || count > 255) return acptExcp(ACPT_EXCP_VALUE_OUT_OF_RANGE);

       ACPT_CHECKR(acptMakeStreamSpaceFor(stream, count+1), result);
       *stream->buffer++ = count;

       memcpy(stream->buffer, bigint+1, count);
       stream->buffer += count;

       from_addr += ACPT_PTR_SIZE;
    }

    return ACPT_OK;
}

xstatic AcptResult acptReadBigInt(AcptTypeDesc *type, AcptStream *stream,
              Buffer to_addr, Int32 n)
    /*
     * Read N elements of TYPE from STREAM into TO_ADDR, byte-swapping
     * if necessary.
     */
{
    AcptResult result;
    Int32 i;
    unsigned char *buffer;
    unsigned char count;

    n *= type->compound_length;

    for (i=0; i<n; i++) {
   ACPT_CHECKF(acptCheckStreamLeft(stream, 1), result);
   count = *stream->buffer++;

   ACPT_CHECKF(acptCheckStreamLeft(stream, count), result);
   ACPT_CHECKF(acptAllocDataSpace(count+1, unsigned char, &buffer), result);

   *(unsigned char **)to_addr = buffer;
   *buffer = count;

   memcpy(buffer+1, stream->buffer, count);

   stream->buffer += count;
   to_addr += ACPT_PTR_SIZE;
    }

    return ACPT_OK;

  errors:
    while(i--) {
   to_addr -= ACPT_PTR_SIZE;
   acptFreeDataSpace(*(unsigned char **)to_addr);
    }

    return result;
}

xstatic AcptResult acptFreeBigInt(AcptTypeDesc *type, Buffer from_addr, Int32 n_els)
    /*
     * Free N_ELS BigInts starting from FROM_ADDR
     */
{
    n_els *= type->compound_length;

    while(n_els--) {
   acptFreeDataSpace(*(unsigned char **)from_addr);
   from_addr += ACPT_PTR_SIZE;
    }

    return ACPT_OK;
}

xstatic AcptResult acptWrite1_2(AcptTypeDesc *type, Buffer from_addr,
                AcptStream* stream, Int32 n)
    /*
     * Write N elements of TYPE (which must be Char26) to STREAM from TO_ADDR,
     * converting to Char8.
     */
{
    AcptResult result;
    unsigned short *from = (unsigned short *)from_addr;
    Buffer p, pe;

    n *= type->compound_length;

    ACPT_CHECKR(acptMakeStreamSpaceFor(stream, n), result);

    for(p=stream->buffer, pe=stream->buffer+n; p<pe; p++, from++) {
   if (*from > 255) return acptExcp(ACPT_EXCP_VALUE_OUT_OF_RANGE);
   *p = (unsigned char)(*from);
    }

    stream->buffer += n;

    return ACPT_OK;
}


xstatic AcptResult acptRead1_2(AcptTypeDesc *type, AcptStream *stream,
               Buffer to_addr, Int32 n)
    /*
     * Read N elements of TYPE (which must be Char8) from STREAM into TO_ADDR,
     * converting to Char16.
     */
{
    AcptResult result;
    unsigned short *to = (unsigned short *)to_addr;
    Buffer p, pe;

    n *= type->compound_length;

    ACPT_CHECKR(acptCheckStreamLeft(stream, n), result);

    for(p=stream->buffer, pe=stream->buffer+n; p<pe; to++, p++) *to = *p;

    stream->buffer += n;

    return ACPT_OK;
}


xstatic AcptResult acptStringLength1(Buffer buffer, Int32 *length,
                 Int32 max_length)
    /*
     * Returns the length of the string *including* trailing nul. If MAX_LENGTH
     * isn't 0, then only this number of characters are checked. The actual result
     * will lie between 1 and MAX_LENGTH+1. In the latter case, the string is
     * actually too long (except in ds cases)
     */
{
    char *p = (char *)buffer;

    if (!max_length) for(;*p; p++);
    else {
   for (;max_length && *p; max_length--, p++);
    }

    *length = (p-(char *)buffer)+1;

    return ACPT_OK;
}

xstatic AcptResult acptStringLength2(Buffer buffer, Int32 *length,
                 Int32 max_length)
    /*
     * Returns the length of the string *including* trailing nul.
     */
{
    unsigned short *p = (unsigned short *)buffer;

    if (!max_length) for(;*p; p++);
    else {
   for (;max_length && *p; max_length--, p++);
    }

    *length = (p-(unsigned short *)buffer)+1;
    return ACPT_OK;
}

static AcptResult acptReadKleene(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
             Int32 n)
    /*
     * Read N 1 bye elements from STREAM and convert them to 4 byte values.
     */
{
    AcptResult result;
    UInt32 *t;
    Buffer p, e;

    n *= type->compound_length;

    ACPT_CHECKR(acptCheckStreamLeft(stream, n), result);
    for (t=(UInt32 *)to_addr, p=stream->buffer, e=p+n; p < e;) *t++ = *p++;
    stream->buffer += n;

    return ACPT_OK;
}


static AcptResult acptWriteKleene(AcptTypeDesc *type, Buffer from_addr,
              AcptStream *stream, Int32 n)
    /*
     * Read N 4 byte values at from_addr as 1 byte values to stream.
     */
{
    AcptResult result;
    UInt32 *f;
    Buffer p, e;

    n *= type->compound_length;

    ACPT_CHECKR(acptMakeStreamSpaceFor(stream, n), result);
    for (f=(UInt32 *)from_addr, p=stream->buffer, e=p+n; p < e;) *p++ = *f++;
    stream->buffer += n;

    return ACPT_OK;
}


static AcptResult acptReadDI(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
              Int32 n)
    /*
     * Read N Int32-aligned elements of TYPE from STREAM into and
     * store them as doubles at TO_ADDR byte-swapping if necessary.
     */
{
    AcptResult result;
    double *dp;
    Int32 *i;

    n *= type->compound_length;
    dp = (double *)to_addr+n;
    i = (Int32 *)to_addr+n;

    /*
     * As we have twice as much space as we need, read the Ints into the user's
     * space, then run backwards converting each element to a double.
     */
    ACPT_CHECKR(acptReadUInt32(stream, (UInt32 *)to_addr, n), result);

    while(dp > (double *)to_addr) *(--dp) = (double)*(--i);

    return ACPT_OK;
}


static AcptResult acptWriteDI(AcptTypeDesc *type, Buffer from_addr, AcptStream *stream,
               Int32 n)
    /*
     * Write N double-as-ints from C address FROM_ADDR into STREAM, converting
     * them to Int32s as they are written. Note that no range checking is
     * performed.
     */
{
    AcptResult result;
    double *dp=(double *)from_addr;
    Int32 ibuff[32], *i=ibuff, inc;

    if (type) n *= type->compound_length; /* type omitted once recursing */

    /*
     * As the stream buffer isn't necc. byte-aligned, work in a
     * buffer. However, rather than allocate a new buffer each time,
     * use a fixed size one and do one level recursion to shovel all
     * but the last bucket... then do that here.
     */
    for (;n>32;n-=32,dp+=32)
      ACPT_CHECKR(acptWriteDI(NULL, (Buffer)dp, stream, 32), result);

    inc = n*4;
    while(n--) *i++ = (Int32)(*dp++);

    ACPT_CHECKR(acptMakeStreamSpaceFor(stream, inc), result);
    memcpy(stream->buffer, ibuff, inc);
    stream->buffer += inc;

    return ACPT_OK;
}

static AcptResult acptByteSwap4InSitu(UInt32 *values, Int32 n)
    /*
     * Byteswap the N 4-byte values in VALUES, in-situ.
     */
{
    while(n--) {
   UInt32 v = *values;
   v = v << 16 | v >> 16;
   *values++ = (v & 0x00ff00ffU) << 8 | (v >> 8) & 0x00ff00ffU;
    }
    return ACPT_OK;
}


/*
 *          S T R E A M S
 *          _____________
 */

AcptResult ACPT_DLL_CALL
acptNewStream(void *handle,
         AcptResult (*Transfer)(void *handle, unsigned char *buffer,
                 Int32 n_bytes),
         Bool reader, AcptStream **streamp)
    /*
     * Create a new acpt stream handle, and return the handle via STREAMP. HANDLE,
     * is opaque data which is handed to the transfer function TRANSER when moving
     * data. READER should be TRUE if this is a read (input) stream, and FALSE if
     * it is a write (output) stream. The behavoir of the TRANSFER function should
     * correspond - ie be a reader if READER is TRUE, otherwise be a writer. The
     * TRANSFER function should simply read or write the N_BYTES in BUFFER.
     */
{
    AcptResult result=ACPT_OK;
    Int32 i;

    ACPT_CHECKR(acptAllocType(1, AcptStream, streamp), result);
    (*streamp)->handle = handle;
    (*streamp)->Transfer = Transfer;
    (*streamp)->reader = reader;
    (*streamp)->buffer_abs_max = 0; /* implies no limit */
#if defined(_ACPTM_C_ONLY_)
    acptmSetupStream(*streamp);
#endif
    if (reader) return ACPT_OK;

    /*
     * Register as a write stream, so we can clean up the sent types when
     * types are freed.
     */
    START_EXCLUSIVE {

   for (i=0; i<writeStreamsMax && writeStreams[i]; i++);
   if (i==writeStreamsMax) {
       if ((result = acptRealloc(writeStreamsMax+1, AcptStream *,
                  &writeStreams)) != ACPT_OK) {
      acptFree(*streamp);
      goto end;
       }
       writeStreamsMax++;
   }

   writeStreams[i] = *streamp;

    end:;
    } END_EXCLUSIVE;

    return result;
}

AcptResult ACPT_DLL_CALL acptFreeStream(AcptStream *stream)
    /*
     * Free any resources associated with acpt stream STREAM.
     */
{
    Int32 i;
    AcptTypeDesc *type_desc;

    /*
     * Free the local stream types
     */
    for (i=0; i < stream->typeDescListMax; i++)
      if (type_desc=stream->typeDescs[i]) {
     acptFree(type_desc->fields);
     acptFree(type_desc);
      }

    acptFree(stream->buffer_start);
    acptFree(stream->typeDescs);
    acptFree(stream->SentTypes);
    acptFree(stream->SentTypesThisGo);
    acptFree(stream->Errors.stack);

#if defined(_ACPTM_C_ONLY_)
    acptmFreeStream(stream);
#endif

    if (!stream->reader) {
   START_EXCLUSIVE {

       for (i=0; i<writeStreamsMax && writeStreams[i] != stream; i++);
       writeStreams[i] = NULL;

   } END_EXCLUSIVE;
    }

    acptFree(stream);

    return ACPT_OK;
}

AcptResult ACPT_DLL_CALL acptSetStreamLimit(AcptStream *stream, Int32 limit)
    /*
     * Limits the buffer size for STREAM to LIMIT. This is only used for
     * writers: a read streams will always absorb whatever data it is
     * given and ignore LIMIT. LIMIT=0 is used to indicate unlimited.
     */
{
    stream->buffer_abs_max = limit;

    return ACPT_OK;
}

/*
 *       S T R E A M   H E L P E R S
 *       ___________________________
 */

xstatic AcptResult acptMakeStreamSpaceForHelper(AcptStream *stream, Int32 size)
    /*
     * Make space in the stream buffer for SIZE bytes.
     */
{
    AcptResult result;
    Int32 diff = stream->buffer-stream->buffer_start, new_size;
    Bool buffer_bounded = stream->buffer_abs_max != 0;

    size += diff;
    if (size <= stream->buffer_max) return ACPT_OK;
    if (buffer_bounded && size > stream->buffer_abs_max)
      return acptExcp(ACPT_EXCP_MAX_BUFSIZ_EXCEEDED);

    new_size = size+ACPT_STREAM_BUFFER_INCR; /* give some slop */

    if (buffer_bounded && new_size > stream->buffer_abs_max)
      new_size = stream->buffer_abs_max;

    ACPT_CHECKR(acptRealloc(new_size, BuffEl, &stream->buffer_start), result);

    stream->buffer_max = new_size;
    stream->buffer = stream->buffer_start+diff;

    return ACPT_OK;
}

xstatic AcptResult acptSendBuffer(AcptStream *stream)
    /*
     * Actually transfer the current data to the stream.
     */
{
    AcptResult result;
    UInt32 total_size, count;
    Buffer start;
    /*
     * the code here is much more complicated than might be expected, because of
     * the way in which Magik interacts. Normally each Transfer actually transfers
     * data to the stream, but in the Magik case the data is simply left sitting
     * in the buffer and extracted manually at the end. This is complicated in the
     * case of WriteObjects, because the count must be inserted into the right
     * place in the stream.
     */
    if (stream->started) {
   start = stream->buffer_start+stream->buffer_count_offset;
   total_size = stream->buffer-start;
   count = total_size-ACPT_INT_SIZE;
    } else {
   ACPT_CHECKR(acptSendIntro(stream), result);
   start = stream->buffer_start;
   total_size = stream->buffer-start;
   count = total_size - (ACPT_INT_SIZE+stream->buffer_count_offset);
    }

    memcpy(stream->buffer_start+stream->buffer_count_offset, (Buffer)&count,
      ACPT_INT_SIZE);

    ACPT_CHECKR((*stream->Transfer)(stream->handle, start, total_size), result);

    /*
     * prepare for next data, in case we are doing WriteObjects.
     */
    stream->buffer_count_offset = stream->buffer-stream->buffer_start;
    stream->buffer += ACPT_INT_SIZE;

    stream->started = TRUE;   /* ie wait till we know everything's worked */

    return ACPT_OK;
}

xstatic AcptResult acptReceiveBuffer(AcptStream *stream)
    /*
     * Retrieve the next glob of data on the stream.
     */
{
    AcptResult result;
    UInt32 n_bytes;

    stream->buffer = stream->buffer_start; /* clear everything out */
    stream->buffer_count = 0;

    if (!stream->started) ACPT_CHECKR(acptReceiveIntro(stream), result);

    ACPT_CHECKR(acptMakeStreamSpaceFor(stream, ACPT_INT_SIZE), result);

    ACPT_CHECKR((*stream->Transfer)(stream->handle, stream->buffer,
                ACPT_INT_SIZE), result);
    stream->buffer_count += ACPT_INT_SIZE;

    ACPT_CHECKR(acptReadUInt32(stream, &n_bytes, 1), result);

    ACPT_CHECKR(acptMakeStreamSpaceFor(stream, n_bytes), result);

    ACPT_CHECKR((*stream->Transfer)(stream->handle, stream->buffer, n_bytes), result);

    stream->buffer_count += n_bytes;
    return ACPT_OK;
}

xstatic AcptResult acptReadUInt32(AcptStream *stream, UInt32 *buffer, Int32 n)
    /*
     * Read N values from STREAM, byte-swapping if necessary.
     */
{
    AcptResult result;
    UInt32 *to = (UInt32 *)buffer;
    Int32 inc;

    inc = n*4;

    ACPT_CHECKR(acptCheckStreamLeft(stream, inc), result);
    memcpy(to, stream->buffer, inc);
    stream->buffer += inc;

    if (stream->byte_swapped) {
   while (n--) {
       UInt32 v = *to;
       v = v << 16 | v >> 16;
       *to++ = (v & 0x00ff00ffU) << 8 | (v >> 8) & 0x00ff00ffU;
   }
    }
    return ACPT_OK;
}

xstatic AcptResult acptWriteUInt32(AcptStream *stream, UInt32 *buffer, Int32 n)
    /*
     * Write N_INTs values to STREAM.
     */
{
    AcptResult result;
    n *= 4;

    ACPT_CHECKR(acptMakeStreamSpaceFor(stream, n), result);
    memcpy(stream->buffer, buffer, n);
    stream->buffer += n;

    return ACPT_OK;
}



/*
 *    A D V A N C E D   E R R O R   H A N D L I N G
 *    _____________________________________________
 */
static AcptResult acptPrepareForError(AcptStream *stream)
    /*
     * Called when an error has occurred which requires a recursive traceback,
     * prepares the variable sized vector for holding the data.
     */
{
    AcptResult result;

    if (stream->Errors.depth > stream->Errors.max_depth) {
   ACPT_CHECKF(acptRealloc(stream->Errors.depth, AcptStreamErrorDetail,
            &stream->Errors.stack), result);

   stream->Errors.max_depth = stream->Errors.depth;
    }
    stream->Errors.sp = stream->Errors.stack+stream->Errors.depth;

    return stream->Errors.code;

 errors:
    stream->Errors.code = result;
    return result;
}

static AcptResult acptProcessError(AcptStream *stream, Int32 i,
               AcptTypeDesc *s_type, AcptResult result)
    /*
     * Called after a recursive error has occured, populates the next element
     * up the error traceback stack with where we've been.
     */
{
    if (result != ACPT_EXCP_MEMORY_EXHAUSTED) {
   stream->Errors.sp--;

   stream->Errors.sp->el = i;
   stream->Errors.sp->flavour = acptInterestingFlags(s_type->type_detail);
    }

    return result;
}


char *flavourNames[64];
char *scalarNames[ACPT_MAX_BUILTIN_TYPES];

static AcptResult acptInitFlavourNames()
    /*
     * Initialise global variable flavourNames to sensible names for
     * flavours, to help with debugging.
     */
{
    Int32 i;

    for (i=0; i<ARRAY_LEN(flavourNames); i++) flavourNames[i] = "*unknown*";

    flavourNames[ACPT_SCALAR] = "scalar";
    flavourNames[ACPT_FIXED_VECTOR] = "fixed_vector";
    flavourNames[ACPT_SHORT_V_VECTOR] = "short_v_vector";
    flavourNames[ACPT_LONG_V_VECTOR] = "long_v_vector";
    flavourNames[ACPT_SHORT_V_STRING] = "short_string";
    flavourNames[ACPT_LONG_V_STRING] = "long_string";
    flavourNames[ACPT_SHORT_BOUNDED_STRING] = "short_bounded_string";
    flavourNames[ACPT_LONG_BOUNDED_STRING] = "long_bounded_string";
    flavourNames[ACPT_STRUCT] = "struct";
    flavourNames[ACPT_UNION] = "union";

    return ACPT_OK;
}

static AcptResult acptInitScalarNames()
    /*
     * Initialise global variable scalarNames to sensible names for
     * scalar types, to help with debugging.
     */
{
    Int32 i;

    for (i=0; i<ARRAY_LEN(scalarNames); i++) scalarNames[i] = "*unknown*";

    scalarNames[ACPT_TYPE_UNSET] = "unset";

    scalarNames[ACPT_TYPE_UINT8] = "uint8";
    scalarNames[ACPT_TYPE_UINT16] = "uint16";
    scalarNames[ACPT_TYPE_UINT32] = "uint32";
    scalarNames[ACPT_TYPE_UINT64] = "uint64";

    scalarNames[ACPT_TYPE_INT8] = "int8";
    scalarNames[ACPT_TYPE_INT16] = "int16";
    scalarNames[ACPT_TYPE_INT32] = "int32";
    scalarNames[ACPT_TYPE_INT64] = "int64";

    scalarNames[ACPT_TYPE_FLOAT32] = "float32";
    scalarNames[ACPT_TYPE_FLOAT64] = "float64";

    scalarNames[ACPT_TYPE_CHAR8] = "char8";
    scalarNames[ACPT_TYPE_CHAR16] = "char16";

    scalarNames[ACPT_TYPE_BOOL] = "bool";
    scalarNames[ACPT_TYPE_KLEENE] = "kleene";

    scalarNames[ACPT_TYPE_SIMPLE_TIME] = "simple_time";
    scalarNames[ACPT_TYPE_DATE] = "date";
    scalarNames[ACPT_TYPE_DATE_TIME] = "date_time";

    scalarNames[ACPT_TYPE_DOUBLE_AS_INT] = "double_as_int";

    scalarNames[ACPT_TYPE_SHORT_STRING8] = "short_string8";
    scalarNames[ACPT_TYPE_SHORT_STRING16] = "short_string16";
    scalarNames[ACPT_TYPE_LONG_STRING8] = "long_string8";
    scalarNames[ACPT_TYPE_LONG_STRING16] = "long_string16";

    scalarNames[ACPT_TYPE_CHAR8_TO_16] = "char8_to_16";

    scalarNames[ACPT_TYPE_EMPTY_VECTOR] = "empty_vector";

    scalarNames[ACPT_TYPE_SHORT_STRING8_TO_16] = "short_string8_to_16";
    scalarNames[ACPT_TYPE_LONG_STRING8_TO_16] = "long_string8_to_16";

    scalarNames[ACPT_TYPE_COORD] = "coord";
    scalarNames[ACPT_TYPE_COORD3D] = "coord3d";
    scalarNames[ACPT_TYPE_COORD2H] = "coord2h";

    scalarNames[ACPT_TYPE_SIMPLE_SECTOR] = "simple_sector";
    scalarNames[ACPT_TYPE_SECTOR_Z] = "sector_z";
    scalarNames[ACPT_TYPE_SECTOR_H] = "sector_h";

    scalarNames[ACPT_TYPE_CIRCLE] = "circle";
    scalarNames[ACPT_TYPE_TAN_ARC] = "tan_pt_arc";
    scalarNames[ACPT_TYPE_CP_ARC] = "centre_pt_arc";

    scalarNames[ACPT_TYPE_SECTOR] = "sector";

    scalarNames[ACPT_TYPE_SECTOR_ROPE] = "sector_rope";

    scalarNames[ACPT_TYPE_GIS_ID] = "gis_id";

    return ACPT_OK;
}

void ACPT_DLL_CALL acptPrintStreamMismatchError(AcptStreamError *error)
    /*
     * Handler for stream errors which writes the error information to
     * stderr.
     */
{
    static char *title = "\nAcptError: Type Mismatch -";
    char detail_mess[128];
    Int32 i, el;
    UInt32 flavour;

    switch(error->code) {
    case ACPT_EXCP_FLAVOUR_MISMATCH:
   fprintf(stderr, "%s flavours differ.\n", title);
   sprintf(detail_mess, "Stream flavour:    %s\nRequested flavour: %s\n",
      flavourNames[error->s_data], flavourNames[error->c_data]);
   break;

    case ACPT_EXCP_LENGTH_MISMATCH:
   fprintf(stderr, "%s vector array lengths differ.\n", title);
   sprintf(detail_mess, "Stream length:    %d\nRequested length: %d\n",
      error->s_data, error->c_data);
   break;

    case ACPT_EXCP_NFIELDS_MISMATCH:
   fprintf(stderr, "%s struct/union field counts differ.\n", title);
   sprintf(detail_mess,
      "Stream field-count:    %d\nRequested field-count: %d\n",
      error->s_data, error->c_data);
   break;

    case ACPT_EXCP_FIELD_FLAGS_MISMATCH:
   fprintf(stderr, "%s struct field flags differ.\n", title);
   sprintf(detail_mess, "Stream field:    %s\nRequested field: %s\n",
      error->s_data ? "May Be UNSET" : "May *Not* Be UNSET",
      error->c_data ? "May Be UNSET" : "May *Not* Be UNSET");
   break;

    case ACPT_EXCP_DISCRIM_MISMATCH:
   fprintf(stderr, "%s union discriminator value differs.\n", title);
   sprintf(detail_mess, "Stream value:    %d\nRequested value: %d\n",
      error->s_data, error->c_data);
   break;

    case ACPT_EXCP_SCALAR_TYPE_MISMATCH:
   fprintf(stderr, "%s scalar type differs.\n", title);
   sprintf(detail_mess, "Stream type:    %s\nRequested type: %s\n",
      scalarNames[error->s_data], scalarNames[error->c_data]);
   break;

    default:
   return;        /* error is something like memory exhausted */
    }

    /*
     * Now give the path to the mismatch
     */
    for (i=0; i<error->depth; i++) {
   el = error->stack[i].el;
   flavour = error->stack[i].flavour;
   if (el >= 0) {
       if (i < error->depth-1) fprintf(stderr, "via ");
       if (flavour == ACPT_UNION || flavour == ACPT_STRUCT)
         fprintf(stderr, "element %d of %s\n", el, flavourNames[flavour]);
       else
         fprintf(stderr, "subtype of %s\n", flavourNames[flavour]);
   }
    }

    /*
     * And the mismatch itself
     */
    fprintf(stderr, "%s", detail_mess);
}

/**********************************************************************
/**** END acpt.c
/**********************************************************************/

/**********************************************************************
/######################################################################
/#### acpt_tics implementation
/######################################################################
/*********************************************************************/

/**********************************************************************
/**** BEGIN acpt_tics.c
/**********************************************************************/

static AcptResult acpt_tics_write(void *handle, unsigned char *data, Int32 n_bytes)
    /*
     * TICS stream write transfer
     */
{
    /*
     * ticsPutBytes() takes a *UShort* value for NBYTES, so send the data in
     * chunks if necc.
     */
    while(1) {
   if (n_bytes > 65534) {
       ticsPutBytes((TicsHandle) handle, data, 65534);
       data += 65534;
       n_bytes -= 65534;
   } else {
       ticsPutBytes((TicsHandle) handle, data, (UShort)n_bytes);
       break;
   }
    }

    return ACPT_OK;
}

static AcptResult acpt_tics_read(void *handle, unsigned char *data, Int32 n_bytes)
    /*
     * TICS stream read transfer
     */
{
    while(1) {
   if (n_bytes > 65534) {
       ticsGetBytes((TicsHandle) handle, data, 65534);
       data += 65534;
       n_bytes -= 65534;
   } else {
       ticsGetBytes((TicsHandle) handle, data, (UShort)n_bytes);
       break;
   }
    }

    return ACPT_OK;
}

/*
 * Interfaces to make actual acpt Streams (which for convenience are then
 * embedded in the tics handle) and to free them.
 */
AcptResult ACPT_DLL_CALL ticsMakeAcptStreams(TicsHandle thHandle)
{
    AcptResult res;

    if ((res = acptNewStream(thHandle, acpt_tics_read, TRUE,
              (AcptStream **)(&(thHandle)->read_stream)))
   != ACPT_OK) return res;

    return acptNewStream(thHandle, acpt_tics_write, FALSE,
          (AcptStream **)(&(thHandle)->write_stream));
}


AcptResult ACPT_DLL_CALL ticsFreeAcptStreams(TicsHandle thHandle)
{
    AcptResult res;

    if ((res = acptFreeStream(TH_READ(thHandle)) != ACPT_OK)) return res;
    return acptFreeStream(TH_WRITE(thHandle));
}

/**********************************************************************
/**** END acpt_tics.c
/**********************************************************************/

