// VERSION FOR SMALLWORLD 3.3

/*
 * Header file acpt_tics.h
 * Includes interface definitions between TICS and acpt - the data marshalling library.
 */

/**********************************************************************
/######################################################################
/#### tics32 implementation
/######################################################################
/*********************************************************************/

#include <stdio.h>
#include <winsock.h>
// #include <Winsock2.h>

/*********************************************************************
/**** BEGIN tics_int.h
/*********************************************************************/

/*
 * Winsock uses a SOCKET type which is neater than the int used elsewhere,
 * Also has an INVALID_SOCKET, a closesocket function and different error
 * handling.
 */

#ifndef  INVALID_SOCKET

   typedef int     SOCKET;
   #include <errno.h>
   #define INVALID_SOCKET     (SOCKET)(~0)
   #define SOCKET_ERROR    -1
   #define closesocket(s)     close(s)
   #define GetLastSockError() (errno)

#else

   #define GetLastSockError() WSAGetLastError()

#endif

#ifndef MAXHOSTNAMELEN

   #define MAXHOSTNAMELEN     255

#endif

/*
 * Convenient TRUE and FALSE
 */

#ifndef FALSE
   #define  FALSE       0
#endif

#ifndef  TRUE
   #define  TRUE        1
#endif

#define IS_BIG_ENDIAN      0
#define IS_LITTLE_ENDIAN   1
#define IS_PDP_11_INTEGER  2

#define IS_IEEE_FLOAT      0
#define IS_VAX_FLOAT    1

#define IS_UNKNOWN      99

/* Flags used when send out-of-band data */

#define OOB_CONDITION      1      /* User generated ticsCondition */

/* Diagnostic displayers for SW development debugging purposes only */

#if defined (DEBUG) && defined (DIAGNOSTIC_LEVEL)
   #define DEBUG_PRINTF(l,args) if (l>=DIAGNOSTIC_LEVEL)fprintf args
   #define DEBUG_DO(l,expr) if (l>=DIAGNOSTIC_LEVEL)expr
#else
   #define DEBUG_PRINTF(l,args)
   #define DEBUG_DO(l,expr)
#endif

#define HOST_BIG_ENDIAN    (iThisEndian == IS_BIG_ENDIAN)
#define HOST_LITTLE_ENDIAN (iThisEndian == IS_LITTLE_ENDIAN)

#define  MIN_PROTOCOL       0
#define  MAX_PROTOCOL       1

/* select() uses different argument types on HP UX  before version 10.20.
   which is why SELECT is wrapped up in macro. The macro remains but
   the definition of select is now common across platforms.*/

#define SELECT(width, readfds, writefds, exceptfds, timeout) select(width, readfds, writefds, exceptfds, timeout)

/*
 * The following arre used with type_check option to check the datatype sent
 * against the datatype expected
 */

#define TICS_BYTE 0
#define TICS_SHORT   1
#define TICS_USHORT     2
#define TICS_INT  3
#define TICS_UINT 4
#define TICS_SFLOAT  5
#define TICS_FLOAT   6
#define TICS_STRING  7
#define TICS_STRING16   8
#define TICS_LONG 9
#define TICS_ULONG   10

/* With the following the datatype ID is folled by the number sent */

#define TICS_MULTIPLE   0xff00

#define TICS_SHORTS  (TICS_MULTIPLE+1)
#define TICS_USHORTS (TICS_MULTIPLE+2)
#define TICS_INTS (TICS_MULTIPLE+3)
#define TICS_UINTS   (TICS_MULTIPLE+4)
#define TICS_SFLOATS (TICS_MULTIPLE+5)
#define TICS_FLOATS  (TICS_MULTIPLE+6)
#define TICS_LONGS   (TICS_MULTIPLE+7)
#define TICS_ULONGS  (TICS_MULTIPLE+8)

/*********************************************************************
/**** END tics_int.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN mc_type.h
/*********************************************************************/

// WIN32

#define WINDOWS
// #define WIN32

#define x86

/*********************************************************************
/**** END mc_type.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN mc_defs.h
/*********************************************************************/

/* Big or little-endian? */

#define LITTLE_ENDIAN_INTEGERS

/* 64 bit flags */

// #define HAVE_64_BIT_INTS

#define IEEE_FLOAT

/*********************************************************************
/**** END mc_defs.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN datadefs.h
/*********************************************************************/

typedef void            Void;

typedef unsigned char   Byte;
typedef unsigned char   Char;
typedef unsigned short  Char16;
typedef Char           *Str;
typedef Char16         *Str16;
typedef Void           *Ptr;         /* DGK addition */
typedef int             Int;         /* always */
typedef unsigned int    UInt;        /* always */
typedef double          Double;         /* always */
typedef float           Float;          /* Only for special occasions */
typedef short           Short;          /* ditto */
typedef unsigned short  UShort;         /* ditto */

#ifdef NATIVE_LONG_IS_64_BIT
typedef int             Long;        /* ditto - defined thus for
                    * consistency */
typedef unsigned int    ULong;          /* ditto - defined thus for
                    * consistency */
#else
typedef long            Long;        /* ditto */
typedef unsigned long   ULong;          /* ditto */
#endif

#if !defined(Bool)             /* to avoid conflict with Xlib.h */
#define Bool         Int
#endif

#ifdef UNSIGNED_CHARS
typedef signed char     Int8;        /* an integer exactly 8 bits long */
#else
typedef char            Int8;
#endif
typedef unsigned char   UInt8;          /* an integer exactly 8 bits long */
typedef short           Int16;          /* an integer exactly 16 bits long */
typedef unsigned short  UInt16;         /* a uint exactly 16 bits long */
#ifdef HAVE_16_BIT_INTS
typedef long            Int32;          /* an integer exactly 32 bits long */
typedef unsigned long   UInt32;         /* a uint exactly 32 bits long */
#else
typedef int             Int32;          /* an integer exactly 32 bits long */
typedef unsigned int    UInt32;         /* a uint exactly 32 bits long */
#endif

typedef __int64         Int64;          /* an integer exactly 64 bits long */
typedef unsigned __int64 UInt64;       /* a uint exactly 64 bits long */

#define FORMAT_LONG "%I64d"
#define FORMAT_ULONG "%I64u"

/*********************************************************************
/**** END datadefs.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN tics_pri.h
/*********************************************************************/

typedef struct TicsHandle_s {
    /* public members */
    SOCKET          in;
    SOCKET          out;
    Ptr         write_stream; /* pointers not used in TICS itself, */
    Ptr         read_stream;  /* but by wrappers such as Marshalling */
    /* private members */
    Void            (*eofHandler) (struct TicsHandle_s * handle, Int32 status);
    Char           *buffer;
    UShort         maxBufferSize;
    UShort          curBufferSize;
    Char            bProtocol;
    Int32           autoFlush;
    Int             endianness;
    Int             floatieee;
    FILE           *Logfile;
    Bool     TypeCheck;
    Int32           Status;
} TicsHandleRec, *TicsHandle;

/*********************************************************************
/**** END tics_pri.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN tics.h
/*********************************************************************/

#include <windows.h>

/*
 * On windows the tics library can be built as a DLL by defining TICSDLL.
 * In this case all the tics calls require prefixes to get the correct
 * calling conventions. The ones here are specific to Microsoft Visual C++.
 * If not building TICS as a DLL the DLL_CALL macro is defined to nothing.
 *
 * When building an application which uses the TICS DLL also make sure that
 * TICSDLL is defined in order to match the correct calling conventions.
 *
 * Note that the calling convention declaration is compiler specific, the
 * following works with MS VC 4.0 and may not work with other compilers.
 */

#if defined(TICSDLL)
#  if defined(WIN16)
#     define DLL_CALL FAR PASCAL _export _loadds
#  elif defined(WIN32)
#     define DLL_CALL WINAPI
#  else
#     include "unrecognised windows platform"
#  endif
#else
#  define DLL_CALL
#endif

typedef         Void(*TicsEofHandler) (TicsHandle handle, Int32 status);

extern Int32    DLL_CALL ticsConnect(Str host, Str service, TicsHandle *handlep, Str opts);
extern Int32    DLL_CALL ticsConnectRegisteredService(Str registryRoot, Str service, TicsHandle *handlep, Str opts);
extern Void     DLL_CALL ticsDisconnect(TicsHandle handle);
extern Int32    DLL_CALL ticsCondition(TicsHandle thHandle, Byte Message);

/*
 * protocol exchange. This routine does not return until
 * the Magik end of the TICS has sent a suitable protocol,
 * which is returned. The usage is
 *
 * UShort protocolToUse = ticsEstablishProtocol(handle,
 *                   minProtocol,
 *                   maxProtocol);
 */

extern UShort     DLL_CALL ticsEstablishProtocol(TicsHandle handle, UShort min, UShort max);

/*
 * ticsSetEofHandler(Void (*proc)(UShort))
 *
 * Set EOF handling procedure. The procedure is called when any of the
 * ticsGetXXX procedures detect an EOF or when an ticsPutXXX routine
 * fails to write its data.  It is called with one Int argument, which
 * is 0 for EOF detected on input and nonzero for errors on output.
 *
 * Specifying NULL as the procedure reinstates the default handler,
 * which simply calls exit(0) under all circumstances.
 */

extern TicsEofHandler DLL_CALL ticsSetEofHandler(TicsHandle handle, TicsEofHandler func);

/*
 * Int32 oldAutoFlush = ticsSetAutoFlush(handle,newAutoFlush)
 *    TicsHandle handle;
 * Int32 newAutoFlush;
 *
 * Output flushing control. By default the ticsGetXXX routines will call
 * ticsFlush() to ensure all output is transmitted before trying to read
 * data. Depending on the machine and TICS, this may be inefficient. To
 * control this behaviour call ticsSetAutoFlush with a value that is
 * nonzero (zero) if automatic flushing by the get routines is wanted
 * (not wanted). The old value of the flag is returned.
 */

extern Int32    DLL_CALL ticsSetAutoFlush(TicsHandle handle, Int32 newAutoFlush);

/*
 * Int32 iStatus = ticsLogToFile(handle, FileName)
 *    TicsHandle handle;
 * Str FileName;
 *
 * If Filename is non NULL then all tics calls are logged to the given file.
 * If Filename is NULL then any existing log file is closed and logging
 * terminated.
 */

extern Int32    DLL_CALL ticsLogToFile(TicsHandle handle, Str FileName);

/*
 * When using the default TICS eofHandler the following may be used to
 * obtain the current status of the given connection. If everything is OK
 * the the value returned will be TICS_OK, otherwise it will be a TICS error
 * number.
 */

extern Int32   DLL_CALL ticsError(TicsHandle handle);

/*
 * ensure data gets through to Magik. If automatic flushing is disabled,
 * output data should usually be flushed before calling a get routine.
 */

extern Void     DLL_CALL ticsFlush(TicsHandle handle);

/*
 * following can be used to ensure TICS library/DLL version is suitable
 * for the application
 */

extern void DLL_CALL ticsVersion(UInt *major, UInt *minor);

/*
 * Base TICS get and put routine. Should only be used when client and
 * server are sure the data is formatted correctly, since no endian
 * conversion will be performed. Otherwise use ticsGetXXX and ticsPutXXX
 * convenience routines
 */

/*
 * ticsGet(handle, buffer, uiSize, uiCount)
 * TicsHandle handle;
 * Ptr buffer;
 * UShort uiSize;
 * UShort uiCount;

 * Read binary data from the TICS input. This is like fread(2) except
 * that no file is needed and the EOF handler is called if the correct
 * number of elements cannot be acquired.
 */

extern Void     DLL_CALL ticsGet(TicsHandle handle, Ptr buffer, UShort uiSize, UShort uiCount);

/*
 * ticsPut(handle, buffer, uiSize, uiCount)
 * TicsHandle *handle;
 * Ptr buffer;
 * UShort uiSize;
 * UShort uiCount;

 * This routine bears the same resemblance to fwrite(2) that ticsGet bears
 * to fread(2).
 */

extern Void     DLL_CALL ticsPut(TicsHandle handle, Ptr buffer, UShort uiSize, UShort uiCount);

/*
 * TICS i/o convenience routines.
 */

/*
 * NOTE: the floating point types follow Magik naming conventions, not C
 * conventions. A Magik float is a C double, a Magik short float is a C
 * float. Also, note that the Float, Short Byte and Bool functions deal in
 * terms of Doubles, and Ints because this is the only reliable way to
 * pass data to and from C functions.
 */

/*
 * reader functions. Singleton readers return the value,
 * multiple readers need a pointer and maximum count
 * Note: Strings are transmitted as <ushort count> <bytes>.
 * If the string is too long for the buffer, the initial part of
 * the string will be placed in the buffer and the rest discarded.
 * The return value in this case is NULL.
 */

#define ticsGetBytes(h, b, n) ticsGet(h, (Ptr)b, sizeof (Char), n)
#define ticsGetBool(h)     ticsGetByte(h)

extern Void     DLL_CALL ticsGetShorts(TicsHandle handle, Short *ptr, UShort count);
extern Void     DLL_CALL ticsGetUShorts(TicsHandle handle, UShort *ptr, UShort count);
extern Void     DLL_CALL ticsGetInts(TicsHandle handle, Int32 *ptr, UShort count);
extern Void     DLL_CALL ticsGetUInts(TicsHandle handle, UInt32 *ptr, UShort count);
extern Void     DLL_CALL ticsGetShortFloats(TicsHandle handle, Float *ptr, UShort count);
extern Void     DLL_CALL ticsGetFloats(TicsHandle handle, Double *ptr, UShort count);
extern Byte     DLL_CALL ticsGetByte(TicsHandle handle);
extern Short    DLL_CALL ticsGetShort(TicsHandle handle);
extern UShort   DLL_CALL ticsGetUShort(TicsHandle handle);
extern Int32    DLL_CALL ticsGetInt(TicsHandle handle);
extern UInt32   DLL_CALL ticsGetUInt(TicsHandle handle);
extern Float    DLL_CALL ticsGetShortFloat(TicsHandle handle);
extern Double   DLL_CALL ticsGetFloat(TicsHandle handle);
extern Str      DLL_CALL ticsGetString(TicsHandle handle, Str s, UShort maxLength);
extern Str16    DLL_CALL ticsGetString16(TicsHandle handle, Str16 s, UShort maxLength);

#if defined(HAVE_64_BIT_INTS)
extern Void     DLL_CALL ticsGetLongs(TicsHandle handle, Int64 *ptr, UShort count);
extern Void     DLL_CALL ticsGetULongs(TicsHandle handle, UInt64 *ptr, UShort count);
extern Int64    DLL_CALL ticsGetLong(TicsHandle handle);
extern UInt64   DLL_CALL ticsGetULong(TicsHandle handle);
#endif

/*
 * Writer functions. Only implement ticsPutBytes as a macro, all others
 * may need to change the input data.
 */

#define ticsPutBytes(h, b, n) ticsPut((h), (Ptr)b, sizeof (Char), (n))
#define ticsPutBools(h, b, n) ticsPut((h), (Ptr)b, sizeof (Byte), (n))
#define ticsPutBool(h,b)   ticsPutByte (h, b)

extern Void     DLL_CALL ticsPutByte(TicsHandle handle, Byte b);
extern Void     DLL_CALL ticsPutShort(TicsHandle handle, Short s);
extern Void     DLL_CALL ticsPutUShort(TicsHandle handle, UShort us);
extern Void     DLL_CALL ticsPutInt(TicsHandle handle, Int32 i);
extern Void     DLL_CALL ticsPutUInt(TicsHandle handle, UInt32 ui);
extern Void     DLL_CALL ticsPutShortFloat(TicsHandle handle, Float f);
extern Void     DLL_CALL ticsPutFloat(TicsHandle handle, Double d);
extern Void     DLL_CALL ticsPutShorts(TicsHandle handle, Short *ptr, UShort count);
extern Void     DLL_CALL ticsPutUShorts(TicsHandle handle, UShort *ptr, UShort count);
extern Void     DLL_CALL ticsPutInts(TicsHandle handle, Int32 *ptr, UShort count);
extern Void     DLL_CALL ticsPutUInts(TicsHandle handle, UInt32 *ptr, UShort count);
extern Void     DLL_CALL ticsPutShortFloats(TicsHandle handle, Float *ptr, UShort count);
extern Void     DLL_CALL ticsPutFloats(TicsHandle handle, Double *ptr, UShort count);
extern Void     DLL_CALL ticsPutString(TicsHandle handle, Str s);
extern Void     DLL_CALL ticsPutLenString(TicsHandle handle, Str s, UShort uslen);
extern Void     DLL_CALL ticsPutString16(TicsHandle handle, Str16 s);
extern Void     DLL_CALL ticsPutLenString16(TicsHandle handle, Str16 s, UShort uslen);

#if defined(HAVE_64_BIT_INTS)
extern Void     DLL_CALL ticsPutLong(TicsHandle handle, Int64 i);
extern Void     DLL_CALL ticsPutULong(TicsHandle handle, UInt64 ui);
extern Void     DLL_CALL ticsPutLongs(TicsHandle handle, Int64 *ptr, UShort count);
extern Void     DLL_CALL ticsPutULongs(TicsHandle handle, UInt64 *ptr, UShort count);
#endif

/*
 * TICS status codes
 */
#define TICS_OK          0     /* no problem */
#define TICS_NETWORK_DOWN   1     /* network software on local host not
                    * working */
#define TICS_HOST_UNKNOWN   2     /* don't recognise name of machine */
#define TICS_SERVICE_UNKNOWN   3     /* name of service unrecognised at low
                    * level */
#define TICS_SERVICE_UNAVAILABLE 4     /* service not running or unreachable */
#define TICS_TRY_AGAIN      5     /* service busy or max number of
                    * clients reached */
#define TICS_ACCESS_DENIED  6     /* access to service denied at low
                    * level */
#define TICS_PROTOCOL_MISMATCH    7     /* protocol mismatch (at any level) */
#define TICS_DISCONNECTED   8     /* connection to service unexpectedly
                    * broken */
#define  TICS_NO_MEMORY     9     /* attempt to allocate memory failed */
#define TICS_EXIT_SUCCESS  10     /* suitable exit success code for this
                    * platform */
#define TICS_EXIT_FAILURE  11     /* suitable exit failure code for this
                    * platform */

/*********************************************************************
/**** END tics.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN eswap.h
/*********************************************************************/

/*
 * end swapping on 2, 4 & 8 byte units.
 * nItems is the number of things to swap, *not* the length in bytes.
 * The pointers *must* be suitably aligned.
 *
 * 2 is for shorts,
 * 4 is for int, longs, floats,
 * 8 is for doubles.
 */

extern Void     endSwap2(Ptr ptr, UInt nItems);
extern Void     endSwap4(Ptr ptr, UInt nItems);
extern Void     endSwap8(Ptr ptr, UInt nItems);

/*
 * fancier names for those who want them
 */

#define endSwapShorts   endSwap2
#define endSwapInts  endSwap4
#define endSwapLongs endSwap4
#define endSwapFloats   endSwap4
#define endSwapDoubles  endSwap8

/*********************************************************************
/**** END eswap.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN swdp.h
/*********************************************************************/

#define _SWDP_H_ 1

/*
 * Publicly visible definitions for the TCP/IP dispatcher protocol.
 * $Id: swdp.h,v 1.2 1999/05/05 11:26:48 laurence Exp $
 * $Log: swdp.h,v $
 * Revision 1.2  1999/05/05  11:26:48  laurence
 * Chnage 12914_1: Added new client side messages.
 *
 * Revision 1.1  1997/08/07  09:21:06  laurence
 * Initial revision
 *
 */

#define DISPATCHER_NAME "swdp"
#define DISPATCHER_NAME_LENGTH 4
#define DISPATCHER_CLIENT_SERVICE "swdp-client"
#define DISPATCHER_CLIENT_DEFAULT_PORT 30001
#define DISPATCHER_SERVER_SERVICE "swdp-server"
#define DISPATCHER_SERVER_DEFAULT_PORT 30002
#define DISPATCHER_PROTOCOL 128

#define DISPATCHER_PRIORITY_DEFAULT 0
#define DISPATCHER_PRIORITY_LOW 1
#define DISPATCHER_PRIORITY_NORMAL 2
#define DISPATCHER_PRIORITY_HIGH 3

#define DISPATCHER_REDIRECTION_OK 0
#define DISPATCHER_REDIRECTION_FAILED 1

#define DISPATCHER_CONNECTION_OK 0
#define DISPATCHER_CONNECTION_REDIRECT 1
#define DISPATCHER_CONNECTION_NO_SERVICE 2
#define DISPATCHER_CONNECTION_PROTOCOL_ERROR 3
#define DISPATCHER_CONNECTION_ERROR 4

/*********************************************************************
/**** END swdp.h
/*********************************************************************/

/**********************************************************************
/######################################################################
/#### acpt32 implementation
/######################################################################
/*********************************************************************/

/*********************************************************************
/**** BEGIN acpt.h
/*********************************************************************/

/*
 * acpt.[ch]: C code for use in ACPs which wish to transfer
 * objects between C and Magik.
 */

#define _ACPT_H_DONE_

/*********************************************************************
/**** BEGIN acpt_mc.h
/*********************************************************************/

/*
 * acpt_mc.h: header which works out whether we need 16-bit ints and whether
 * we need multi-threading.
 */
#if defined(_WIN16) || defined(WIN16) || (defined(_WINDOWS) && !defined(WIN32) && !defined(_WIN32))
#  if !defined(HAVE_16_BIT_INTS)
#  define HAVE_16_BIT_INTS
#  endif
#endif

#if defined(_WINDOWS) || defined(_WIN16) || defined(WIN16) || defined(_WIN32) || defined (WIN32)
#  define ACPT_MULTI_THREADED
#include <windows.h>
#endif

#if !defined(SW_BASIC_H_DONE) && !defined(DATADEFS_H)
#ifdef HAVE_16_BIT_INTS
typedef long            Int32;          /* an integer exactly 32 bits long */
typedef unsigned long   UInt32;         /* a uint exactly 32 bits long */
#else
typedef int             Int32;          /* an integer exactly 32 bits long */
typedef unsigned int    UInt32;         /* a uint exactly 32 bits long */
#endif
#endif

#if defined(ACPTDLL)
#  if defined(WIN16)
#     define ACPT_DLL_CALL FAR PASCAL _export _loadds
#  elif defined(WIN32)
#     define ACPT_DLL_CALL WINAPI
#  else
#     include "unrecognised windows platform"
#  endif
#else
#  define ACPT_DLL_CALL
#endif

/*********************************************************************
/**** END acpt_mc.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN acpt_excp.h
/*********************************************************************/

#define ACPT_OK            0

#define ACPT_EXCP_BAD_DISCRIM    1
#define ACPT_EXCP_BAD_DISCRIM_ON_WRITE 2
#define ACPT_EXCP_BAD_FLAGS      3
#define ACPT_EXCP_BAD_INTRO      4
#define ACPT_EXCP_BAD_LENGTH     5
#define ACPT_EXCP_BAD_PROTOCOL      6
#define ACPT_EXCP_BAD_SEND_TYPE     7
#define ACPT_EXCP_BAD_VECTOR_TYPE   8
#define ACPT_EXCP_FIELD_SIZE_TOO_SMALL 9
#define ACPT_EXCP_INCOMPATIBLE_TYPE 10
#define ACPT_EXCP_INTS_TOO_SMALL 11
#define ACPT_EXCP_LENGTH_TOO_LONG   12
#define ACPT_EXCP_MEMORY_EXHAUSTED  13
#define ACPT_EXCP_NO_MORE_TYPES     14
#define ACPT_EXCP_NO_SUCH_TYPE      15
#define ACPT_EXCP_NULL_ADDR      16
#define ACPT_EXCP_NULL_TYPE      17
#define ACPT_EXCP_SIZE_TOO_SMALL 18
#define ACPT_EXCP_STREAM_NOT_READER 19
#define ACPT_EXCP_STREAM_NOT_WRITER 20
#define ACPT_EXCP_STRING_TOO_LONG   21
#define ACPT_EXCP_TYPE_MUDDLE    22
#define ACPT_EXCP_UNSUPPORTED_REF_REF  23
#define ACPT_EXCP_VECTOR_SENT    24
#define ACPT_EXCP_BAD_BOUNDED_STRING    25
#define ACPT_EXCP_SCALAR_SENT    26
#define ACPT_EXCP_STREAM_WRITE_FAILED   27
#define ACPT_EXCP_STREAM_EXHAUSTED      28
#define ACPT_EXCP_NO_MORE_STREAMS       29
#define ACPT_EXCP_UNFREED_TYPE          30
#define ACPT_EXCP_STREAM_MANGLED        31
#define ACPT_EXCP_UNSUPPORTED_TYPE      32
#define ACPT_EXCP_CONVERSION_ERROR      33
#define ACPT_EXCP_VALUE_OUT_OF_RANGE    34
#define ACPT_EXCP_HETERO_TYPE_MISMATCH  35
#define ACPT_EXCP_BAD_MAGIK_EXEMPLAR    36
#define ACPT_EXCP_BAD_TYPE              37
#define ACPT_EXCP_STALE_STREAM          38
#define ACPT_EXCP_NO_RECORD_ON_JOIN     39
#define ACPT_EXCP_SLOT_IN_JOINED_REC    40
#define ACPT_EXCP_CANNOT_READ_JOIN      41
#define ACPT_EXCP_BAD_INDIRECT          42
#define ACPT_EXCP_MAX_BUFSIZ_EXCEEDED   43
#define ACPT_EXCP_LENGTH_MISMATCH       44
#define ACPT_EXCP_NFIELDS_MISMATCH      45
#define ACPT_EXCP_FLAVOUR_MISMATCH      46
#define ACPT_EXCP_FIELD_FLAGS_MISMATCH  47
#define ACPT_EXCP_DISCRIM_MISMATCH      48
#define ACPT_EXCP_SCALAR_TYPE_MISMATCH  49
#define ACPT_EXCP_LOOKASIDE_ERROR       50
#define ACPT_EXCP_NONEXISTENT_PRIM      51

/*********************************************************************
/**** END acpt_excp.h
/*********************************************************************/

/*
 * Some very basic typdefs and constants.
 */

#if !defined(NULL)
#   define NULL 0
#endif

#if !defined(Bool)   /* to avoid conflicts */
#define Bool         Int32
#endif

#if !defined(FALSE)
#   define FALSE 0
#endif

#if !defined(TRUE)
#   define TRUE 1
#endif

#   define ACPT_KLEENE_TRUE 2
#   define ACPT_KLEENE_FALSE 0
#   define ACPT_KLEENE_MAYBE 1


#if !defined(ARRAY_LEN)
#define ARRAY_LEN(a) (Int32)(sizeof(a)/sizeof(a[0]))
#endif

/*
 * the following macros are useful in struct field declarations. _TYPE is
 * a struct type, field is the name of a field within it.
 */
#define OFFSET(_type, _field) ((char *)&(((_type *)0)->_field)-((char *)0))
#define SIZEOF(_type, _field) sizeof(((_type *)0)->_field)
#define OFFSIZ(_type, _field) OFFSET(_type, _field), SIZEOF(_type, _field)

/*
 * The following are the hardwired values for basic ACPT types.
 */
#define ACPT_TYPE_UNSET            1    /* Unset */

#define ACPT_TYPE_UINT8       2  /* Byte */
#define ACPT_TYPE_UINT16            3    /* Unsigned Short */
#define ACPT_TYPE_UINT32      4    /* Unsigned Int */
#define ACPT_TYPE_UINT64            5    /* Unsigned 64-bit Int */

#define ACPT_TYPE_INT8        6  /* Signed Byte */
#define ACPT_TYPE_INT16       7    /* Signed Short */
#define ACPT_TYPE_INT32       8    /* Signed Int */
#define ACPT_TYPE_INT64       9    /* Signed 64-bit Int */

#define ACPT_TYPE_FLOAT32    10    /* 32-bit float */
#define ACPT_TYPE_FLOAT64    11    /* 64-bit float (ie double) */

#define ACPT_TYPE_CHAR8      12    /* 8-bit character */
#define ACPT_TYPE_CHAR16     13    /* 16-bit character */

#define ACPT_TYPE_BOOL       14  /* Boolean - actually Int32 underneath */
#define ACPT_TYPE_KLEENE     15  /* Kleenean - ditto */

#define ACPT_TYPE_SIMPLE_TIME     16    /* Time - actually Int seconds */
#define ACPT_TYPE_DATE            17    /* Date - actually Int days */
#define ACPT_TYPE_DATE_TIME       18    /* Date and Time - two ints */

#define ACPT_TYPE_DOUBLE_AS_INT   19    /* Expressed as double, sent as int */

#define ACPT_TYPE_SHORT_STRING8    20  /* String of 8-bit chars < 256 chars */
#define ACPT_TYPE_SHORT_STRING16  21   /* String of 16-bit chars < 255 chars */
#define ACPT_TYPE_LONG_STRING8     22  /* String of 8-bit chars any length */
#define ACPT_TYPE_LONG_STRING16    23  /* String of 16-bit chars any length */

#define ACPT_TYPE_EMPTY_VECTOR    24    /* Special type doing readObjectVector
                * and no type was specified and
                * nothing was sent.
                */
#define ACPT_TYPE_CHAR8_TO_16     25     /* Special type where the local type is
                 * Char16 but the remote type is Char8
                 */
#define ACPT_TYPE_SHORT_STRING8_TO_16 26 /* Special string type < 255 chars where
                 * local type is Char16 but the remote type
                 * is Char8
                 */
#define ACPT_TYPE_LONG_STRING8_TO_16  27 /* Special string type (any length) chars
                 * where local type is Char16 but the remote
                 * type is Char8
                 */
#define ACPT_TYPE_BIGINT         45      /*
                 * Corresponds to a Magik Bignum/Java BigInt.
                 * Represented in C as a vector of unsigned
                 * char (ie Byte) where the first byte is the
                 * number of bytes which follow. The top bit
                 * of the second byte is the sign bit and the
                 * remaining value is 2-s complement. Eg:
                 * {1, 0} = 0
                 * {1, 127} = 127
                 * {1, 128} = -128
                 * {1, 255} = -1
                 * {2, 0, 128} = 128
                 * {2, 127, 255} = 32767
                 * {2, 128, 0} = -32768
                 */
typedef struct AcptDateTime {
    Int32 day;       /* day number, Jan 1st 1970 = 0 */
    Int32 msecond;      /* time in Ms since midnight */
} AcptDateTime;

/* This list not to exceed 63 */

#define ACPT_MAX_BUILTIN_TYPES 128


/* The following values define the vector properties of a type and are
 * mutually exclusive.
 */
#define ACPT_SCALAR     0x0
#define ACPT_FIXED_VECTOR  0x1
#define ACPT_SHORT_V_VECTOR   0x3
#define ACPT_LONG_V_VECTOR 0x7
#define ACPT_FIXED_VECTOR_REF 0x101
#define ACPT_SHORT_V_VECTOR_REF  0x103
#define ACPT_LONG_V_VECTOR_REF   0x107
#define ACPT_BOUNDED_STRING     0x9

/*
 * The following six values should not be used during definition, but can
 * be returned via acptQueryStreamType().
 */
#define ACPT_SHORT_V_STRING      0xb
#define ACPT_LONG_V_STRING      0xf

#define ACPT_SHORT_BOUNDED_STRING     0x9
#define ACPT_LONG_BOUNDED_STRING      0xd

#define ACPT_STRUCT     0x30
#define ACPT_UNION      0x10


/*
 * The following flags define characteristics of a field in a struct,
 * and may be or'd together. REF may *not* be used in an on-the-fly
 * declaration of a 'V' vector.
 */
#define ACPT_UNSET      0x1
#define ACPT_REF     0x2

/* Default values provided for completeness */
#define ACPT_NO_UNSET      0x0
#define ACPT_INLINE     0x0


/*
 * The 'unset flag' value, set to indicate that the field is unset
 */
#define ACPT_UNSET_FLAG 0x01
#define acptFieldIsUnset(_field) ((_field & ACPT_UNSET_FLAG) == ACPT_UNSET_FLAG)
#define acptFieldIsSet(_field) ((_field & ACPT_UNSET_FLAG) != ACPT_UNSET_FLAG)
#define acptSetFieldUnset(_field) ((_field) |= ACPT_UNSET_FLAG)
#define acptSetFieldSet(_field) ((_field) &= ~(ACPT_UNSET_FLAG))

typedef Int32 AcptResult; /* Result of any function. 0 = OK */

typedef UInt32 AcptType; /* Any type */


typedef struct AcptField {
    /*
     * A field in a struct or union can be defined statically in-situ to have
     * vector properties, to save the need to dynamically create a new type
     * for each vector. For a scalar type, simply do {type}.
     */
    AcptType type;       /* The field type, or element type if vector */
    UInt32 vec_type;     /* Scalar/Array type of this field - if 0 or omitted,
                          * the default is ACPT_SCALAR.
                          * Legitimate values comprise:
                          *   ACPT_SCALAR: field is scalar
                          *   ACPT_FIXED_VECTOR: field is fixed length array
                          *       of TYPE, the length being supplied in
                          *       LENGTH_DETAIL.
                          *   ACPT_BOUNDED_STRING: only valid in conjunction with
                          *       CHAR and CHAR16 types, this option is used to
                          *       define a null terminated 'string' field of fixed
                          *       maximum length (supplied in LENGTH_DETAIL). Note
                          *       that the supplied length must *include* space
                          *       for the trailing NULL.
                          *   ACPT_SHORT_V_VECTOR: the field is a variable
                          *       length array of TYPE, which is never longer
                          *       than 255 elements long. LENGTH_DETAIL must be
                          *       set to an offset in the owning struct which
                          *       refers to an Int32 field which will contain
                          *       the length. The vector field will always
                          *       be a reference to dynamically allocated memory.
                          *   ACPT_LONG_V_VECTOR: As for ACPT_SHORT_V_VECTOR,
                          *       except that there is no limit on field length.
                          *   ACPT_FIXED_VECTOR_REF:  } Equivalent array types,
                          *   ACPT_SHORT_V_VECTOR_REF:} but each element of the
                          *   ACPT_LONG_V_VECTOR_REF: } array is a ref to the
                          *                           } value, rather than being
                          *                           } the value itself.
                          *   Note that VECTOR_REF and VECTOR are treated as
                          *   the same for type-comparison purposes, so data
                          *   sent as one can be received as the other.
                          */
    Int32 length_detail; /*
                          * If scalar this can be 0 or omitted. If
                          * FIXED_VECTOR or BOUNDED_STRING, then this should
                          * be set to the array length (in the latter case
                          * *including* the tailing NULL), if a V_VECTOR then this
                          * should be set to an offset in the owning struct
                          * where the actual length of the array is to be
                          * found or deposited. The length will be an Int32.
                          */
} AcptField;

typedef struct AcptStructElement {
    /*
     * Describes an individual field in a structure. Note that the
     * 'TYPE' field is a struct, so when statically declaring a structure
     * it should be enclosed in braces (eg {{ACPT_INT}, offset, size})
     */
    AcptField type;      /* The field's type */
    Int32 offset;        /* its offset within this struct or union */
    Int32 size;          /* the size it occupies in the body of the struct, in bytes */
    UInt32 flags;        /* properties of this field: legitimate flags (or'd
                          * together) comprise:
                          *  ACPT_INLINE/ACPT_REF: says whether the field is
                          *      in-line in the struct, or accessed via a
                          *      reference. If TYPE defines a V_VECTOR on the
                          *      fly, then REF may not be used (as the vector
                          *      data itself will always be via a reference).
                          *      Note that the REF aspect is ignored for type
                          *      comparison purposes, so data sent as REF can
                          *      be received as INLINE and vice-versa.
                          *  ACPT_UNSET/ACPT_NO_UNSET: says whether the field
                          *      may be unset. If so, DETAIL_OFFSET must be
                          *      supplied to indicate where the unset flag
                          *      is to be deposited or read from. Note that
                          *      if a field may be unset, and its value when
                          *      read *is* unset, then the memory corresponding
                          *      to where the genuine value would go is zeroed.
                          */
    Int32 detail_offset; /* The offset at which flags will be deposited. the
                          * field at this offset should be a UInt32, and is
                          * currently only used to indicate that the field is
                          * (or is not) unset (as indicated by acptFieldIsUnset()
                          * and set by acptSetFieldUnset()).
                          */
} AcptStructElement;


typedef struct AcptUnionElement {
    /*
     * Describes an individual field in a union. Note that the
     * 'TYPE' field is a struct, so when statically declaring a structure
     * it should be enclosed in braces (eg {{ACPT_INT}, size, discrim_value})
     */
    AcptField type;      /* The field's type */
    Int32 size;          /* The size it occupies in the body of the struct, in bytes */
    Int32 discrim_value; /* The discriminator value which matches this field. This
                          * is always Int32.
                          */
    UInt32 flags;        /* properties of this field: legitimate flags (or'd
                          * together) comprise:
                          *  ACPT_INLINE/ACPT_REF: says whether the field is
                          *      in-line in the struct, or accessed via a
                          *      reference. If TYPE defines a V_VECTOR on the
                          *      fly, then REF may not be used (but the vector
                          *      data itself will always be via a reference).
                          */
} AcptUnionElement;


/*
 * The following structures are used to return error information in cases
 * where the error occurs down a recursive call and may be one of many
 * sorts.
 */
typedef struct {
    Int32 el;
    UInt32 flavour;
} AcptStreamErrorDetail;

typedef struct {
    Int32 code;                 /* A precise error code */
    Int32 c_data;       /* Data information relating to C */
    Int32 s_data;                  /* Data information relating to the stream */
    Int32 depth;        /* The current error's recursive depth */
    Int32 max_depth;               /* The error stack's HWM */
    AcptStreamErrorDetail *stack;  /* Stack of errors: last el indicates error */
    AcptStreamErrorDetail *sp;     /* Internal use only */
} AcptStreamError;

typedef struct AcptStream {
    /*
     * Defines the properties of a stream. It holds the stream handle,
     * a method_table of functions to read and write data, and details
     * of the current object read (if any). Note that this structure should
     * generally be regarded as opaque by programmers. However, the
     * embedded Errors struct can provide useful information on type-
     * mismatch errors.
     */
    void *handle;                     /* Opaque physical handle for stream */
    AcptResult (*Transfer)(void *handle,
                           unsigned char *buffer,
                           Int32 n_bytes); /* raw data mover */
    char reader;                      /* stream for read (TRUE), write (FALSE) */
    char started;                     /* TRUE if intro read or written */
    char object_current;              /* TRUE if object details meaningful */
    char byte_swapped;                /* TRUE (on reading) if byte-swapped */

    AcptType current_object_type;     /* Current object's type */
    Int32 current_object_is_vector;   /* TRUE if it was writeVector'd */
    Int32 current_object_length;      /* If writeVector'd, its length */

    Int32 buffer_max;                 /* The size the buffer is declared to */
    Int32 buffer_abs_max;        /* Absolute max size buffer allowed */
    Int32 buffer_count;               /* Used during reading: the size used */
    Int32 buffer_count_offset;        /* Says where buffer count is to go */
    Int32 typeDescListMax;            /* The HWM of this list */
    struct AcptTypeDesc **typeDescs;  /* Known inbound stream types */

    unsigned char *buffer;            /* Current pos for reading or writing */
    unsigned char *buffer_start;      /* Physical start of buffer  */
    unsigned char *data_start;        /* Where the data starts, after error  */

    /*
     * The next fields are to do with recording whether a type has been
     * sent, and cleaning this information in case of errors. They are
     * only used by Write streams.
     */
    unsigned char *SentTypes;       /* Bitmap, bit set means type sent */
    unsigned char *SentTypesThisGo;   /* Ditto, but just sent in one Write */
    Int32 SentTypesMax;          /* Max type sent */
    Int32 any_type_sent;         /* TRUE if any type sent in a Write */
    AcptStreamError Errors;         /* Error detail after type_mismatch */
#if defined(_ACPTM_C_ONLY_)
    unsigned char timezone[65];       /* Name of timezone, if any */
    Int32 time_offset;                /* time offset, if set */
    void *transfm;                    /* Datastore transform, if any */
#endif
} AcptStream;


/*
 *     E X T E R N A L   I N T E R F A C E S
 *    ________________________________________
 */

/*
 *       I N I T I A L I S A T I O N
 *       ___________________________
 */
extern AcptResult ACPT_DLL_CALL acptInit();
/*
 * Mandatory call prior to using Marshalling. Note that this is *not*
 * required when using acp-io, as that encapsulates the call in acpInitAcpt.
 * This function should only be called once during a program run.
 */

extern AcptResult ACPT_DLL_CALL acptEnd();
/*
 * Not required in normal programming, this function returns an error if any
 * user defined acpt types have not been freed, and is useful when tracking
 * down memory leaks.
 */

extern AcptResult ACPT_DLL_CALL acptVersion(UInt32 *major, UInt32 *minor,
                   UInt32 *bugfix);
/*
 * Populates the three arguments with Major, Minor and Bugfix version numbers.
 */

/*
 *          S T R E A M S
 *          _____________
 */
extern AcptResult ACPT_DLL_CALL
acptNewStream(void *handle,
         AcptResult (*Transfer)(void *handle,
                 unsigned char *buffer,
                 Int32 n_bytes),
         Bool reader, AcptStream **streamp);
/*
 * Populates STREAMP with a pointer to a stream handle to be used subsequently
 * by WriteObject/ReadObject etc calls. TRANSFER is a caller supplied function
 * which either reads (READER is TRUE) or writes (READER is FALSE) N_BYTES from
 * the BUFFER argument from or to the physical stream identified by HANDLE.
 * HANDLE is passed to acptNewStream which then stores it and subsequently
 * passes it to TRANSFER when data is to be moved.
 */


extern AcptResult ACPT_DLL_CALL acptFreeStream(AcptStream *stream);
/*
 * Frees any acpt resources associated with STREAM, but does *not* do anything
 * to close the underlying physical stream (identified by HANDLE in acptNewStream()).
 */


extern AcptResult ACPT_DLL_CALL acptSetStreamLimit(AcptStream *stream, Int32 limit);
/*
 * Imposes an artificial limit on the number of bytes which may be buffered
 * before transfer. If this limit is breached during a Write call, an exception
 * is returned and the Write fails. The limit only applies to writing, not
 * reading. The default is unlimited (limit = 0).
 */



/*
 *       T Y P E   D E F I N I T I O N
 *       _____________________________
 */
extern AcptResult ACPT_DLL_CALL
acptNewStructType(AcptStructElement field_specs[], int n_fields,
        int size, AcptType *typep);
/*
 * Populates TYPEP with a reference to a new type which defines a
 * C struct which can be transferred. FIELD_SPECS (N_FIELDS long)
 * describes each field in the struct, and the order in which they
 * will be transferred. SIZE is the size of the C struct, in bytes.
 */


extern AcptResult ACPT_DLL_CALL
acptNewUnionType(AcptUnionElement field_specs[], int n_fields,
       int data_offset, int discrim_offset, int size,
       AcptType *typep);
/*
 * Populates TYPEP with a reference to a new type which defines a
 * C discriminated union which can be transferred. FIELD_SPECS
 * (N_FIELDS long) describes each element in the union. The union
 * will manifest itself as a structure with two fields, the actual
 * data (at DATA_OFFSET) and the Int32 discriminator value at
 * DISCRIM_OFFSET.  SIZE should be provided as the size of the
 * struct which encapsulates the discriminated union.
 */


extern AcptResult ACPT_DLL_CALL
acptNewFixedVectorType(AcptType el_type, unsigned int vec_type,
             int length, AcptType *typep);
/*
 * Populates TYPEP with a reference to a new type which defines a
 * fixed vector type field of length LENGTH. VEC_TYPE describes the
 * field, and should be one of:
 *   ACPT_FIXED_VECTOR: field is fixed length array of TYPE, the
 *       length being supplied in LENGTH.
 *   ACPT_BOUNDED_STRING: only valid in conjunction with CHAR8 and
 *       CHAR16 types, this option is used to define a null
 *       terminated 'string' field of fixed maximum length (supplied
 *       in LENGTH), populated in-situ (ie the data is not via a ref).
 *   ACPT_FIXED_VECTOR_REF:  The vector is a vector of ref's to the data
 *       rather than holding the data in-line. Note that the 'REF' aspect
 *       is ignored when type matching (so data sent as FIXED_VECTOR
 *       can be received as FIXED_VECTOR_REF and vice-versa).
 *
 * The size the field will occupy is assumed to be
 * LENGTH*element_size (where element-size is the size of a
 * pointer in the case of VECTOR_REF).  LENGTH must be >= 0.
 */


extern AcptResult ACPT_DLL_CALL
acptNewVVectorType(AcptType el_type, unsigned int vec_type,
         int vec_offset, int length_offset, int size,
         AcptType *typep);
/*
 * Populates TYPEP with a reference to a new type which defines a
 * variable length vector type field. The variable length array
 * will manifest itself as a structure with two fields, the length
 * (Int32) (at LENGTH_OFFSET) and the vector data (which will be a
 * pointer to dynamically allocated memory) at VEC_OFFSET. SIZE
 * should be provided as the size of the struct (which does not
 * have to comprise simply an Int32 and a pointer).
 *
 * VEC_TYPE describes the field, and should be one of:
 *   ACPT_SHORT_V_VECTOR: the field is a variable length array of TYPE,
 *       which is never longer than 255 elements long. LENGTH_OFFSET must
 *       be set to an offset in an owning struct which refers to an Int32
 *       field which will contain the length. The vector field itself
 *       will always be a reference to dynamically allocated memory.
 *   ACPT_LONG_V_VECTOR: As for ACPT_SHORT_V_VECTOR,
 *       except that there is no limit on field length.
 *   ACPT_SHORT_V_VECTOR_REF:} equivalent types, but each element of the
 *   ACPT_LONG_V_VECTOR_REF: } array is a ref to the value.
 */


extern AcptResult ACPT_DLL_CALL acptFreeType(AcptType type);
/*
 * Free TYPE so that space it required can be reused.
 *
 * Types are internally heavyweight, so it is desirable to free a
 * type which is no longer required as soon as possible. Note that
 * it is legitimate to Free a type as soon as access is no longer
 * required to it explicitly - even though it is referred to as
 * part of another type. Internal ref-counting ensures that the
 * type will only really die when there are genuinely no
 * references. This means that it is legitimate to define a type
 * temporarily, incorporate it into something else (eg a vector or
 * struct) and then immediately free it again.
 */


/*
 *          S T R E A M   I / O
 *          ___________________
 */
extern AcptResult ACPT_DLL_CALL
acptReadObject(AcptStream *stream, AcptType type, void *datum,
          Int32 datum_size);
/*
 * Read the next object from STREAM, which must be compatible with TYPE,
 * into the space at address DATUM which is DATUM_SIZE bytes long. The
 * data must have been sent using writeObject functionality: if
 * writeObjectVector functionality was used, then readObjectVector
 * must be used to read the data.
 *
 * Note: In the majority of cases where this function returns an
 * exception, the inbound type can queried (using
 * acptPeekObject()) and another attempt made to read the
 * data (eg using readObjectVector, or providing a different type
 * or size) - (so read-failure at this juncture is typically
 * non-destructive).
 */


extern AcptResult ACPT_DLL_CALL
acptWriteObject(AcptStream *stream, AcptType type, void *datum);
/*
 * Write the object at DATUM which matches type TYPE to STREAM. Note that
 * any errors which arise are non-destructive - in all cases other than
 * the stream itself being unwritable, nothing is written, and if it was
 * physically unwritable nothing would be written on the stream.
 */


extern AcptResult ACPT_DLL_CALL
acptReadObjectVector(AcptStream *stream, AcptType type,
           void *datums, Int32 space, Int32 *n_objects);
/*
 * Read a homogeneous vector of objects from STREAM, which must be
 * of a type compatible with TYPE, into the space at address
 * DATUMS which is SPACE bytes long. The data must have been sent
 * using writeObjectVector functionality: if writeObject
 * functionality was used, then readObject must be used to read
 * the data. N_OBJECTS (if not NULL) will be populated with the
 * number of objects retrieved.
 *
 * DATUMS will be populated with in-line objects.
 *
 * Note: In the majority of cases where this function returns an
 * exception, the inbound type can queried (using
 * acptPeekObject()) and another attempt made to read the
 * data (eg using readObject, or providing a different type
 * or size) - (so read-failure at this juncture is typically
 * non-destructive).
 */


extern AcptResult ACPT_DLL_CALL
acptWriteObjectVector(AcptStream *stream, AcptType type, void *datums,
            Int32 length);
/*
 * Write LENGTH objects in DATUMS which match type TYPE to STREAM.
 *
 * The objects in DATUMS are assumed to be in-line (ie not references).
 *
 * Note that any errors which arise are non-destructive - in all
 * cases other than the stream itself being unwritable, nothing is
 * written, and if it was physically unwritable nothing would be
 * written on the stream.
 */


extern AcptResult ACPT_DLL_CALL
acptPeekObject(AcptStream *stream, UInt32 *typeno, Bool *is_vectorp,
          Int32 *lengthp);
/*
 * Interrogates the item on the stream which would be consumed by
 * the next call to ReadObject, ReadObjectVector or SkipNext. This
 * call is non-destructive, and may be made repeatedly before one
 * of the above calls, and also after a failed call to ReadObject
 * or ReadObjectVector. TYPENO is populated with the *stream type*
 * - this will only match a local type if it is one of the
 * built-in ones; IS_VECTORP is populated with TRUE if
 * ReadObjectVector() is required to read the object, FALSE if
 * ReadObject is required; and LENGTHP with the length if
 * IS_VECTORP is TRUE. Any of these three arguments may
 * legitimately be NULL. Note that under some circumstances, if
 * IS_VECTORP is TRUE and if LENGTHP is 0, TYPENO will be
 * ACPT_TYPE_EMPTY_VECTOR.
 */


extern AcptResult ACPT_DLL_CALL acptSkipObject(AcptStream *stream);
/*
 * Skip the next or pending (if the previous request was a failed Read,
 * or a Peek) object (or vector of objects) on STREAM.
 */


extern AcptResult ACPT_DLL_CALL
acptFreeObject(AcptType type, void *datum);
/*
 * Free any space allocated during acptReadObject() to DATUM which
 * is an object of type TYPE.
 *
 * Note that the object memory itself is *not* freed - only space which it
 * refers to. For successful operation, the _unset flag on a field, and the
 * length field of a variable length array, must be used correctly:
 * If a field can be unset, and *is* unset, then space it may potentially
 * refer to is ignored (so if a field is marked _unset, the caller should
 * also ensure that the object it referred to is freed at the same time).
 * Similarly, if the length of a variable length array is 0, the space
 * reference by the array is not touched.
 */


extern AcptResult ACPT_DLL_CALL
acptFreeObjectVector(AcptType type, void *datums, Int32 length);
/*
 * Free any space allocated during acptReadObjectVector() to DATUMs which
 * is an in-line vector of LENGTH objects of type TYPE.
 *
 * Note that DATUMS itself is *not* freed - only space which it
 * refers to. For successful operation, the _unset flag on a field, and the
 * length field of a variable length array, must be used correctly:
 * If a field can be unset, and *is* unset, then space it may potentially
 * refer to is ignored (so if a field is marked _unset, the caller should
 * also ensure that the object it referred to is freed at the same time).
 * Similarly, if the length of a variable length array is 0, the space
 * reference by the array is not touched.
 */


/*
 *    S T R E A M   T Y P E   Q U E R Y I N G
 *    _______________________________________
 */

extern AcptResult ACPT_DLL_CALL acptTypenoForType(AcptType type, UInt32 *typeno);
/*
 * Populates TYPENO with the type number which will be sent
 * associated with TYPE. Note that although this is often TYPE, it
 * isn't guaranteed to be so.
 */


extern AcptResult ACPT_DLL_CALL
acptQueryStreamType(AcptStream *stream, UInt32 typeno,
          UInt32 *flavour, Int32 *data_detail,
          Int32 *subtype, Int32 **subtypes);
/*
 * Returns details of the stream type from STREAM whose number is
 * TYPENO.  FLAVOUR will be populated with the flags which define
 * what sort of type this is, and will be one of:
 *     ACPT_SCALAR
 *     ACPT_FIXED_VECTOR
 *     ACPT_SHORT_V_VECTOR
 *     ACPT_LONG_V_VECTOR
 *     ACPT_SHORT_BOUNDED_STRING
 *     ACPT_LONG_BOUNDED_STRING
 *     ACPT_SHORT_V_STRING
 *     ACPT_LONG_V_STRING
 *     ACPT_STRUCT
 *     ACPT_UNION
 *
 * DATA_DETAIL will be populated with a value depending on
 * FLAVOUR: if FLAVOUR is is ACPT_FIXED_VECTOR or
 * ACPT_[LONG/SHORT]_BOUNDED_STRING then DATA_DETAIL is populated
 * with the length value; if FLAVOUR is ACPT_STRUCT or ACPT_UNION
 * then it is populated with the number of subtypes. If FLAVOUR
 * indicates a vector type, SUBTYPE is populated with the typeno
 * of the subtype (otherwise it's NULLed). If FLAVOUR is
 * ACPT_STRUCT or ACPT_UNION, SUBTYPES is populated with an
 * integer vector containing the typnos or the underlying fields
 * and followed by the same number of elements containing flags
 * (indicating _unset allowed, if FLAVOUR is ACPT_STRUCT) or
 * discriminators (if FLAVOUR is ACPT_UNION) - so the vector is
 * 2*data_detail long.
 *
 * If SUBTYPES is not relevant the argument is populated with
 * NULL; if it is relevant then it is the caller's responsibility
 * to free the space using acptFree() when no longer required.
 */


extern AcptResult ACPT_DLL_CALL
acptSetTypeMismatchHandler(void (ACPT_DLL_CALL *handler)(AcptStreamError *error));
/*
 * Declare a handler which will be called when a type-mismatch
 * error occurs. The handler takes one argument, which will be
 * supplied as a pointer to an AcptStreamError struct, from which
 * various information about the error can be established. This
 * handler could then, for example, be used to print details of
 * the mismatch on standard-error. An acpt function
 * acptPrintStreamMismatchError() is provided for this purpose,
 * but this is *not* the default. The default is a NULL handler
 * (which therefore does nothing).
 */

extern void ACPT_DLL_CALL acptPrintStreamMismatchError(AcptStreamError *error);
/*
 * Convenience Handler which can be supplied to
 * acptSetTypeMismatchHandler which will write the error information to
 * stderr.
 */


/*
 *       M E M O R Y   M A N A G E M E N T
 *       _________________________________
 */

extern AcptResult ACPT_DLL_CALL
acptSetDataAllocators(void * (*calloc)(), void (*free)());
/*
 * Set the allocators to be used for generating and freeing space
 * used in object data. These allocators are *not* used for internal
 * space (eg for types or streams). The precise spec of CALLOC and
 * FREE is the defined spec for the ANSII standard versions, and
 * should be rigorously adhered to. In particular, CALLOC should
 * return NULL on memory_exhausted.
 */

#define acptFree(_loc) if (_loc) free(_loc); else;

/*
 * flavourNames is a vector of strings indexed by a FLAVOUR which is
 * something like ACPT_SHORT_V_VECTOR or ACPT_STRUCT. FLAVOUR will be
 * encountered in type-mismatch error information and also in the
 * result of acptQueryStreamType(). The values in these cases permit
 * direct lookup into the flavourNames vector.
 */
extern char *flavourNames[64];

/*********************************************************************
/**** END acpt.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN acpt_private.h
/*********************************************************************/

/*
 * acpt_private.h: Header which defines interfaces used by internal
 * Smallworld software (in particular, Magik primitives) which are
 * note required to be public in a conventional vanilla acpt program.
 */

/*
 * Note: in all application programs which use acpt, the #defined name
 * 'xstatic' is defined to 'static', so sensible encapsulation is preserved.
 * For building internal smallworld code, 'xstatic' is #defined to 'extern',
 * so more internal functions are available.
 */
#if defined(_ACPTM_C_ONLY_)
#define xstatic extern
#else
#define xstatic static
#endif

#define ACPT_INT_SIZE 4            /* must be true for everything to work */

typedef unsigned char BuffEl; /* element of a buffer (needed for AllocType  */
typedef BuffEl       *Buffer; /* for all data transfer  */


typedef struct AcptTypeField {
    /*
     * A structure/union field of a type is referred to in conjunction with
     * the details of where information is to go. Thus this structure
     * is effectively an intersection between owning type and owned
     * type.
     */
    Int32 offset;         /* Offset of this component in its owner */
    Int32 extra_info;           /* Offset of unset flag, or discrim value */
    Int32 inline_size;     /* The number of bytes in the body of the struct */
    UInt32 flags;    /* Says things such as whether Ref or Inline */
    struct AcptTypeDesc *subtype; /* pointer to the subtype itself */
#if defined(_ACPTM_C_ONLY_)
    void *unset_value;          /* For datastore C _unset handling */
#endif
} AcptTypeField;



typedef struct AcptTypeDesc {
    /*
     * This struct describes a particular application respository for
     * information being transferred to or from C. It is in two parts:
     * the first part is common for local types and types read from the
     * stream; the second part is only used for local types (and has no
     * space reserved for stream types.
     */
    AcptType base_id;          /* Equivalent most basic type. Note that most
                 * operations use BASE_ID, not ID, to perform
                 * comparisons etc, as this value typically ignores
                 * local structuring concepts (eg REF).
                 */

    UInt32 matching_read_seq;  /* Used as a fast mechanism to determine whether
                 * an incoming stream type matches this type. We
                 * can't compare type_ids, because there is no
                 * correspondence, so what we do instead is set
                 * this field to a unique value in any locally
                 * defined type, and when an incoming type matches
                 * it then that has it's (stream-local) value set
                 * to the same value. If an incoming type on a
                 * particular stream has a matching_read_seq which
                 * matches a local one, then the types themselves
            * match.
                 */

    UInt32 type_detail;        /* Flags saying whether vector, struct etc */
    Int32 data_detail;         /* Length, length_offset or discrim_offset */

    struct AcptTypeDesc *el_subtype;   /* If vector, the subtype of its elements */

    Int32 n_fields;         /* The number of fields if struct or union */
    AcptTypeField *fields;     /* References to those fields */

    /*
     * The remaining types are only used in local types, not in ones
     * read from the stream. NOTE: ID is used to mark the break,
     * expressed in macro ACPT_STREAM_TYPE_SIZE.
     */
    AcptType id;         /* _self's id, for fast reverse lookup */
    Int32 n_refs;        /* Ref counted, so can be GC'd */
    struct AcptTypeMT *MT;     /* Type specific methods for Read, Write etc */

    Int32 size;             /* Size of one of this type in C */
    Int32 stream_size;         /* The size of one scalar element on the stream */

    AcptType owner_id;         /* Types are 'shared', so if two requests are made
                 * for the same shape of type the same use will be
                 * used again. To do this, a type has a list of
                 * 'flavours' of itself (so an integer may have various
                 * vector flavours). These are scanned for a match
                 * when creating a new type. OWNER_ID is the back
                 * reference... it is an ID, because the thing from
                 * which it was originally derived is allowed to die
                      * before the flavour itself dies.
                      */
    Int32 n_flavours;          /* Number of different flavours of this type  */
    struct AcptTypeDesc **flavours; /* List of pointers to different flavours */

    Int32 data_offset;         /* Offset in owning struct if V vector or union  */
    Int32 compound_length;     /* If not 0, raw copier will move this length */
} AcptTypeDesc;


/*
 * Type-based functions which know how to move data of a specific type.
 */
typedef struct AcptTypeMT {
    /*
     * 'Method table of functions used to copy data of a particular type,
     * size it and free space for it. Some fields may be NULL in which case
     * no call is made.
     */
    AcptResult (*Read)(AcptTypeDesc *type,      /* Read from stream into C */
             AcptStream *stream, Buffer c_addr,
             Int32 n_els);

    AcptResult (*Write)(AcptTypeDesc *type,     /* Write from C to stream */
         Buffer c_addr, AcptStream *stream,
         Int32 n_els);

    AcptResult (*Free)(AcptTypeDesc *type,      /* Free alloc'd memory at C_ADDR */
             Buffer c_addr, Int32 n_els); /* May be NULL */

    AcptResult (*stringLength)(Buffer c_addr, Int32 *length, Int32 max_length);/* Returns length of string, including NULL */
} AcptTypeMT;


xstatic AcptResult acptNewType(AcptTypeDesc *owner_type, AcptTypeDesc *temp_type,
                AcptTypeDesc **new_type, Bool internal);

xstatic AcptResult acptPrepareToReadObject(AcptStream *stream, AcptType type,
                  Bool want_vector, Int32 *length,
                  AcptTypeDesc **type_desc);
xstatic AcptResult acptPrepareToWriteObject(AcptStream *stream, AcptType type,
                   Int32 length, Bool is_vector,
                   AcptTypeDesc **type_desc);

xstatic AcptResult acptGetTypeDesc(AcptType type, AcptTypeDesc **type_desc);

xstatic AcptResult acptSendBuffer(AcptStream *stream);
xstatic AcptResult acptReceiveBuffer(AcptStream *stream);
xstatic AcptResult acptMakeStreamSpaceForHelper(AcptStream *stream, Int32 size);
xstatic AcptResult acptReadUInt32(AcptStream *stream, UInt32 *buffer, Int32 n_ints);
xstatic AcptResult acptWriteUInt32(AcptStream *stream, UInt32 *buffer, Int32 n_ints);

xstatic AcptResult acptUnsendTypes(AcptStream *stream);


xstatic AcptResult acptWriteScalars(AcptTypeDesc *type, Buffer from_addr,
                AcptStream *stream, Int32 n);
xstatic AcptResult acptReadScalars(AcptTypeDesc *type, AcptStream *stream,
               Buffer to_addr, Int32 n);
xstatic AcptResult acptRead1_2(AcptTypeDesc *type, AcptStream *stream, Buffer to_addr,
                Int32 n);
xstatic AcptResult acptWrite1_2(AcptTypeDesc *type, Buffer from_addr,
            AcptStream* stream, Int32 n);

xstatic AcptResult acptStringLength1(Buffer buffer, Int32 *length, Int32 max_length);
xstatic AcptResult acptStringLength2(Buffer buffer, Int32 *length, Int32 max_length);

xstatic AcptResult acptReadBigInt(AcptTypeDesc *type, AcptStream *stream,
              Buffer to_addr, Int32 n);
xstatic AcptResult acptWriteBigInt(AcptTypeDesc *type, Buffer from_addr,
              AcptStream *stream, Int32 n);
xstatic AcptResult acptFreeBigInt(AcptTypeDesc *type, Buffer from_addr, Int32 n_els);


xstatic AcptResult acptExcp(AcptResult result);

#define acptDoRead(_type_desc, _stream, _datums, _n_els) \
  (*(_type_desc)->MT->Read)(_type_desc, _stream, _datums, _n_els)

#define acptDoWrite(_type_desc, _datums, _stream, _n_els) \
  (*(_type_desc)->MT->Write)(_type_desc, _datums, _stream, _n_els)

#define acptDoFree(_type_desc, _datums, _n_els) \
  (_type_desc->MT->Free) ? (*(_type_desc)->MT->Free)(_type_desc, _datums, _n_els) : ACPT_OK

xstatic AcptTypeDesc *builtinTypeDescList[ACPT_MAX_BUILTIN_TYPES];
xstatic UInt32 matchSeq;   /* the source of type_desc.matching_read_seq,
             * incremented for each new C type made */


/*
 * Flags which describe the specific details of a type.
 */

/*
 * Bits are as follows:
 *     8         7       6       5      4            3     2    1    0
 *  Vec-ref? | Indir? Unused Strct? Has-Strct? | String? Long? Vlen? Vec?
 *
 * The bottom 6 bits are interpreted the same everywhere, and describe
 * the structure on the stream as well as the strucure in
 * memory. Indir? is used as the Ref flag in the ACP, and the
 * undiscriminated union flag in Magik, and says that checking should
 * go via subtype (ie this type is just a holder). The higher bits are
 * specific to the local implementation (ie C or Magik), and won't be
 * looked at by generic code.
 */
#define ACPT_VECTOR_MASK   0x1
#define ACPT_V_VECTOR_MASK 0x2
#define ACPT_LONG_MASK     0x4
#define ACPT_STRING_MASK   0x8

#define ACPT_STRUCT_MASK   0x30
#define ACPT_HAS_STRUCTURE_MASK  0x10
#define ACPT_STRUCT     0x30
#define ACPT_UNION      0x10

#define ACPT_INDIR_MASK         0x80

#define ACPT_UNDISCRIMINATED_MASK ACPT_INDIR_MASK
#define ACPT_UNDISCRIMINATED_UNION (ACPT_UNDISCRIMINATED_MASK | ACPT_UNION)

#define ACPT_REF_MASK      ACPT_INDIR_MASK

#define ACPT_VECREF_MASK        0x100

/* Internal type tags */

#define ACPT_V_VECTOR      0x3




#define acptIsScalar(_flags)     (((_flags) & ACPT_VECTOR_MASK) != ACPT_VECTOR_MASK)
#define acptIsVector(_flags)     (((_flags) & ACPT_VECTOR_MASK) == ACPT_VECTOR_MASK)
#define acptIsFixedVector(_flags)  (((_flags) & (ACPT_VECTOR_MASK | ACPT_V_VECTOR_MASK)) == ACPT_FIXED_VECTOR)
#define acptIsVVector(_flags)      (((_flags) & (ACPT_VECTOR_MASK | ACPT_V_VECTOR_MASK)) == ACPT_V_VECTOR)
#define acptVectorIsShort(_flags)  (((_flags) & ACPT_LONG_MASK) != ACPT_LONG_MASK)
#define acptIsString(_flags)       (((_flags) & ACPT_STRING_MASK) == ACPT_STRING_MASK)
#define acptIsBoundedString(_flags) (((_flags) & (ACPT_STRING_MASK | ACPT_V_VECTOR_MASK)) == ACPT_STRING_MASK)
#define acptIsRef(_flags)          (((_flags) & ACPT_REF_MASK) == ACPT_REF_MASK)
#define acptIsVecRef(_flags)       (((_flags) & ACPT_VECREF_MASK) == ACPT_VECREF_MASK)

#define acptHasStructure(_flags)   (((_flags) & ACPT_HAS_STRUCTURE_MASK) == ACPT_HAS_STRUCTURE_MASK)
#define acptIsStruct(_flags)     (((_flags) & ACPT_STRUCT_MASK) == ACPT_STRUCT)
#define acptIsUnion(_flags)        (((_flags) & ACPT_STRUCT_MASK) == ACPT_UNION)
#define acptIsDiscriminated(_flags)(((_flags) & ACPT_UNDISCRIMINATED_MASK) != ACPT_UNDISCRIMINATED_MASK)

#define acptIsIndirect(_flags)     (((_flags) & ACPT_INDIR_MASK) == ACPT_INDIR_MASK)
#define acptSizeIsEmbedded(_flags) (((_flags) & ACPT_LONG_MASK) != ACPT_LONG_MASK)

#define acptSkipIndirection(_desc) \
    ((acptIsIndirect((_desc)->type_detail)) ? (_desc)->el_subtype : (_desc))


/* Interesting type flags are the bottom 6 bits */
#define acptInterestingFlags(_flags) ((_flags) & 0x3f)

/* Interesting field flags are the bottom 8 bits */
#define acptInterestingFieldFlags(_flags) ((_flags) & 0xff)

/*
 * malloc type wrappers. These are written to take an *address* to be populated
 * with the result (even though, as macros, this isn't strictly necessary) as
 * it would make converting them to functions trivial. Note also (a), that
 * memory exhausted yields an error return, not simply a NULL value; (b)
 * AllocType() clears the memory (like calloc), because this makes cleanup
 * more bombproof; and (c) Free() can legitimately take a NULL pointer (as
 * this cleans up the calling); (d) you can legitimately ask for 0 bytes.
 */

#define acptAllocType(_n, _type, _loc) \
   ((*(_loc) = (_type *)calloc((_n) ? (_n) : 1, sizeof(_type))) ? ACPT_OK : acptExcp(ACPT_EXCP_MEMORY_EXHAUSTED))

#define acptRealloc(_n, _type, _loc) \
   ((*(_loc) = *(_loc) ? (_type *)realloc(*(_loc), ((_n) ? (_n) : 1)*sizeof(_type)) : (_type *)malloc(((_n) ? (_n) : 1)*sizeof(_type))) ? ACPT_OK : acptExcp(ACPT_EXCP_MEMORY_EXHAUSTED))

/* acptFree() is defined in acpt.h, as it is required by queryStreamType */

/*
 * Helper macros for error handling: all functions return a result flag - if
 * this isn't OK then jump to label 'errors'.
 */
#define ACPT_CHECKF(_to_do, _result) if ((_result = (_to_do)) != ACPT_OK) goto errors; else
#define ACPT_CHECKR(_to_do, _result) if ((_result = (_to_do)) != ACPT_OK) return result; else
#define ACPT_CHECK(_ok, _excp, _result) if (!(_ok)) {_result=acptExcp(_excp); goto errors;} else


/*
 * This code does not itself write to a physical stream, save at the
 * very beginning or end of an operation when a vector of bytes is
 * read or transmitted. The data is actually operated on within the
 * AcptStream buffer. These macros control its state.
 */
#define acptInvalidateCurrentObject(_stream)  (_stream)->object_current = FALSE
#define acptInitStreamForWrite(_stream) \
  {(_stream)->buffer = (_stream)->buffer_start+ACPT_INT_SIZE; stream->buffer_count_offset=0;}
#define acptCheckStreamLeft(_stream, _n) \
(((_stream)->buffer+(_n) <= (_stream)->buffer_start+(_stream)->buffer_count) ? ACPT_OK : acptExcp(ACPT_EXCP_STREAM_MANGLED))
#define acptMakeStreamSpaceFor(_stream, _size) \
    (((_stream)->buffer+(_size) <= (_stream)->buffer_start+(_stream->buffer_max)) ? ACPT_OK : acptMakeStreamSpaceForHelper(_stream, _size))

#define acptClearSentThisGo(stream) \
   memset(stream->SentTypesThisGo, 0, stream->SentTypesMax)

#if defined(ACPTM_C_ONLY)
/* When included as part of Magik, used to initialise the Magik-specific
 * portions of the stream structure.
 */
xstatic Void acptmSetupStream(AcptStream *stream);
xstatic Void acptmFreeStream(AcptStream *stream);
#endif

/*********************************************************************
/**** END acpt_private.h
/*********************************************************************/

/*********************************************************************
/**** BEGIN acpt_sectors.h
/*********************************************************************/

/*
 * acpt_sector.h: Contains type information specific to geometry.
 */

/*
 * Sectors are very special. There are currently 6 flavours, which are
 * then bunched together into a union, discriminated by 'type' (which
 * will have the same encoding as used internally in the GIS). Note that
 * although circle and arc could have been implemented as fixed vectors,
 * this would have meant that NControlPoints wouldn't have been set.
 */
#define ACPT_TYPE_COORD       32 /* 2D coordinate */
#define ACPT_TYPE_COORD3D     33 /* 2D coordinate with Z (ie ~3D) */
#define ACPT_TYPE_COORD2H     34 /* 2D homogeneous coordinate with W */

#define ACPT_TYPE_SIMPLE_SECTOR     36       /* vvec of 2D coords */
#define ACPT_TYPE_SECTOR_Z    37 /* vvec of 3D coords */
#define ACPT_TYPE_SECTOR_H    38 /* vvec of 2D tan pt arcs with w's */
#define ACPT_TYPE_CP_ARC                39   /* vvec of 2D centre-pt arcs with w's */
/*
 * the next two aren't strictly necessary, except to make the Magik
 * work properly
 */
#define ACPT_TYPE_CIRCLE            40 /* C-point circle... same as sector */
#define ACPT_TYPE_TAN_ARC           41 /* vvec of 2D tan-pt arcs with w's */

#define ACPT_TYPE_SECTOR      42 /* Union of the above */

#define ACPT_TYPE_SECTOR_ROPE    43 /* VVec of SECTORs */

#define ACPT_TYPE_GIS_ID                44      /* Conventional GIS ID */

typedef struct {double x; double y;} AcptCoord;
typedef struct {double x; double y; double z;} AcptCoord3D;
typedef struct {double x; double y; double w;} AcptCoord2H;

typedef struct {AcptCoord   *coords; Int32 NCoords;} AcptSimpleSector;
typedef struct {AcptCoord3D *coords; Int32 NCoords;} AcptSectorZ;
typedef struct {AcptCoord2H *coords; Int32 NCoords;} AcptSectorH;

typedef struct {
    Int32 type;               /* type flags */
    union {
   AcptSimpleSector SimpleSector;
   AcptSectorZ      SectorZ;
   AcptSectorH  SectorH;
    } u;
} AcptSector;


#define acptIsCircle(_sectorp) (((_sectorp)->type & 2048) == 2048)
#define acptIsSectorZ(_sectorp) (((_sectorp)->type & 128) == 128)
#define acptIsSectorH(_sectorp) (!acptIsCircle(_sectorp) && ((_sectorp)->type & 1) == 1)
#define acptIsSectorTanArc(_sectorp) (((_sectorp)->type & 1037) == 1037)
#define acptIsSectorCPArc(_sectorp) (((_sectorp)->type & 3085) == 3085)
#define acptIsSimpleSector(_sectorp) (!(acptIsSectorZ(_sectorp) || ((_sectorp)->type & 1) == 1))

#define acptNCoords(_sectorp) ((_sectorp)->u.SimpleSector.NCoords)
#define acptCoords(_sectorp)    ((AcptCoord *)(_sectorp)->u.SimpleSector.coords)

typedef struct {
    AcptSector *Sectors;
    Int32 NSectors;
} AcptSectorRope;

typedef UInt32 AcptGIS_ID[3];

/*
 * Discriminators for different sorts of sector.
 */
#define ACPT_SIMPLE_SECTOR_DISCRIM     8
#define ACPT_SECTOR_Z_DISCRIM        136
#define ACPT_SECTOR_H_DISCRIM         13
#define ACPT_CIRCLE_DISCRIM         2061
#define ACPT_SECTOR_TAN_ARC_DISCRIM 1037
#define ACPT_SECTOR_CP_ARC_DISCRIM  3085

static AcptUnionElement SectorDef[] = {
    {{ACPT_TYPE_SIMPLE_SECTOR}, sizeof(AcptSimpleSector),    8},  /* straight */
    {{ACPT_TYPE_SECTOR_Z},      sizeof(AcptSectorZ),       136},  /* straight 3D */
    {{ACPT_TYPE_SECTOR_H},      sizeof(AcptSectorH),        13},  /* rat. B-Spline */
    {{ACPT_TYPE_CIRCLE},        sizeof(AcptSimpleSector), 2061},  /* circle */
    {{ACPT_TYPE_TAN_ARC},       sizeof(AcptSectorH),      1037},  /* tangent arc */
    {{ACPT_TYPE_CP_ARC},        sizeof(AcptSectorH),      3085}}; /* centre-point arc */

/*********************************************************************
/**** END acpt_sector.h
/*********************************************************************/

/**********************************************************************
/######################################################################
/#### acpt_tics implementation
/######################################################################
/*********************************************************************/

/*********************************************************************
/**** BEGIN acpt_tics.h
/*********************************************************************/

/* Public parts of TicsHandle: the acpt read and write streams */

#define TH_WRITE(_th) ((AcptStream *)((_th)->write_stream))
#define TH_READ(_th)  ((AcptStream *)((_th)->read_stream))


extern AcptResult ACPT_DLL_CALL ticsMakeAcptStreams(TicsHandle thHandle);
/*
 * Enable Tics THHANDLE to use acpt functionality. This function
 * *must* be called for any handle which is subsequently going to be
 * used with acpt or 'Object' interfaces.
 *
 * When a handle is no longer required, ticsFreeAcptStreams() must be
 * called to free resources. Note, however, that this does *not* close
 * any physical connections, and conventional tics protocols must be
 * used to do this.
 *
 * Note that this interface does *not* do acptInit() [because this
 * must only be done once], and it is the caller's responsibility to
 * call acptInit() at the very beginning of the program.
 */


extern AcptResult ACPT_DLL_CALL ticsFreeAcptStreams(TicsHandle thHandle);
/*
 * Free any acpt resources associated with THHANDLE. This function
 * should be called when THHANDLE is about to be discarded, but note
 * that it does *not* close any physical connections, and conventional
 * tics protocols must be used to do this.
 */



/*
 * Tics wrappers for acpt stream functions
 */

#define ticsPutObject(_th, _type, _datum) \
        acptWriteObject(TH_WRITE(_th), _type, _datum)
/*
 * Send the object at DATUM which matches type TYPE to TH. Note that
 * any errors which arise are non-destructive - in all cases other than
 * the underlying stream itself being unwritable, nothing is sent.
 */


#define ticsPutObjectVector(_th, _type, _datums, _length) \
        acptWriteObjectVector(TH_WRITE(_th), _type, _datums, _length)
/*
 * Send LENGTH objects in DATUMS which match type TYPE to TH.
 *
 * The objects in DATUMS are assumed to be in-line (ie not references).
 *
 * Note that any errors which arise are non-destructive - in all
 * cases other than the stream itself being unwritable, nothing is
 * sent.
 */


#define ticsGetObject(_th, _type, _datum, _datum_size) \
        acptReadObject(TH_READ(_th), _type, _datum, _datum_size)
/*
 * Retrieve the next object, which must be compatible with TYPE, into
 * the space at address DATUM which is DATUM_SIZE bytes long, from
 * TH. The data must have been sent using putObject functionality: if
 * putObjectVector functionality was used, then getObjectVector must
 * be used to retrieve the data.
 *
 * Note: In the majority of cases where this function returns an
 * exception, the inbound type can queried (using acpPeekObject()) and
 * another attempt made to get the data (eg using getObjectVector, or
 * providing a different type or size) - (so read-failure at this
 * juncture is typically non-destructive).
 */


#define ticsGetObjectVector(_th, _type, _datums, _space, _n_objects) \
        acptReadObjectVector(TH_READ(_th), _type, _datums, _space, _n_objects)
/*
 * Retrieve a homogeneous vector of objects, which must be of a type
 * compatible with TYPE, into the space at address DATUMS which is
 * SPACE bytes long from TH. The data must have been sent using
 * putObjectVector functionality: if putObject functionality was used,
 * then getObject must be used to retrieve the data. N_OBJECTS (if not
 * NULL) will be populated with the number of objects retrieved.
 *
 * DATUMS will be populated with in-line objects.
 *
 * Note: In the majority of cases where this function returns an
 * exception, the inbound type can queried (using acpPeekObject()) and
 * another attempt made to get the data (eg using getObject, or
 * providing a different type or size) - (so read-failure at this
 * juncture is typically non-destructive).
 */

#define ticsPeekObject(_th, _typeno, _is_vectorp, _lengthp) \
        acptPeekObject(TH_READ(_th), _typeno, _is_vectorp, _lengthp)
/*
 * Interrogates the item which would be consumed by the next call to
 * GetObject, GetObjectVector or SkipObject from TH. This call is
 * non-destructive, and may be made repeatedly before one of the above
 * calls, and also after a failed call to GetObject or
 * GetObjectVector. TYPENO is populated with the *stream type* - this
 * will only match a local type if it is one of the built-in ones;
 * IS_VECTORP is populated with TRUE if GetObjectVector() is required
 * to get the object, FALSE if GetObject is required; and LENGTHP with
 * the length if IS_VECTORP is TRUE. Any of these three arguments may
 * legitimately be NULL. Note that under some circumstances, if
 * IS_VECTORP is TRUE and if LENGTHP is 0, TYPENO will be
 * ACPT_TYPE_EMPTY_VECTOR.
 */

#define ticsSkipObject(_th) acptSkipObject(TH_READ(_th))
/*
 * Skip the next or pending (if the previous request was a failed Get)
 * object (or vector of objects) from TH.
 */

#define ticsQueryStreamType(_th, _typeno, _flavour, _data_detail, _subtype, _subtypes) \
         acptQueryStreamType(TH_READ(_th), _typeno, _flavour, _data_detail, _subtype, _subtypes)
/*
 * Returns details of the incoming type identified by TYPENO from TH.
 * FLAVOUR will be populated with the flags which define what sort of
 * type this is, and will be one of:
 *     ACPT_SCALAR
 *     ACPT_FIXED_VECTOR
 *     ACPT_SHORT_V_VECTOR
 *     ACPT_LONG_V_VECTOR
 *     ACPT_SHORT_BOUNDED_STRING
 *     ACPT_LONG_BOUNDED_STRING
 *     ACPT_SHORT_V_STRING
 *     ACPT_LONG_V_STRING
 *     ACPT_STRUCT
 *     ACPT_UNION
 *
 * DATA_DETAIL will be populated with a value depending on
 * FLAVOUR: if FLAVOUR is is ACPT_FIXED_VECTOR or
 * ACPT_[LONG/SHORT]_BOUNDED_STRING then DATA_DETAIL is populated
 * with the length value; if FLAVOUR is ACPT_STRUCT or ACPT_UNION
 * then it is populated with the number of subtypes. If FLAVOUR
 * indicates a vector type, SUBTYPE is populated with the typeno
 * of the subtype (otherwise it's NULLed). If FLAVOUR is
 * ACPT_STRUCT or ACPT_UNION, SUBTYPES is populated with an
 * integer vector containing the typnos or the underlying fields
 * and followed by the same number of elements containing flags
 * (indicating _unset allowed, if FLAVOUR is ACPT_STRUCT) or
 * discriminators (if FLAVOUR is ACPT_UNION) - so the vector is
 * 2*data_detail long.
 *
 * If SUBTYPES is not relevant the argument is populated with
 * NULL; if it is relevant then it is the caller's responsibility
 * to free the space using acptFree() when no longer required.
 */

/*********************************************************************
/**** END acpt_tics.h
/*********************************************************************/

