
#include "acpt_tics.h"
#include <math.h>

// class coord2 is same as AcptCoord
// but has some convenient constructors and operators

class coord2
{
  public:
    double x, y;

    coord2(){};
    coord2(double X, double Y){ x=X; y=Y; };
    coord2(AcptCoord p){ x=p.x; y=p.y; };
    void Set(double X, double Y){ x=X; y=Y; };

    coord2& operator = (const coord2& p)
      { x=p.x; y=p.y;
        return(*this);  };
    coord2 operator + (coord2 p)
      { coord2 p2(x+p.x, y+p.y);
        return(p2); };
    coord2 operator - (coord2 p)
      { coord2 p2(x-p.x, y-p.y);
        return(p2); };
    coord2 operator / (double s)
      { coord2 p2(x/s, y/s);
        return(p2); };

    double distance_to(coord2 p2)
      { double xd = p2.x - x;
        double yd = p2.y - y;
        return sqrt((xd * xd) + (yd * yd));
      }
    coord2 normalized()
      { coord2 p2(x, y);
        return (p2 / p2.distance_to(coord2(0.0, 0.0)));
      }
    double cross(coord2 p2)
      { return (x * p2.y - p2.x * y);
      }
    double angle_of()
      {
         // Return angle of normalized coord2
         // where -pi <= t <= pi

         double t = asin(y);
         if (x < 0)
         {
            double pi = acos( (double) -1);
            t = pi - t;
         }
         return t;
      }

};

// class coord2h is same as AcptCoord2H
// but has some convenient constructors and operators

class coord2h
{
  public:
    double x, y, w;

    coord2h(){};
    coord2h(double X, double Y, double W){ x=X; y=Y; w=W; };
    coord2h(AcptCoord2H p){ x=p.x; y=p.y; w=p.w; };
    void Set(double X, double Y, double W){ x=X; y=Y; w=W; };
    coord2 as_coord2()
       { coord2 p2(x / w, y / w);
        return (p2); };

    coord2h& operator = (const coord2h& p)
      { x=p.x; y=p.y; w=p.w;
        return(*this);  };
    coord2h operator + (coord2h p)
      { coord2h p2(x+p.x, y+p.y, w+p.w);
        return(p2); };
    coord2h operator * (double s)
      { coord2h p2(x*s, y*s, w*s);
        return(p2); };
};

#define DEFAULT_NSEGS        32
// #define DEFAULT_TOL          0.001
#define DEFAULT_SUBDIVS      5

class SGrinder
{
public:

   // constructor

   SGrinder(){};

   // operations

   int grind(AcptSector * s, AcptCoord ** clist, int numsegs);

private:

   // attributes

   int NSegs;
// double Tol;          not implemented at this time
   coord2h * cPts;
   coord2 * c2Pts;
   int NumCtrlPts;
   double * Knots;
   int NumKnots;
   AcptCoord * gPts;
   int NumPts;

   // operations

   void grind_circle();
   void grind_arc();
   bool arc_is_clockwise();
   void get_ellipse_parameters(coord2 p1, coord2 p2, double *A, double *B, double *sin_t, double *cos_t);
   void grind_bezier();
   void grind_bspline();
   double R(int i, int k, double u);
   double B(int k, int m, double t);
   int factorial(int n);

};


